Package: ppforest2
Type: Package
Title: Projection Pursuit Oblique Decision Trees and Random Forests
Version: 0.1.1
Date: 2026-07-10
Authors@R: c(
    person("Andrés", "Vidal", email = "andres@andresvidal.dev", role = c("aut", "cre", "cph")),
    person("Natalia", "da Silva", role = "aut"))
Description: Builds decision trees by splitting on linear combinations of
    randomly chosen variables. Projection pursuit is used to choose a
    projection of the variables that best separates the groups. Using linear
    combinations of variables to separate groups takes the correlation between
    variables into account, which allows the model to outperform a traditional
    decision tree when the separation between groups occurs in combinations of
    variables. Single trees can be assembled into random forests for improved
    accuracy. Implements projection pursuit classification trees (Lee, Cook,
    Park and Lee (2013) <doi:10.1214/13-EJS810>) and projection
    pursuit forests (da Silva, Cook and Lee (2021)
    <doi:10.1080/10618600.2020.1870480>), following the earlier 'PPforest'
    package.
License: MIT + file LICENSE
Imports: Rcpp (>= 1.0.11)
LinkingTo: Rcpp, RcppEigen
Depends: R (>= 3.5.0)
LazyData: true
RoxygenNote: 7.3.2
Encoding: UTF-8
Suggests: ggplot2, jsonlite, knitr, parsnip, patchwork, rlang,
        rmarkdown, rsample, testthat (>= 3.0.0), tibble, tune, vdiffr,
        withr, workflows, yardstick
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/testthat/parallel: false
NeedsCompilation: yes
Packaged: 2026-07-10 20:07:48 UTC; andres
Author: Andrés Vidal [aut, cre, cph],
  Natalia da Silva [aut]
Maintainer: Andrés Vidal <andres@andresvidal.dev>
Repository: CRAN
Date/Publication: 2026-07-19 13:10:07 UTC

# Generated by roxygen2: do not edit by hand

S3method(bag_samples,default)
S3method(bag_samples,pprf)
S3method(fitted,ppmodel)
S3method(formula,ppmodel)
S3method(nobs,ppmodel)
S3method(oob_error,default)
S3method(oob_error,pprf)
S3method(oob_predictions,default)
S3method(oob_predictions,pprf_classification)
S3method(oob_predictions,pprf_regression)
S3method(oob_samples,default)
S3method(oob_samples,pprf)
S3method(permuted_importance,default)
S3method(permuted_importance,pprf)
S3method(plot,pprf)
S3method(plot,pptr)
S3method(predict,pprf_classification)
S3method(predict,pprf_regression)
S3method(predict,pptr_classification)
S3method(predict,pptr_regression)
S3method(print,pprf)
S3method(print,pptr)
S3method(print_node,pptr_classification)
S3method(print_node,pptr_regression)
S3method(projection_importance,default)
S3method(projection_importance,ppmodel)
S3method(residuals,ppmodel)
S3method(save_json,ppmodel)
S3method(summary,ppmodel)
S3method(summary,pprf)
S3method(summary,pprf_classification)
S3method(summary,pprf_regression)
S3method(summary,pptr)
S3method(summary,pptr_classification)
S3method(summary,pptr_regression)
S3method(update,pp_rand_forest)
S3method(update,pp_tree)
S3method(weighted_importance,default)
S3method(weighted_importance,pprf)
export(bag_samples)
export(binarize_disabled)
export(binarize_largest_gap)
export(cutpoint_mean_of_means)
export(grouping_by_cutpoint)
export(grouping_by_label)
export(leaf_majority_vote)
export(leaf_mean_response)
export(load_json)
export(oob_error)
export(oob_predictions)
export(oob_samples)
export(permuted_importance)
export(pp_pda)
export(pp_rand_forest)
export(pp_tree)
export(pprf)
export(pptr)
export(projection_importance)
export(save_json)
export(stop_any)
export(stop_max_depth)
export(stop_min_size)
export(stop_min_variance)
export(stop_pure_node)
export(vars_all)
export(vars_uniform)
export(weighted_importance)
importFrom(Rcpp,evalCpp)
importFrom(stats,fitted)
importFrom(stats,formula)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.response)
importFrom(stats,nobs)
importFrom(stats,predict)
importFrom(stats,residuals)
importFrom(stats,sd)
importFrom(stats,terms)
importFrom(stats,update)
useDynLib(ppforest2)

# ppforest2

<!-- badges: start -->
[![R-CMD-check](https://github.com/andres-vidal/ppforest2/actions/workflows/run-r-check.yml/badge.svg)](https://github.com/andres-vidal/ppforest2/actions/workflows/run-r-check.yml)
<!-- badges: end -->

ppforest2 provides projection pursuit oblique decision trees and random
forests for classification.  Instead of splitting on single variables,
each node projects the data onto a linear combination of features,
capturing structure that axis-aligned trees miss.

The package wraps a high-performance C++ core and is intended as a modern
successor to [`PPforest`](https://cran.r-project.org/package=PPforest).

Key capabilities: oblique splits via projection pursuit, multi-threaded
forest training (OpenMP), cross-platform reproducibility, three variable
importance measures (projection-based, weighted, permutation), LDA/PDA
optimisation, OOB error estimation, and
[parsnip](https://parsnip.tidymodels.org/) / tidymodels integration.

## Installation

```r
# install.packages("devtools")
devtools::install_github("andres-vidal/ppforest2", subdir = "bindings/R", build = FALSE)
```

## Usage

### Single tree

```r
library(ppforest2)

model <- pptr(Species ~ ., data = iris)
predict(model, iris[1:5, ])
summary(model)
```

### Random forest

```r
forest <- pprf(Species ~ ., data = iris, size = 500)
predict(forest, iris[1:5, ])
predict(forest, iris[1:5, ], type = "prob")   # vote proportions
summary(forest)
```

### Regularisation (PDA)

When classes are highly correlated or the number of variables is large
relative to the sample size, penalised discriminant analysis can improve
separation:

```r
pptr(Species ~ ., data = iris, lambda = 0.5)
```

### Visualisation

ppforest2 provides four diagnostic plot types (requires
[ggplot2](https://ggplot2.tidyverse.org/)):

```r
# Mosaic overview: structure + importance + boundaries
plot(model)

# Individual plot types
plot(model, type = "structure")     # tree diagram with per-node histograms
plot(model, type = "importance")    # variable importance bar chart
plot(model, type = "projection")    # projected data at each split
plot(model, type = "boundaries")    # decision boundaries in feature space

# Forest: importance across all trees, or inspect individual trees
plot(forest)
plot(forest, type = "structure", tree_index = 1)
plot(forest, type = "boundaries", tree_index = 1)
```

### tidymodels integration

ppforest2 integrates with [parsnip](https://parsnip.tidymodels.org/):

```r
library(parsnip)

# Single tree
spec <- pp_tree(lambda = 0) |> set_engine("ppforest2") |> set_mode("classification")
fit  <- fit(spec, Species ~ ., data = iris)

# Random forest
spec <- pp_rand_forest(trees = 50, mtry = 2) |> set_engine("ppforest2")
fit  <- spec |> fit(Species ~ ., data = iris)
predict(fit, iris, type = "prob")
```

### JSON serialisation

Models can be saved and loaded in JSON format, enabling interoperability
with the C++ CLI and other language bindings:

```r
save_json(model, "model.json")
restored <- load_json("model.json")
```

## Learning more

- `vignette("introduction")` — a tutorial covering trees, forests,
  visualisation, and tidymodels integration.
- [C++ API Reference](https://andres-vidal.github.io/ppforest2/main/cpp/) —
  core library documentation (Doxygen).
- [GitHub repository](https://github.com/andres-vidal/ppforest2) —
  source code, build instructions, and benchmarks.

## ----setup, include = FALSE---------------------------------------------------
# Cap OpenMP to 2 threads so forest training in this vignette never uses more
# than two cores when the vignette is rebuilt under R CMD check (CRAN policy).
Sys.setenv(OMP_NUM_THREADS = "2")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----tree---------------------------------------------------------------------
library(ppforest2)

tree <- pptr(Species ~ ., data = iris, seed = 0)
tree

## ----tree-summary-------------------------------------------------------------
summary(tree)

## ----tree-predict-------------------------------------------------------------
preds <- predict(tree, iris)
table(predicted = preds, actual = iris$Species)

## ----forest-------------------------------------------------------------------
forest <- pprf(Species ~ ., data = iris, size = 100, n_vars = 2, seed = 0)
summary(forest)

## ----forest-predict-----------------------------------------------------------
probs <- predict(forest, iris[1:5, ], type = "prob")
probs

## ----plot-structure, eval = requireNamespace("ggplot2", quietly = TRUE)-------
# Tree structure with projected data histograms at each node
plot(tree, type = "structure")

## ----plot-importance, eval = requireNamespace("ggplot2", quietly = TRUE)------
# Variable importance bar chart
plot(tree, type = "importance")

## ----plot-projection, eval = requireNamespace("ggplot2", quietly = TRUE)------
# Data projected onto the first split's projection vector
plot(tree, type = "projection")

## ----plot-boundaries, eval = requireNamespace("ggplot2", quietly = TRUE)------
# Decision boundaries for two selected variables
plot(tree, type = "boundaries")

## ----plot-forest, eval = requireNamespace("ggplot2", quietly = TRUE)----------
plot(forest)

## ----plot-forest-tree, eval = requireNamespace("ggplot2", quietly = TRUE)-----
plot(forest, type = "structure", tree_index = 1)

## ----pda----------------------------------------------------------------------
tree_pda <- pptr(Species ~ ., data = iris, lambda = 0.5, seed = 0)
summary(tree_pda)

## ----strategies---------------------------------------------------------------
# These two calls produce identical results:
forest_shortcut <- pprf(Species ~ ., data = iris, size = 10, lambda = 0.5, n_vars = 2, seed = 0)

forest_explicit <- pprf(Species ~ ., data = iris, size = 10, pp = pp_pda(0.5), vars = vars_uniform(n_vars = 2), seed = 0)

all.equal(predict(forest_shortcut, iris), predict(forest_explicit, iris))

## ----strategies-parsnip, eval = requireNamespace("parsnip", quietly = TRUE)----
library(parsnip)

spec <- pp_rand_forest(trees = 10) |>
  set_engine("ppforest2", pp = pp_pda(0.5), vars = vars_uniform(n_vars = 2)) |>
  set_mode("classification")

fit <- spec |> fit(Species ~ ., data = iris)
predict(fit, iris[1:5, ])

## ----parsnip, eval = requireNamespace("parsnip", quietly = TRUE)--------------
library(parsnip)

# Single tree
spec <- pp_tree(penalty = 0.5) |>
  set_engine("ppforest2") |>
  set_mode("classification")

fit <- spec |> fit(Species ~ ., data = iris)
predict(fit, iris[1:5, ])

## ----parsnip-forest, eval = requireNamespace("parsnip", quietly = TRUE)-------
# Random forest
spec <- pp_rand_forest(trees = 50, mtry = 2, penalty = 0.5) |>
  set_engine("ppforest2") |>
  set_mode("classification")

fit <- spec |> fit(Species ~ ., data = iris)
predict(fit, iris[1:5, ], type = "prob")

# Suppress R CMD check NOTEs for rlang expressions used in parsnip registration.
# `x` / `y` appear in `set_fit(... protect = c("x", "y"))` symbol references;
# `object` / `new_data` appear in `set_pred(... args = list(...))`.
utils::globalVariables(c("object", "new_data", "x", "y"))

#' @keywords internal
.onLoad <- function(libname, pkgname) {
  if (requireNamespace("parsnip", quietly = TRUE)) {
    register_pp_rand_forest()
    register_pp_tree()
  }
}

#' Parsnip model specification for pprf.
#'
#' Creates a model specification for a Projection Pursuit random forest.
#' Use \code{set_engine("ppforest2")} to select the ppforest2 engine.
#'
#' @param mode A character string for the model type. Either \code{"classification"} or \code{"regression"}.
#' @param trees The number of trees in the forest (maps to \code{size}).
#' @param mtry The number of variables to consider at each split (maps to \code{n_vars}).
#' @param mtry_prop The proportion of variables to consider at each split (maps to \code{p_vars}). An alternative to \code{mtry} that expresses the feature subsample as a fraction in (0, 1], tunable via the \code{dials} \code{mtry_prop()} parameter. Supply \code{mtry} or \code{mtry_prop}, not both.
#' @param penalty The regularization parameter (maps to \code{lambda}).
#' @return A parsnip model specification.
#' @seealso \code{\link{pprf}} for the underlying training function, \code{\link{pp_tree}} for single trees
#' @examples
#' \donttest{
#' if (requireNamespace("parsnip", quietly = TRUE)) {
#'   library(parsnip)
#'   spec <- pp_rand_forest(trees = 50, mtry = 2) %>% set_engine("ppforest2")
#'   fit <- spec %>% fit(Species ~ ., data = iris)
#'   predict(fit, iris)
#'   predict(fit, iris, type = "prob")
#' }
#' }
#' @export
pp_rand_forest <- function(mode = "classification", trees = NULL, mtry = NULL, mtry_prop = NULL, penalty = NULL) {
  if (!requireNamespace("parsnip", quietly = TRUE)) {
    stop("Package 'parsnip' is required for pp_rand_forest().", call. = FALSE)
  }

  args <- list(
    trees = rlang::enquo(trees),
    mtry = rlang::enquo(mtry),
    mtry_prop = rlang::enquo(mtry_prop),
    penalty = rlang::enquo(penalty)
  )

  parsnip::new_model_spec(
    "pp_rand_forest",
    args = args,
    eng_args = NULL,
    mode = mode,
    method = NULL,
    engine = NULL
  )
}

#' Update a \code{pp_rand_forest} model specification.
#'
#' Implements parsnip's \code{update} protocol so that \code{tune::tune_grid()}
#' (and any other caller that finalises a spec via \code{update()}) can fill
#' in the tuned values for \code{trees}, \code{mtry}, and \code{penalty}.
#' Without this method, \code{update()} falls back to \code{stats::update.default}
#' and fails with "need an object with call component".
#'
#' @param object A \code{pp_rand_forest} model specification.
#' @param parameters A named list of parameters to update (alternative to passing them as args).
#' @param trees,mtry,mtry_prop,penalty New values for the corresponding parameters.
#' @param fresh If \code{TRUE}, drop any previously-set arg whose new value is not provided.
#' @param ... Engine-specific arguments to update.
#' @return An updated \code{pp_rand_forest} model specification.
#' @export
update.pp_rand_forest <- function(object, parameters = NULL, trees = NULL, mtry = NULL, mtry_prop = NULL, penalty = NULL, fresh = FALSE, ...) {
  args <- list(
    trees = rlang::enquo(trees),
    mtry = rlang::enquo(mtry),
    mtry_prop = rlang::enquo(mtry_prop),
    penalty = rlang::enquo(penalty)
  )
  update_pp_spec(object, parameters = parameters, args_enquo_list = args, fresh = fresh, cls = "pp_rand_forest", ...)
}

#' Parsnip model specification for pptr.
#'
#' Creates a model specification for a single Projection Pursuit decision tree.
#' Use \code{set_engine("ppforest2")} to select the ppforest2 engine.
#'
#' @param mode A character string for the model type. Either \code{"classification"} or \code{"regression"}.
#' @param penalty The regularization parameter (maps to \code{lambda}).
#' @return A parsnip model specification.
#' @seealso \code{\link{pptr}} for the underlying training function, \code{\link{pp_rand_forest}} for forests
#' @examples
#' \donttest{
#' if (requireNamespace("parsnip", quietly = TRUE)) {
#'   library(parsnip)
#'   spec <- pp_tree(penalty = 0) %>% set_engine("ppforest2")
#'   fit <- spec %>% fit(Species ~ ., data = iris)
#'   predict(fit, iris)
#' }
#' }
#' @export
pp_tree <- function(mode = "classification", penalty = NULL) {
  if (!requireNamespace("parsnip", quietly = TRUE)) {
    stop("Package 'parsnip' is required for pp_tree().", call. = FALSE)
  }

  args <- list(
    penalty = rlang::enquo(penalty)
  )

  parsnip::new_model_spec(
    "pp_tree",
    args = args,
    eng_args = NULL,
    mode = mode,
    method = NULL,
    engine = NULL
  )
}

#' Update a \code{pp_tree} model specification.
#'
#' Companion to \code{\link{update.pp_rand_forest}} — exists for the same
#' reason (\code{tune::tune_grid()} relies on \code{update()} to finalise the
#' spec at each grid point).
#'
#' @param object A \code{pp_tree} model specification.
#' @param parameters A named list of parameters to update (alternative to passing them as args).
#' @param penalty A new value for the regularisation parameter.
#' @param fresh If \code{TRUE}, drop any previously-set arg whose new value is not provided.
#' @param ... Engine-specific arguments to update.
#' @return An updated \code{pp_tree} model specification.
#' @export
update.pp_tree <- function(object, parameters = NULL, penalty = NULL, fresh = FALSE, ...) {
  args <- list(
    penalty = rlang::enquo(penalty)
  )
  update_pp_spec(object, parameters = parameters, args_enquo_list = args, fresh = fresh, cls = "pp_tree", ...)
}

# Local re-implementation of parsnip's internal `update_spec()`. We avoid the
# triple-colon dependency to keep R CMD check clean. Behaviour mirrors
# `parsnip:::update_spec`: replace main args (and engine args from `...`)
# either by overlay (default) or completely (`fresh = TRUE`).
update_pp_spec <- function(object, parameters, args_enquo_list, fresh, cls, ...) {
  eng_dots <- rlang::enquos(...)
  if (fresh) {
    object$args <- args_enquo_list
    object$eng_args <- if (length(eng_dots)) eng_dots else NULL
  } else {
    is_null_arg <- vapply(args_enquo_list, function(q) {
      e <- rlang::quo_get_expr(q)
      is.null(e) || (is.symbol(e) && identical(e, quote(expr =)))
    }, logical(1))
    new_args <- args_enquo_list[!is_null_arg]
    if (length(new_args) > 0) {
      object$args[names(new_args)] <- new_args
    }
    if (length(eng_dots) > 0) {
      object$eng_args <- utils::modifyList(object$eng_args %||% list(), eng_dots)
    }
  }

  if (!is.null(parameters)) {
    if (!is.list(parameters) && !is.data.frame(parameters)) {
      stop("`parameters` must be a named list or one-row tibble.", call. = FALSE)
    }
    for (nm in names(parameters)) {
      val <- if (is.data.frame(parameters)) parameters[[nm]][[1]] else parameters[[nm]]
      object$args[[nm]] <- rlang::new_quosure(val, env = rlang::empty_env())
    }
  }

  parsnip::new_model_spec(
    cls,
    args = object$args,
    eng_args = object$eng_args,
    mode = object$mode,
    method = NULL,
    engine = object$engine
  )
}

register_pp_rand_forest <- function() {
  try(parsnip::set_new_model("pp_rand_forest"), silent = TRUE)
  parsnip::set_model_mode("pp_rand_forest", "classification")
  parsnip::set_model_mode("pp_rand_forest", "regression")

  for (mode in c("classification", "regression")) {
    parsnip::set_model_engine("pp_rand_forest", mode = mode, eng = "ppforest2")

    # Use the matrix interface (not formula) on purpose. With `interface =
    # "formula"`, parsnip's dispatch for x/y-input callers (workflows,
    # fit_resamples, tune_grid) is `xy_form()`, which synthesises a
    # `..y ~ x1 + x2 + ...` formula plus a data frame whose response column
    # is renamed to `..y`. `pprf()` then stores that synthetic formula on
    # the model, and prediction blows up with "object '..y' not found"
    # because the test data has no `..y` column. The matrix interface skips
    # that synthesis: parsnip's dispatch becomes `xy_xy(target = "matrix")`,
    # which calls the engine as `pprf(x = maybe_matrix(x), y = y, ...)`. We
    # also flip `predictor_indicators` from "none" to "traditional" so the
    # formula path (`form_xy()`) dummy-codes factor predictors into a numeric
    # matrix before handing them to `pprf()` — the old setup relied on
    # `pprf()` itself running `model.matrix()`, which doesn't happen on the
    # matrix path.
    parsnip::set_encoding(
      model = "pp_rand_forest",
      eng = "ppforest2",
      mode = mode,
      options = list(
        predictor_indicators = "traditional",
        compute_intercept = TRUE,
        remove_intercept = TRUE,
        allow_sparse_x = FALSE
      )
    )

    parsnip::set_fit(
      model = "pp_rand_forest",
      eng = "ppforest2",
      mode = mode,
      value = list(
        interface = "matrix",
        protect = c("x", "y"),
        func = c(pkg = "ppforest2", fun = "pprf"),
        # Forward the parsnip-declared mode to the engine function so it
        # round-trips through the wrapper instead of being re-derived from
        # `y`. parsnip's mode is the user's explicit intent; `pprf()`'s
        # `mode` arg then asserts on type compatibility (factor-y +
        # mode="regression" errors fast).
        defaults = list(mode = mode)
      )
    )
  }

  parsnip::set_model_arg(
    "pp_rand_forest", eng = "ppforest2",
    parsnip = "trees", original = "size",
    func = list(pkg = "dials", fun = "trees"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    "pp_rand_forest", eng = "ppforest2",
    parsnip = "mtry", original = "n_vars",
    func = list(pkg = "dials", fun = "mtry"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    "pp_rand_forest", eng = "ppforest2",
    parsnip = "mtry_prop", original = "p_vars",
    func = list(pkg = "dials", fun = "mtry_prop"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    "pp_rand_forest", eng = "ppforest2",
    parsnip = "penalty", original = "lambda",
    func = list(pkg = "dials", fun = "penalty"),
    has_submodel = FALSE
  )

  parsnip::set_pred(
    model = "pp_rand_forest",
    eng = "ppforest2",
    mode = "classification",
    type = "class",
    value = list(
      pre = NULL,
      post = NULL,
      func = c(fun = "predict"),
      args = list(
        object = rlang::expr(object$fit),
        new_data = rlang::expr(new_data),
        type = "class"
      )
    )
  )

  parsnip::set_pred(
    model = "pp_rand_forest",
    eng = "ppforest2",
    mode = "classification",
    type = "prob",
    value = list(
      pre = NULL,
      post = NULL,
      func = c(fun = "predict"),
      args = list(
        object = rlang::expr(object$fit),
        new_data = rlang::expr(new_data),
        type = "prob"
      )
    )
  )

  parsnip::set_pred(
    model = "pp_rand_forest",
    eng = "ppforest2",
    mode = "regression",
    type = "numeric",
    value = list(
      pre = NULL,
      post = function(results, object) tibble::tibble(.pred = as.numeric(results)),
      func = c(fun = "predict"),
      args = list(
        object = rlang::expr(object$fit),
        new_data = rlang::expr(new_data),
        type = "response"
      )
    )
  )
}

register_pp_tree <- function() {
  try(parsnip::set_new_model("pp_tree"), silent = TRUE)
  parsnip::set_model_mode("pp_tree", "classification")
  parsnip::set_model_mode("pp_tree", "regression")

  for (mode in c("classification", "regression")) {
    parsnip::set_model_engine("pp_tree", mode = mode, eng = "ppforest2")

    # Matrix interface — see the matching comment in
    # `register_pp_rand_forest` for the rationale (`..y` formula synthesis
    # in parsnip's `xy_form()` breaks prediction through workflows).
    parsnip::set_encoding(
      model = "pp_tree",
      eng = "ppforest2",
      mode = mode,
      options = list(
        predictor_indicators = "traditional",
        compute_intercept = TRUE,
        remove_intercept = TRUE,
        allow_sparse_x = FALSE
      )
    )

    parsnip::set_fit(
      model = "pp_tree",
      eng = "ppforest2",
      mode = mode,
      value = list(
        interface = "matrix",
        protect = c("x", "y"),
        func = c(pkg = "ppforest2", fun = "pptr"),
        # See the matching comment in `register_pp_rand_forest`.
        defaults = list(mode = mode)
      )
    )
  }

  parsnip::set_model_arg(
    "pp_tree", eng = "ppforest2",
    parsnip = "penalty", original = "lambda",
    func = list(pkg = "dials", fun = "penalty"),
    has_submodel = FALSE
  )

  parsnip::set_pred(
    model = "pp_tree",
    eng = "ppforest2",
    mode = "classification",
    type = "class",
    value = list(
      pre = NULL,
      post = NULL,
      func = c(fun = "predict"),
      args = list(
        object = rlang::expr(object$fit),
        new_data = rlang::expr(new_data),
        type = "class"
      )
    )
  )

  parsnip::set_pred(
    model = "pp_tree",
    eng = "ppforest2",
    mode = "classification",
    type = "prob",
    value = list(
      pre = NULL,
      post = NULL,
      func = c(fun = "predict"),
      args = list(
        object = rlang::expr(object$fit),
        new_data = rlang::expr(new_data),
        type = "prob"
      )
    )
  )

  parsnip::set_pred(
    model = "pp_tree",
    eng = "ppforest2",
    mode = "regression",
    type = "numeric",
    value = list(
      pre = NULL,
      post = function(results, object) tibble::tibble(.pred = as.numeric(results)),
      func = c(fun = "predict"),
      args = list(
        object = rlang::expr(object$fit),
        new_data = rlang::expr(new_data),
        type = "response"
      )
    )
  )
}

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

skip_if_not_installed("parsnip")

library(parsnip)

describe("parsnip integration", {
  describe("pp_rand_forest", {
    it("can be created as a parsnip model specification", {
      spec <- pp_rand_forest() |>
        set_engine("ppforest2") |>
        set_mode("classification")
      expect_s3_class(spec, "pp_rand_forest")
    })

    it("can be fit using a parsnip workflow", {
      spec <- pp_rand_forest() |>
        set_engine("ppforest2") |>
        set_mode("classification")
      fit <- spec |> fit(Species ~ ., data = iris)
      expect_s3_class(fit, "model_fit")
      expect_s3_class(fit$fit, "pprf")
    })

    it("can predict groups via parsnip", {
      spec <- pp_rand_forest() |>
        set_engine("ppforest2") |>
        set_mode("classification")
      fit <- spec |> fit(Species ~ ., data = iris)
      preds <- predict(fit, iris)
      expect_s3_class(preds, "tbl_df")
      expect_equal(nrow(preds), nrow(iris))
      expect_true(".pred_class" %in% colnames(preds))
    })

    it("can predict probabilities via parsnip", {
      spec <- pp_rand_forest() |>
        set_engine("ppforest2") |>
        set_mode("classification")
      fit <- spec |> fit(Species ~ ., data = iris)
      preds <- predict(fit, iris, type = "prob")
      expect_s3_class(preds, "tbl_df")
      expect_equal(nrow(preds), nrow(iris))
      prob_cols <- grep("^\\.pred_", colnames(preds), value = TRUE)
      expect_equal(length(prob_cols), length(levels(iris$Species)))
    })

    it("can fit with mtry_prop (the p_vars proportion)", {
      spec <- pp_rand_forest(trees = 5, mtry_prop = 0.5) |>
        set_engine("ppforest2") |>
        set_mode("classification")
      fit <- spec |> fit(Species ~ ., data = iris)
      expect_s3_class(fit$fit, "pprf")
      # iris has 4 predictors; 0.5 * 4 = 2.
      expect_equal(fit$fit$training_spec$vars$count, 2)
    })

    it("exposes mtry_prop as a tunable parameter", {
      spec <- pp_rand_forest() |>
        set_engine("ppforest2") |>
        set_mode("classification")
      expect_true("mtry_prop" %in% tunable(spec)$name)
    })
  })

  describe("pp_tree", {
    it("can be created as a parsnip model specification", {
      spec <- pp_tree() |>
        set_engine("ppforest2") |>
        set_mode("classification")
      expect_s3_class(spec, "pp_tree")
    })

    it("can be fit using a parsnip workflow", {
      spec <- pp_tree() |>
        set_engine("ppforest2") |>
        set_mode("classification")
      fit <- spec |> fit(Species ~ ., data = iris)
      expect_s3_class(fit, "model_fit")
      expect_s3_class(fit$fit, "pptr")
    })

    it("can predict groups via parsnip", {
      spec <- pp_tree() |>
        set_engine("ppforest2") |>
        set_mode("classification")
      fit <- spec |> fit(Species ~ ., data = iris)
      preds <- predict(fit, iris)
      expect_s3_class(preds, "tbl_df")
      expect_equal(nrow(preds), nrow(iris))
      expect_true(".pred_class" %in% colnames(preds))
    })

    it("can predict probabilities via parsnip", {
      spec <- pp_tree() |>
        set_engine("ppforest2") |>
        set_mode("classification")
      fit <- spec |> fit(Species ~ ., data = iris)
      preds <- predict(fit, iris, type = "prob")
      expect_s3_class(preds, "tbl_df")
      expect_equal(nrow(preds), nrow(iris))
    })
  })

  describe("regression mode", {
    make_reg_df <- function() {
      set.seed(0)
      df <- data.frame(x1 = runif(50), x2 = runif(50))
      df$y <- df$x1 + df$x2 + rnorm(50, sd = 0.1)
      df
    }

    it("pp_tree can fit and predict regression", {
      spec <- pp_tree(mode = "regression") |> set_engine("ppforest2")
      df <- make_reg_df()
      fit <- spec |> fit(y ~ ., data = df)
      preds <- predict(fit, df)
      expect_s3_class(preds, "tbl_df")
      expect_equal(nrow(preds), nrow(df))
      expect_true(".pred" %in% colnames(preds))
      expect_true(is.numeric(preds$.pred))
    })

    it("pp_rand_forest can fit and predict regression", {
      spec <- pp_rand_forest(mode = "regression", trees = 10) |> set_engine("ppforest2")
      df <- make_reg_df()
      fit <- spec |> fit(y ~ ., data = df)
      preds <- predict(fit, df)
      expect_s3_class(preds, "tbl_df")
      expect_equal(nrow(preds), nrow(df))
      expect_true(".pred" %in% colnames(preds))
      expect_true(is.numeric(preds$.pred))
    })
  })

  describe("seed reproducibility through parsnip", {
    # parsnip's `fit()` does not wrap the engine call in a private RNG
    # context, so `pprf`'s internal `sample.int()` draws from the caller's
    # R RNG state. Lock in the three ways a user can control reproducibility:
    #   (1) outer `set.seed()`
    #   (2) `set_engine("ppforest2", seed = X)`  [engine-arg forwarding]
    #   (3) `withr::with_seed()`                 [tune's idiom]
    # If any of these regresses, reproducibility-by-seed under parsnip
    # silently breaks.
    it("outer set.seed flows through parsnip fit", {
      spec <- pp_rand_forest(trees = 5) |> set_engine("ppforest2")

      set.seed(42)
      fit1 <- spec |> fit(Species ~ ., data = iris)
      set.seed(42)
      fit2 <- spec |> fit(Species ~ ., data = iris)

      expect_identical(fit1$fit$seed, fit2$fit$seed)
      expect_equal(predict(fit1, iris), predict(fit2, iris))
    })

    it("set_engine(seed = X) is forwarded to pprf and reproducible", {
      spec <- pp_rand_forest(trees = 5) |> set_engine("ppforest2", seed = 777L)

      fit1 <- spec |> fit(Species ~ ., data = iris)
      fit2 <- spec |> fit(Species ~ ., data = iris)

      expect_identical(fit1$fit$seed, 777L)
      expect_identical(fit2$fit$seed, 777L)
      expect_equal(predict(fit1, iris), predict(fit2, iris))
    })

    it("engine-arg seed overrides outer set.seed", {
      spec <- pp_rand_forest(trees = 5) |> set_engine("ppforest2", seed = 777L)

      set.seed(1)
      fit_a <- spec |> fit(Species ~ ., data = iris)
      set.seed(999)
      fit_b <- spec |> fit(Species ~ ., data = iris)

      # Engine arg wins: both fits use seed 777 regardless of outer state.
      expect_identical(fit_a$fit$seed, 777L)
      expect_identical(fit_b$fit$seed, 777L)
      expect_equal(predict(fit_a, iris), predict(fit_b, iris))
    })

    it("withr::with_seed controls the engine RNG (tune-package idiom)", {
      skip_if_not_installed("withr")
      spec <- pp_rand_forest(trees = 5) |> set_engine("ppforest2")

      fit1 <- withr::with_seed(100, spec |> fit(Species ~ ., data = iris))
      fit2 <- withr::with_seed(100, spec |> fit(Species ~ ., data = iris))

      expect_identical(fit1$fit$seed, fit2$fit$seed)
      expect_equal(predict(fit1, iris), predict(fit2, iris))
    })
  })
})

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

skip_if_not_installed("parsnip")
skip_if_not_installed("workflows")
skip_if_not_installed("rsample")
skip_if_not_installed("tune")
skip_if_not_installed("yardstick")

library(parsnip)
library(workflows)
library(rsample)
library(tune)

# Regression suite for the "object '..y' not found" bug. The bug surfaced
# whenever a `pp_rand_forest` / `pp_tree` spec was used through workflows
# (or anything downstream of it: `fit_resamples`, `tune_grid`). Root cause
# was that `set_fit(..., interface = "formula")` triggered parsnip's
# `xy_form()` dispatch when callers passed x/y data, which synthesises a
# `..y ~ ...` formula. `pprf()` stored that formula on the model, and
# prediction blew up because new data has no `..y` column. Switching the
# registration to `interface = "matrix"` made parsnip pass x/y straight
# through. These tests lock that wiring in.

# The ..y bug surfaced as silent per-fold fit *failures*, which tune records as
# error-type notes; assert there are none of those. Degenerate-split *warnings*
# (legitimate on small resampling folds, and more frequent now that the forest
# default subsamples features) are warning-type notes and must not fail this
# plumbing test.
expect_no_fit_errors <- function(rs, label) {
  expect_equal(sum(collect_notes(rs)$type == "error"), 0L, info = label)
}

describe("parsnip + workflows integration (..y bug regression)", {
  data(crabs, package = "ppforest2")

  make_classification_split <- function(seed = 0, prop = 0.8) {
    set.seed(seed)
    initial_split(crabs, prop = prop, strata = Type)
  }

  describe("pp_rand_forest", {
    it("fits and predicts through workflow() + add_formula()", {
      split <- make_classification_split()
      train <- training(split); test <- testing(split)
      spec <- pp_rand_forest(trees = 5, mtry = 2, penalty = 0.5) |>
        set_engine("ppforest2") |> set_mode("classification")
      wf <- workflow() |> add_model(spec) |> add_formula(Type ~ .)

      fit_wf <- fit(wf, data = train)
      expect_s3_class(fit_wf, "workflow")
      preds <- predict(fit_wf, new_data = test)
      expect_s3_class(preds, "tbl_df")
      expect_equal(nrow(preds), nrow(test))
      expect_true(".pred_class" %in% colnames(preds))
      expect_s3_class(preds$.pred_class, "factor")
    })

    it("fits and predicts through workflow() + add_variables()", {
      split <- make_classification_split()
      train <- training(split); test <- testing(split)
      spec <- pp_rand_forest(trees = 5, mtry = 2, penalty = 0.5) |>
        set_engine("ppforest2") |> set_mode("classification")
      wf <- workflow() |> add_model(spec) |>
        add_variables(outcomes = Type,
                      predictors = c(FL, RW, CL, CW, BD))

      fit_wf <- fit(wf, data = train)
      preds <- predict(fit_wf, new_data = test)
      expect_equal(nrow(preds), nrow(test))
      expect_true(".pred_class" %in% colnames(preds))
    })

    it("runs end-to-end through fit_resamples()", {
      split <- make_classification_split()
      train <- training(split)
      spec <- pp_rand_forest(trees = 5, mtry = 2, penalty = 0.5) |>
        set_engine("ppforest2") |> set_mode("classification")
      wf <- workflow() |> add_model(spec) |> add_formula(Type ~ .)
      folds <- vfold_cv(train, v = 3, strata = Type)

      rs <- fit_resamples(wf, resamples = folds)
      expect_no_fit_errors(rs, "fit_resamples produced fit-error notes")
      metrics <- collect_metrics(rs)
      expect_true("accuracy" %in% metrics$.metric)
      expect_equal(metrics$n[metrics$.metric == "accuracy"], 3L)
    })

    it("runs end-to-end through tune_grid()", {
      split <- make_classification_split()
      train <- training(split)
      spec <- pp_rand_forest(trees = 5, mtry = tune(), penalty = tune()) |>
        set_engine("ppforest2") |> set_mode("classification")
      wf <- workflow() |> add_model(spec) |> add_formula(Type ~ .)
      folds <- vfold_cv(train, v = 3, strata = Type)
      grid <- tibble::tibble(mtry = c(2L, 3L), penalty = c(0.0, 0.5))

      rs <- tune_grid(wf, resamples = folds, grid = grid)
      expect_no_fit_errors(rs, "tune_grid produced fit-error notes")
      metrics <- collect_metrics(rs)
      expect_equal(length(unique(metrics$.config)), nrow(grid))
    })

    it("runs end-to-end through tune_grid() over mtry_prop", {
      split <- make_classification_split()
      train <- training(split)
      spec <- pp_rand_forest(trees = 5, mtry_prop = tune(), penalty = 0) |>
        set_engine("ppforest2") |> set_mode("classification")
      wf <- workflow() |> add_model(spec) |> add_formula(Type ~ .)
      folds <- vfold_cv(train, v = 3, strata = Type)
      grid <- tibble::tibble(mtry_prop = c(0.4, 0.8))

      rs <- tune_grid(wf, resamples = folds, grid = grid)
      expect_no_fit_errors(rs, "tune_grid over mtry_prop produced fit-error notes")
      metrics <- collect_metrics(rs)
      expect_equal(length(unique(metrics$.config)), nrow(grid))
    })
  })

  describe("pp_tree", {
    it("runs end-to-end through fit_resamples()", {
      split <- make_classification_split()
      train <- training(split)
      spec <- pp_tree(penalty = 0) |>
        set_engine("ppforest2") |> set_mode("classification")
      wf <- workflow() |> add_model(spec) |> add_formula(Type ~ .)
      folds <- vfold_cv(train, v = 3, strata = Type)

      rs <- fit_resamples(wf, resamples = folds)
      expect_no_fit_errors(rs, "fit_resamples produced fit-error notes")
      expect_true("accuracy" %in% collect_metrics(rs)$.metric)
    })

    it("runs end-to-end through tune_grid()", {
      split <- make_classification_split()
      train <- training(split)
      spec <- pp_tree(penalty = tune()) |>
        set_engine("ppforest2") |> set_mode("classification")
      wf <- workflow() |> add_model(spec) |> add_formula(Type ~ .)
      folds <- vfold_cv(train, v = 3, strata = Type)
      grid <- tibble::tibble(penalty = c(0.0, 0.5))

      rs <- tune_grid(wf, resamples = folds, grid = grid)
      expect_no_fit_errors(rs, "tune_grid produced fit-error notes")
      expect_equal(length(unique(collect_metrics(rs)$.config)), nrow(grid))
    })
  })

  describe("regression mode", {
    it("pp_rand_forest fits and predicts through workflow + fit_resamples()", {
      set.seed(0)
      df <- data.frame(x1 = runif(60), x2 = runif(60))
      df$y <- df$x1 + 0.5 * df$x2 + rnorm(60, sd = 0.1)
      spec <- pp_rand_forest(trees = 5, mode = "regression") |>
        set_engine("ppforest2")
      wf <- workflow() |> add_model(spec) |> add_formula(y ~ .)
      folds <- vfold_cv(df, v = 3)

      rs <- fit_resamples(wf, resamples = folds)
      expect_no_fit_errors(rs, "regression fit_resamples produced fit-error notes")
      metrics <- collect_metrics(rs)
      expect_true("rmse" %in% metrics$.metric)
    })
  })
})

#' Uniform random variable selection strategy.
#'
#' Creates a variable selection strategy that randomly selects a subset
#' of variables at each split. Used with \code{\link{pprf}} for random forests.
#'
#' Exactly one of \code{n_vars} or \code{p_vars} may be specified. When
#' \code{p_vars} is used, it is stored as-is and resolved to an integer count
#' later by \code{\link{pprf}} once the number of features is known.
#'
#' @param n_vars The number of variables to consider at each split (integer).
#'   Cannot be used together with \code{p_vars}.
#' @param p_vars The proportion of variables to consider at each split
#'   (number between 0 and 1, exclusive). Resolved to an integer count when
#'   the number of features is known. Cannot be used together with \code{n_vars}.
#' @return A \code{vars_strategy} object.
#' @seealso \code{\link{pprf}}, \code{\link{vars_all}}, \code{\link{pp_pda}}, \code{\link{cutpoint_mean_of_means}}
#' @examples
#' # Select 2 variables at each split
#' vars_uniform(n_vars = 2)
#'
#' # Select half the variables at each split
#' vars_uniform(p_vars = 0.5)
#'
#' # Use with pprf
#' pprf(Species ~ ., data = iris, vars = vars_uniform(n_vars = 2))
#'
#' @export
vars_uniform <- function(n_vars = NULL, p_vars = NULL) {
  if (!is.null(n_vars) && !is.null(p_vars))
    stop("Only one of `n_vars` or `p_vars` may be specified, not both.")

  if (!is.null(n_vars) && (!is.numeric(n_vars) || length(n_vars) != 1 || n_vars < 1 || n_vars != as.integer(n_vars)))
    stop("`n_vars` must be a positive integer greater than 0.")

  if (!is.null(p_vars) && (!is.numeric(p_vars) || length(p_vars) != 1 || p_vars <= 0 || p_vars > 1))
    stop("`p_vars` must be a single number between 0 (exclusive) and 1 (inclusive).")

  structure(list(name = "uniform", display_name = "Uniform random", count = n_vars, p_vars = p_vars), class = "vars_strategy")
}

#' All-variables selection strategy.
#'
#' Creates a variable selection strategy that uses all variables at each
#' split. This is the default for single trees (\code{\link{pptr}}).
#'
#' @return A \code{vars_strategy} object.
#' @seealso \code{\link{pptr}}, \code{\link{vars_uniform}}, \code{\link{pp_pda}}, \code{\link{cutpoint_mean_of_means}}
#' @examples
#' vars_all()
#'
#' @export
vars_all <- function() {
  structure(list(name = "all", display_name = "All variables"), class = "vars_strategy")
}

describe("vars_uniform", {
  it("creates a vars_strategy object with n_vars", {
    s <- vars_uniform(n_vars = 2)
    expect_s3_class(s, "vars_strategy")
    expect_equal(s$name, "uniform")
    expect_equal(s$count, 2)
    expect_null(s$p_vars)
  })

  it("creates a vars_strategy object with p_vars", {
    s <- vars_uniform(p_vars = 0.5)
    expect_s3_class(s, "vars_strategy")
    expect_equal(s$name, "uniform")
    expect_null(s$count)
    expect_equal(s$p_vars, 0.5)
  })

  it("creates a vars_strategy object with p_vars=1", {
    s <- vars_uniform(p_vars = 1)
    expect_s3_class(s, "vars_strategy")
    expect_equal(s$name, "uniform")
    expect_null(s$count)
    expect_equal(s$p_vars, 1)
  })

  it("rejects both n_vars and p_vars", {
    expect_error(vars_uniform(n_vars = 2, p_vars = 0.5), "not both")
  })

  it("rejects invalid n_vars", {
    expect_error(vars_uniform(n_vars = 0), "positive integer greater than 0")
    expect_error(vars_uniform(n_vars = 1.5), "positive integer greater than 0")
  })

  it("rejects invalid p_vars", {
    expect_error(vars_uniform(p_vars = 0), "between 0 (exclusive) and 1 (inclusive)", fixed = TRUE)
    expect_error(vars_uniform(p_vars = 2), "between 0 (exclusive) and 1 (inclusive)", fixed = TRUE)
  })
})

describe("vars_all", {
  it("creates a vars_strategy object", {
    s <- vars_all()
    expect_s3_class(s, "vars_strategy")
    expect_equal(s$name, "all")
  })
})

describe("ppforest2_proportion_to_count (shared C++/R core logic)", {
  it("rounds half to even and clamps to >= 1, matching the CLI", {
    f <- ppforest2:::ppforest2_proportion_to_count
    expect_equal(f(0.5, 4), 2L)
    expect_equal(f(0.5, 5), 2L)    # 2.5 -> 2 (ties to even)
    expect_equal(f(0.5, 9), 4L)    # 4.5 -> 4 (ties to even)
    expect_equal(f(0.1, 4), 1L)    # 0.4 -> 0 -> clamped to 1
    expect_equal(f(1.0, 10), 10L)
  })
})

# Null-coalescing helper. Package-private; used from json.R, stop-strategy.R,
# and anywhere else that wants rlang's `%||%` without depending on rlang.
`%||%` <- function(a, b) if (is.null(a)) b else a


# Resolve strategy objects vs shortcut params into strategy objects.
#
# When shortcut params are used (lambda, n_vars, p_vars), builds the
# corresponding strategy objects. When explicit strategy objects are
# provided, validates and forwards them. Errors if both APIs are mixed.
#
# When n_features is provided, p_vars is resolved to n_vars and the
# upper bound (n_vars <= n_features) is validated for uniform variable selection.
#
# @param pp A pp_strategy object or NULL.
# @param lambda Shortcut for PDA lambda (caller must pass missing() result).
# @param lambda_missing TRUE if lambda was not explicitly passed by the user.
# @param vars A vars_strategy object or NULL.
# @param n_vars Shortcut for number of variables.
# @param n_vars_missing TRUE if n_vars was not explicitly passed by the user.
# @param p_vars Shortcut for proportion of variables.
# @param p_vars_missing TRUE if p_vars was not explicitly passed by the user.
# @param cutpoint A cutpoint_strategy object or NULL.
# @param stop A stop_strategy object or NULL.
# @param binarize A binarize_strategy object or NULL.
# @param grouping A grouping_strategy object or NULL.
# @param leaf A leaf_strategy object or NULL.
# @param default_vars The default variable selection strategy when none is specified.
# @param n_features Number of features, used to resolve p_vars and validate
#   n_vars upper bound. NULL if not yet known.
# @return A list with resolved strategy objects.
resolve_strategies <- function(
  pp,
  lambda,
  lambda_missing,
  vars = NULL,
  n_vars = NULL,
  n_vars_missing = TRUE,
  p_vars = NULL,
  p_vars_missing = TRUE,
  cutpoint = NULL,
  stop = NULL,
  binarize = NULL,
  grouping = NULL,
  leaf = NULL,
  default_vars = vars_all(),
  n_features = NULL) {
  if (!is.null(pp) && !lambda_missing)
    stop("Cannot use `pp` together with `lambda`. Use one API or the other.")

  if (!is.null(vars) && (!n_vars_missing || !p_vars_missing))
    stop("Cannot use `vars` together with `n_vars`/`p_vars`. Use one API or the other.")

  if (!is.null(pp) && !inherits(pp, "pp_strategy"))
    stop("`pp` must be a pp_strategy object (e.g., pp_pda()).")

  if (!is.null(vars) && !inherits(vars, "vars_strategy"))
    stop("`vars` must be a vars_strategy object (e.g., vars_uniform() or vars_all()).")

  if (!is.null(cutpoint) && !inherits(cutpoint, "cutpoint_strategy"))
    stop("`cutpoint` must be a cutpoint_strategy object (e.g., cutpoint_mean_of_means()).")

  if (!is.null(stop) && !inherits(stop, "stop_strategy"))
    stop("`stop` must be a stop_strategy object (e.g., stop_pure_node()).")

  if (!is.null(binarize) && !inherits(binarize, "binarize_strategy"))
    stop("`binarize` must be a binarize_strategy object (e.g., binarize_largest_gap()).")

  if (!is.null(grouping) && !inherits(grouping, "grouping_strategy"))
    stop("`grouping` must be a grouping_strategy object (e.g., grouping_by_label()).")

  if (!is.null(leaf) && !inherits(leaf, "leaf_strategy"))
    stop("`leaf` must be a leaf_strategy object (e.g., leaf_majority_vote()).")

  # PP strategy
  if (is.null(pp))
    pp <- pp_pda(lambda)

  # Variable selection strategy
  if (is.null(vars)) {
    if (!is.null(n_vars) || !is.null(p_vars)) {
      vars <- vars_uniform(n_vars = n_vars, p_vars = p_vars)
    } else {
      vars <- default_vars
    }
  }

  # Resolve p_vars to count now that we know the number of features
  if (!is.null(n_features) && vars$name == "uniform") {
    if (!is.null(vars$p_vars)) {
      # Shared C++ core logic (banker's rounding + clamp to >= 1) so R and the
      # CLI resolve identical counts for the same proportion.
      vars$count <- ppforest2_proportion_to_count(vars$p_vars, n_features)
      vars$p_vars <- NULL
    }
    if (is.null(vars$count)) {
      vars$count <- n_features
    }
    if (vars$count > n_features) {
      stop("`n_vars` must be an integer between 1 and the number of features (", n_features, ").")
    }
  }

  # Strategy defaults (cutpoint, stop, binarize, grouping, leaf) are
  # resolved by `TrainingSpec::Builder::apply_defaults()` on the C++ side
  # — the single source of truth for mode-aware defaults. Anything left
  # `NULL` here is dropped from the training-spec list passed to C++,
  # and the Builder fills it in at `make()` time. R-side resolution
  # remains responsible only for parsing user-facing shortcuts (lambda,
  # n_vars, p_vars) and validating user-supplied strategy objects.

  list(
    pp = pp,
    vars = vars,
    cutpoint = cutpoint,
    stop = stop,
    binarize = binarize,
    grouping = grouping,
    leaf = leaf)
}

print_training_spec <- function(spec) {
  section_labels <- c(
    pp = "pp method", vars = "vars method", cutpoint = "cutpoint method",
    stop = "stop rule", binarize = "binarize method", grouping = "grouping method",
    leaf = "leaf method")
  for (section in c("pp", "vars", "cutpoint", "stop", "binarize", "grouping", "leaf")) {
    s <- spec[[section]]
    if (is.null(s)) next
    display <- if (!is.null(s$display_name)) s$display_name else s$name
    params <- s[!names(s) %in% c("name", "display_name")]
    if (length(params) == 0) {
      cat(section_labels[[section]], ": ", display, "\n", sep = "")
    } else {
      param_str <- paste(names(params), params, sep = "=", collapse = ", ")
      cat(section_labels[[section]], ": ", display, " (", param_str, ")\n", sep = "")
    }
  }
  cat("\n")
}

print_confusion_matrix <- function(raw_preds, model) {
  preds <- factor(model$groups[raw_preds], levels = model$groups)
  actual <- factor(model$groups[model$y], levels = model$groups)
  cm <- table(Actual = actual, Predicted = preds)
  print(cm)
  cat("\n")
  n <- length(actual)
  correct <- sum(preds == actual)
  cat("Training error: ", round((1 - correct / n) * 100, 2), "%\n", sep = "")
}

print_oob_confusion_matrix <- function(model) {
  # `oob_predictions()` is the lazy accessor; returns a factor with `NA` for
  # observations with no OOB tree.
  oob_preds <- oob_predictions(model)
  oob_mask  <- !is.na(oob_preds)
  preds     <- oob_preds[oob_mask]
  actual    <- factor(model$groups[model$y[oob_mask]], levels = model$groups)
  cm <- table(Actual = actual, Predicted = preds)
  print(cm)
  cat("\n")
  n <- length(actual)
  correct <- sum(as.character(preds) == as.character(actual))
  cat("OOB error: ", round((1 - correct / n) * 100, 2), "%\n", sep = "")
}

process_predict_arguments <- function(object, new_data, ...) {
  if (is.null(new_data)) {
    new_data <- list(...)[[1]]
  }

  # Only run model.matrix when `new_data` is a data.frame — that's what
  # model.matrix actually needs. If the caller already handed us a numeric
  # matrix (e.g. `predict(fit, fit$x)` or a hand-prepared design matrix),
  # accept it as-is. Previously this path errored with "data must be a
  # data.frame, not a matrix or an array" whenever the model was fit via
  # the formula interface, which is surprising and blocks any downstream
  # generic that wants to predict on `model$x` (notably `fitted()`).
  if (!is.null(object$formula) && is.data.frame(new_data)) {
    new_data <- model.matrix(object$formula, new_data)
  }

  as.matrix(new_data)
}

resolve_model_data <- function(formula, data, x, y, mode = NULL) {
  if (!is.null(formula) && !is.null(data)) {
    if (!inherits(formula, "formula")) {
      stop("`formula` must be a formula object.")
    }

    if (!is.data.frame(data)) {
      stop("`data` must be a data frame.")
    }

    formula <- update(formula(terms(formula, data = data)), . ~ . - 1)

    y <- model.response(model.frame(formula, data))
    x <- model.matrix(formula, data, response = TRUE)
  } else if (is.null(x) || is.null(y)) {
    stop("For the matrix interface, both `x` and `y` must be provided.")
  }

  x <- as.matrix(x)

  if (!is.numeric(x)) {
    stop("All columns in `x` must be numeric.")
  }

  if (anyNA(x)) {
    stop("`x` must not contain NA or NaN values.")
  }

  # Mode resolution. Three sources of truth, in priority order:
  #
  #   1. Explicit `mode` argument: caller stated their intent. Validate the
  #      response against it — error fast on a real mismatch (e.g. regression
  #      asked for, factor `y` supplied), coerce silently when the response
  #      can satisfy the request (e.g. classification asked for, integer
  #      labels supplied — the common binary-0/1 case).
  #   2. Response type: factor / character → classification, numeric →
  #      regression. The default when `mode` is left `NULL`.
  #
  # Returning the resolved `mode` (alongside the prepared `y` / `groups`)
  # is what the rest of the pipeline keys off: `resolve_strategies(mode = ...)`
  # for default strategies, `ppforest2_train` (which dispatches on `spec$mode`)
  # for the C++ dispatch, and the Rcpp wrap layer for the S3 class.
  if (!is.null(mode)) {
    if (!is.character(mode) || length(mode) != 1L || !mode %in% c("classification", "regression")) {
      stop("`mode` must be either \"classification\" or \"regression\" (or NULL to auto-detect).")
    }
    if (identical(mode, "regression") && is.factor(y)) {
      stop("`mode = \"regression\"` requires a numeric response; got a factor.")
    }
    is_regression <- identical(mode, "regression")
  } else {
    is_regression <- is.numeric(y) && !is.factor(y)
  }

  if (nrow(x) != length(y)) {
    stop("`x` and `y` must have the same number of observations.")
  }

  if (anyNA(y)) {
    stop("`y` must not contain NA or NaN values.")
  }
  if (is_regression && !all(is.finite(y))) {
    stop("`y` must contain only finite values for regression (no Inf / -Inf).")
  }

  if (is_regression) {
    ord <- order(y)
    x_sorted <- x[ord, , drop = FALSE]
    y_sorted <- as.numeric(y[ord])

    return(
      list(
        x = x_sorted,
        y = as.matrix(y_sorted),
        groups = character(0),
        formula = formula,
        mode = "regression"
      )
    )
  }

  # Classification path.
  if (!is.factor(y)) {
    y <- factor(y)
  }

  if (nlevels(y) < 2) {
    stop("`y` must have at least 2 groups.")
  }

  if (nrow(x) < nlevels(y)) {
    stop("Not enough observations: need at least as many rows as groups.")
  }

  return(
    list(
      x = x,
      y = as.matrix(as.numeric(y)),
      groups = levels(y),
      formula = formula,
      mode = "classification"
    )
  )
}

# ======================================================================
# pptr Visualization — Design system and shared helpers
#
# Named constants for line weights, point sizes, alpha values, and
# structural colors used across all plot types.  These ensure a cohesive
# look and make it easy to adjust the overall visual style in one place.
#
# The design system has two layers:
#
#   1. Structural constants (below): fixed values for non-data elements
#      like edges, borders, cutpoints, and decision boundaries.  These
#      use neutral greys to avoid competing with data-mapped colors.
#
#   2. Data-mapped colors: per-group fill and colour generated by
#      get_group_colors() using evenly-spaced HCL hues.  Applied via
#      scale_fill_manual / scale_color_manual so users can override
#      with + scale_fill_manual(values = ...).
#
# Architecture:
#   - C++ (Visualization.hpp/cpp) computes geometry: node data routing,
#     boundary line clipping, region polygon clipping (Sutherland-Hodgman),
#     and tree layout positioning.
#   - R (plot-*.R files) handles rendering: translates C++ output into
#     ggplot2 layers and assembles composite layouts via patchwork.
# ======================================================================

# Suppress R CMD check NOTEs for ggplot2 aes() column references
utils::globalVariables(c(
  "from_x", "from_y", "to_x", "to_y",
  "importance", "variable",
  "projected",
  "x", "y",
  "x_start", "y_start", "x_end", "y_end",
  "alpha",
  "edge_label", "group", "xmin", "xmax", "ymin", "ymax", "label",
  "xend", "yend", "region_id", "region_group",
  "metric", "x_var", "y_var"
))

#' Guard: abort with a helpful message if ggplot2 is not installed.
#'
#' Called at the top of every public plot method (plot.pptr, plot.pprf).
#' @noRd
check_ggplot2 <- function() {
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop(
      "Package 'ggplot2' is required for plot methods. ",
      "Install it with install.packages('ggplot2').",
      call. = FALSE
    )
  }
}

#' Guard: abort with a helpful message if patchwork is not installed.
#'
#' Called by composite layouts (mosaic, importance grid) that use
#' patchwork to compose multiple ggplot objects into a single plot.
#' @noRd
check_patchwork <- function() {
  if (!requireNamespace("patchwork", quietly = TRUE)) {
    stop(
      "Package 'patchwork' is required for composite plot layouts. ",
      "Install it with install.packages('patchwork').",
      call. = FALSE
    )
  }
}

# ------------------------------------------------------------------
# Design option accessor
#
# Each constant is read from options() with a default fallback, so
# users can customize any value globally:
#
#   options(ppforest2.col_bar = "steelblue")
#   options(ppforest2.alpha_region = 0.4)
#
# Or restore defaults by setting to NULL:
#
#   options(ppforest2.col_bar = NULL)
# ------------------------------------------------------------------
ppforest2_opt <- function(name, default) {
  getOption(paste0("ppforest2.", name), default)
}

# Node dimensions (must match C++ LayoutParams)
ppforest2_node_w <- function() ppforest2_opt("node_w", 0.8)
ppforest2_node_h <- function() ppforest2_opt("node_h", 0.7)
ppforest2_leaf_w <- function() ppforest2_opt("leaf_w", 0.5)
ppforest2_leaf_h <- function() ppforest2_opt("leaf_h", 0.3)

# Line weights
ppforest2_lw_light  <- function() ppforest2_opt("lw_light",  0.4)
ppforest2_lw_medium <- function() ppforest2_opt("lw_medium", 0.6)

# Point sizes
ppforest2_pt_small  <- function() ppforest2_opt("pt_small",  1.2)
ppforest2_pt_medium <- function() ppforest2_opt("pt_medium", 1.8)

# Alpha values
ppforest2_alpha_region <- function() ppforest2_opt("alpha_region", 0.25)
ppforest2_alpha_hist   <- function() ppforest2_opt("alpha_hist",   0.65)
ppforest2_alpha_leaf   <- function() ppforest2_opt("alpha_leaf",   0.30)
ppforest2_alpha_proj   <- function() ppforest2_opt("alpha_proj",   0.60)

# Structural colors (not data-mapped)
ppforest2_col_edge     <- function() ppforest2_opt("col_edge",     "grey55")
ppforest2_col_border   <- function() ppforest2_opt("col_border",   "grey50")
ppforest2_col_cutpoint <- function() ppforest2_opt("col_cutpoint", "grey25")
ppforest2_col_boundary <- function() ppforest2_opt("col_boundary", "grey30")
ppforest2_col_tick     <- function() ppforest2_opt("col_tick",     "grey40")
ppforest2_col_bar      <- function() ppforest2_opt("col_bar",      "#5B8BA0")

#' Shared ggplot2 theme for all non-structure plots.
#'
#' Uses \code{theme_minimal} as a base with slightly larger titles and no
#' minor gridlines.  The tree structure plot uses \code{theme_void()}
#' instead (no axes, no grid) since it renders on an abstract 2D canvas.
#' @noRd
ppforest2_theme <- function() {
  base_size <- ppforest2_opt("base_size", 11)
  ggplot2::theme_minimal(base_size = base_size) +
  ggplot2::theme(
    plot.title       = ggplot2::element_text(size = ggplot2::rel(1.1), hjust = 0),
    plot.title.position = "plot",
    axis.title       = ggplot2::element_text(size = ggplot2::rel(0.85)),
    panel.grid.minor = ggplot2::element_blank()
  )
}

# ======================================================================
# Shared helpers
# ======================================================================

#' Extract variable names from a model.
#'
#' Prefers column names from the training matrix (model$x).  Falls back
#' to "x1", "x2", ... if column names are absent.
#'
#' @param model A pptr or pprf model with \code{$x} and optionally
#'   \code{$vi}.
#' @return Character vector of length p (number of features).
#' @noRd
get_variable_names <- function(model) {
  p <- if (!is.null(model$vi$projections)) {
    length(model$vi$projections)
  } else {
    ncol(model$x)
  }

  if (!is.null(colnames(model$x))) {
    colnames(model$x)
  } else {
    paste0("x", seq_len(p))
  }
}

#' Generate a perceptually-uniform color palette for group labels.
#'
#' Colors are evenly spaced in HCL hue space (lightness = 65, chroma = 100)
#' to ensure maximal visual distinction between groups.  Named by the
#' group label so they can be passed directly to \code{scale_fill_manual()}.
#'
#' @param classes Character vector of unique group labels.
#' @return Named character vector of hex colours, one per group.
#' @noRd
get_group_colors <- function(groups) {
  n <- length(groups)
  hues <- seq(15, 375, length.out = n + 1)[seq_len(n)]
  colors <- grDevices::hcl(h = hues, l = 65, c = 100)
  names(colors) <- groups
  colors
}

#' Equalize two ranges so both span the same distance (for square plots).
#'
#' Both ranges are expanded symmetrically about their midpoints to match
#' the larger span.  Used before \code{coord_fixed(ratio = 1)} to ensure
#' the 2D boundary plot is square without distortion.
#'
#' @param x_range Numeric vector of length 2 (x_min, x_max).
#' @param y_range Numeric vector of length 2 (y_min, y_max).
#' @return List with elements \code{$x} and \code{$y}, each a length-2
#'   numeric vector.
#' @noRd
equalize_ranges <- function(x_range, y_range) {
  x_span <- diff(x_range)
  y_span <- diff(y_range)
  max_span <- max(x_span, y_span)
  x_mid <- mean(x_range)
  y_mid <- mean(y_range)
  list(
    x = c(x_mid - max_span / 2, x_mid + max_span / 2),
    y = c(y_mid - max_span / 2, y_mid + max_span / 2)
  )
}

#' Pad a numeric range symmetrically by a fraction of its span.
#'
#' @param rng Numeric vector of length 2 (min, max).
#' @param frac Fraction of the span to add on each side (default 0.05).
#' @return Numeric vector of length 2 (padded min, padded max).
#' @noRd
pad_range <- function(rng, frac = 0.05) {
  pad <- diff(rng) * frac
  rng + c(-pad, pad)
}

# This file is part of the standard setup for testthat.
# It is recommended that you do not modify it.
#
# Where should you do additional test configuration?
# Learn more about the roles of various files in:
# * https://r-pkgs.org/testing-design.html#sec-tests-files-overview
# * https://testthat.r-lib.org/articles/special-files.html

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

test_check("ppforest2")

#' Save a model to a JSON file.
#'
#' Serializes a \code{pptr} or \code{pprf} model to JSON format compatible
#' with the C++ CLI. The JSON includes the model structure, group labels,
#' training parameters, and optionally variable importance metrics.
#'
#' @param model A \code{pptr} or \code{pprf} model (inheriting from \code{ppmodel}).
#' @param path File path to write the JSON to.
#' @param include_metrics If \code{TRUE} (default), include variable importance
#'   and (for forests) OOB error in the output. Set to \code{FALSE} to save
#'   only the model structure and metadata.
#' @param ... Additional arguments (currently unused).
#' @return Invisibly returns the file \code{path} that was written (a length-one
#'   character string). Called for its side effect of serializing \code{model}
#'   to a JSON file at \code{path}.
#' @seealso \code{\link{load_json}}, \code{\link{pptr}}, \code{\link{pprf}}
#' @examples
#' model <- pptr(Species ~ ., data = iris, seed = 0)
#' path <- tempfile(fileext = ".json")
#' save_json(model, path)
#'
#' @export
save_json <- function(model, path, ...) {
  UseMethod("save_json")
}

#' @rdname save_json
#' @export
save_json.ppmodel <- function(model, path, include_metrics = TRUE, ...) {
  if (include_metrics && is.null(model$x)) {
    warning("Training data not available; saving without metrics.", call. = FALSE)
    include_metrics <- FALSE
  }
  x <- if (!is.null(model$x)) model$x else matrix(0, nrow = 0, ncol = 0)
  feature_names <- if (!is.null(model$x)) colnames(model$x) else character(0)
  groups <- if (!is.null(model$groups)) model$groups else character(0)

  # `y` is the unified response vector: integer class labels (classification)
  # or the continuous response (regression). Single field, mode-agnostic —
  # matches the C++ unified-y convention.
  y <- if (!is.null(model$y)) model$y else {
    if (identical(model$mode, "regression")) numeric(0) else integer(0)
  }

  json_str <- ppforest2_save_model_json(
    model, groups, include_metrics, x, y, feature_names
  )
  writeLines(json_str, path)
  invisible(path)
}


#' Load a model from a JSON file.
#'
#' Deserializes a \code{pptr} or \code{pprf} model from a JSON file. The model
#' can be used for prediction immediately. If variable importance metrics and
#' OOB fields were saved, they are restored into the model's lazy-computation
#' cache so later accessor calls (\code{oob_error()}, \code{permuted_importance()},
#' etc.) return the stored values without recomputation.
#'
#' Note that \code{formula}, \code{x}, and \code{y} are not stored in the JSON
#' and will be \code{NULL} on the loaded model. Formula-based prediction and
#' any accessor that requires training data will error unless the user
#' re-attaches those fields.
#'
#' @param path File path to read the JSON from.
#' @return A \code{pptr} or \code{pprf} model (with the appropriate
#'   \code{_classification} / \code{_regression} subclass).
#' @seealso \code{\link{save_json}}, \code{\link{pptr}}, \code{\link{pprf}}
#' @examples
#' model <- pptr(Species ~ ., data = iris, seed = 0)
#' path <- tempfile(fileext = ".json")
#' save_json(model, path)
#' loaded <- load_json(path)
#' predict(loaded, as.matrix(iris[, 1:4]))
#'
#' @export
load_json <- function(path) {
  # The underlying C++ loader throws `std::runtime_error` (surfaced by Rcpp
  # as a plain R error without the file path) for any structural problem —
  # malformed JSON, mode/groups mismatch, missing required keys, etc. Wrap
  # here so the user sees which file triggered the failure.
  model <- tryCatch(
    ppforest2_load_model_json(path),
    error = function(e) {
      stop(
        "Failed to load model from '", path, "': ", conditionMessage(e),
        call. = FALSE
      )
    }
  )
  model$groups <- as.character(model$groups)

  # Restore mode from the training spec.
  model$mode <- model$training_spec$mode %||% "classification"

  is_forest     <- !is.null(model$trees)
  is_regression <- identical(model$mode, "regression")

  if (is_forest && !is_regression) {
    for (i in seq_along(model$trees)) {
      model$trees[[i]]$groups <- model$groups
    }
  }

  # Cache env for lazy accessors. Prime with values that travelled through JSON
  # (the C++ save/load path persists OOB metrics and weighted/permuted VI).
  # `oob_metrics` mirrors the C++ Export shape: a list with `confusion_matrix`
  # + `error_rate` (classification) or `mse`/`mae`/`r_squared` (regression).
  # `oob_error()` extracts the right scalar from it without recomputation.
  model$.cache <- .new_cache()
  oob_scalar <- if (!is.null(model$oob_metrics)) {
    if (is_regression) model$oob_metrics$mse else model$oob_metrics$error_rate
  } else NULL
  .prime_cache(model, "oob_error",            oob_scalar)
  .prime_cache(model, "weighted_importance",  model$vi$weighted)
  .prime_cache(model, "permuted_importance",  model$vi$permuted)
  # Clean up primed VI fields so users access via generics, not raw slots.
  # `oob_metrics` itself stays on the model — it's the richer companion to
  # `oob_error()` and `model$training_metrics`.
  model$vi$weighted <- NULL
  model$vi$permuted <- NULL

  # Class is set by the Rcpp wrap layer (see `make_model_class` in
  # bindings/R/inst/include/ppforest2.h), which derives it from
  # `training_spec->mode`. The deserialized model already carries the
  # correct 3-level S3 class — no reassignment needed.

  model
}

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

describe("pptr formula interface", {
  it("returns a object with s3 class pptr", {
    model <- pptr(Species ~ ., data = iris)
    expect_s3_class(model, "pptr")
  })

  it("preserves the observations (x) as a matrix in the returned model", {
    model <- pptr(Species ~ ., data = iris)
    expected <- model.matrix(
      Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width - 1,
      iris,
      response = TRUE
    )
    expect_equal(model$x, expected)
  })

  it("preserves the labels (y) as a matrix in the returned model", {
    model <- pptr(Species ~ ., data = iris)
    expected <- as.matrix(as.numeric(as.factor(iris$Species)))
    expect_equal(model$y, expected)
  })

  it("preserves the groups in the returned model", {
    model <- pptr(Species ~ ., data = iris)
    expected <- levels(iris$Species)
    expect_equal(model$groups, expected)
  })

  it("preserves the formula in the returned model", {
    model <- pptr(Species ~ ., data = iris)
    expect_equal(
      model$formula,
      Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width - 1
    )
  })
})

describe("pptr matrix interface", {
  it("returns a object with s3 class pptr", {
    model <- pptr(x = iris[, 1:4], y = iris[, 5])
    expect_s3_class(model, "pptr")
  })

  it("preserves the observations (x) as a matrix in the returned model", {
    model <- pptr(x = iris[, 1:4], y = iris[, 5])
    expected <- as.matrix(iris[, 1:4])
    expect_equal(model$x, expected)
  })

  it("preserves the labels (y) as a matrix in the returned model", {
    model <- pptr(x = iris[, 1:4], y = iris[, 5])
    expected <- as.matrix(as.numeric(as.factor(iris$Species)))
    expect_equal(model$y, expected)
  })

  it("preserves the groups in the returned model", {
    model <- pptr(x = iris[, 1:4], y = iris[, 5])
    expected <- levels(iris$Species)
    expect_equal(model$groups, expected)
  })

  it("does not have a formula", {
    model <- pptr(x = iris[, 1:4], y = iris[, 5])
    expect_equal(model$formula, NULL)
  })
})

describe("pptr reproducibility", {
  it("produces identical predictions with set.seed", {
    set.seed(0)
    model1 <- pptr(Species ~ ., data = iris)
    set.seed(0)
    model2 <- pptr(Species ~ ., data = iris)
    expect_equal(predict(model1, iris), predict(model2, iris))
  })

  it("produces identical predictions with explicit seed", {
    model1 <- pptr(Species ~ ., data = iris, seed = 123L)
    model2 <- pptr(Species ~ ., data = iris, seed = 123L)
    expect_equal(predict(model1, iris), predict(model2, iris))
  })

  it("stores the explicit seed in the model", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    expect_equal(model$seed, 0)
  })

  it("stores the generated seed when seed is NULL", {
    set.seed(99)
    model <- pptr(Species ~ ., data = iris)
    expect_true(is.numeric(model$seed))
    expect_true(model$seed > 0)
  })
})

describe("pptr input validation", {
  it("rejects invalid lambda", {
    expect_error(pptr(x = iris[, 1:4], y = iris[, 5], lambda = -1), "between 0 and 1")
    expect_error(pptr(x = iris[, 1:4], y = iris[, 5], lambda = 2), "between 0 and 1")
    expect_error(pptr(x = iris[, 1:4], y = iris[, 5], lambda = "a"), "between 0 and 1")
  })

  it("rejects non-integer seed", {
    expect_error(pptr(x = iris[, 1:4], y = iris[, 5], seed = 1.5), "integer")
  })

  it("rejects single-group y", {
    expect_error(pptr(x = iris[, 1:4], y = rep("a", 150)), "at least 2 groups")
  })

  it("rejects dimension mismatch", {
    expect_error(pptr(x = iris[1:10, 1:4], y = iris[, 5]), "same number of observations")
  })

  # Non-finite y rejection — used to cross the C++ boundary and trip an
  # out-of-range cast. Rejecting here fails fast with a clear message.
  it("rejects NA in factor y", {
    x <- as.matrix(iris[, 1:4])
    y <- iris$Species
    y[10] <- NA
    expect_error(pptr(x = x, y = y), "`y` must not contain NA")
  })
})

describe("pptr training spec", {
  it("preserves the lambda parameter in the returned model", {
    model <- pptr(Species ~ ., data = iris, lambda = 0.5)
    expect_equal(model$training_spec$pp$lambda, 0.5)
  })

  it("the lambda parameter is 0.5 by default", {
    model <- pptr(Species ~ ., data = iris)
    expect_equal(model$training_spec$pp$lambda, 0.5)
  })

  it("the pp strategy is pda", {
    model <- pptr(Species ~ ., data = iris)
    expect_equal(model$training_spec$pp$name, "pda")
  })

  it("the vars strategy is all", {
    model <- pptr(Species ~ ., data = iris)
    expect_equal(model$training_spec$vars$name, "all")
  })

  it("the stop strategy is pure_node", {
    model <- pptr(Species ~ ., data = iris)
    expect_equal(model$training_spec$stop$name, "pure_node")
  })

  it("the binarize strategy is largest_gap", {
    model <- pptr(Species ~ ., data = iris)
    expect_equal(model$training_spec$binarize$name, "largest_gap")
  })

  it("the grouping strategy is by_label", {
    model <- pptr(Species ~ ., data = iris)
    expect_equal(model$training_spec$grouping$name, "by_label")
  })

  it("the leaf strategy is majority_vote", {
    model <- pptr(Species ~ ., data = iris)
    expect_equal(model$training_spec$leaf$name, "majority_vote")
  })
})

describe("pptr regression training spec", {
  # Pins the mode-aware defaults selected by
  # `TrainingSpec::Builder::apply_defaults()` for regression. The C++
  # side is the single source of truth; this block guards against
  # silent drift if the default selection ever changes.
  it("the pp strategy is pda", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)
    expect_equal(model$training_spec$pp$name, "pda")
  })

  it("the vars strategy is all", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)
    expect_equal(model$training_spec$vars$name, "all")
  })

  it("the cutpoint strategy is mean_of_means", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)
    expect_equal(model$training_spec$cutpoint$name, "mean_of_means")
  })

  it("the stop strategy is any(min_size(5), min_variance(0.01))", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)
    stop_spec <- model$training_spec$stop
    expect_equal(stop_spec$name, "any")
    expect_length(stop_spec$rules, 2)
    expect_equal(stop_spec$rules[[1]]$name, "min_size")
    expect_equal(stop_spec$rules[[1]]$min_size, 5)
    expect_equal(stop_spec$rules[[2]]$name, "min_variance")
    expect_equal(stop_spec$rules[[2]]$threshold, 0.01, tolerance = 1e-6)
  })

  it("the binarize strategy is disabled", {
    # Regression's `by_cutpoint` grouping always yields a 2-group
    # partition, so binarize is a no-op placeholder.
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)
    expect_equal(model$training_spec$binarize$name, "disabled")
  })

  it("the grouping strategy is by_cutpoint", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)
    expect_equal(model$training_spec$grouping$name, "by_cutpoint")
  })

  it("the leaf strategy is mean_response", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)
    expect_equal(model$training_spec$leaf$name, "mean_response")
  })
})

describe("pptr with strategy objects", {
  it("trains with pp = pp_pda()", {
    model <- pptr(Species ~ ., data = iris, pp = pp_pda(0.5), seed = 0)
    expect_s3_class(model, "pptr")
    expect_equal(model$training_spec$pp$lambda, 0.5)
  })

  it("trains with pp = pp_pda(0) (LDA)", {
    model <- pptr(Species ~ ., data = iris, pp = pp_pda(0), seed = 0)
    expect_s3_class(model, "pptr")
    expect_equal(model$training_spec$pp$lambda, 0)
  })

  it("produces identical export as lambda shortcut", {
    model_shortcut <- pptr(Species ~ ., data = iris, lambda = 0.5, seed = 0)
    model_strategy <- pptr(Species ~ ., data = iris, pp = pp_pda(0.5), seed = 0)

    path_shortcut <- tempfile(fileext = ".json")
    path_strategy <- tempfile(fileext = ".json")
    save_json(model_shortcut, path_shortcut)
    save_json(model_strategy, path_strategy)

    j_shortcut <- jsonlite::fromJSON(readLines(path_shortcut, warn = FALSE))
    j_strategy <- jsonlite::fromJSON(readLines(path_strategy, warn = FALSE))

    # Compare full export (model, config, meta, metrics)
    j_shortcut$training_duration_ms <- NULL
    j_strategy$training_duration_ms <- NULL
    expect_equal(j_shortcut, j_strategy)
  })

  it("trains with explicit stop, binarize, grouping, leaf", {
    model <- pptr(Species ~ ., data = iris, stop = stop_pure_node(), binarize = binarize_largest_gap(), grouping = grouping_by_label(), leaf = leaf_majority_vote(), seed = 0)
    expect_s3_class(model, "pptr")
    expect_equal(model$training_spec$stop$name, "pure_node")
    expect_equal(model$training_spec$binarize$name, "largest_gap")
    expect_equal(model$training_spec$grouping$name, "by_label")
    expect_equal(model$training_spec$leaf$name, "majority_vote")
  })

  it("rejects non-stop_strategy objects", {
    expect_error(
      pptr(Species ~ ., data = iris, stop = list(name = "pure_node")),
      "stop_strategy")
  })

  it("rejects non-binarize_strategy objects", {
    expect_error(
      pptr(Species ~ ., data = iris, binarize = list(name = "largest_gap")),
      "binarize_strategy")
  })

  it("rejects non-grouping_strategy objects", {
    expect_error(
      pptr(Species ~ ., data = iris, grouping = list(name = "by_label")),
      "grouping_strategy")
  })

  it("rejects non-leaf_strategy objects", {
    expect_error(
      pptr(Species ~ ., data = iris, leaf = list(name = "majority_vote")),
      "leaf_strategy")
  })

  it("default tree and fully explicit defaults produce identical export", {
    model_default  <- pptr(Species ~ ., data = iris, seed = 0)
    model_explicit <- pptr(Species ~ ., data = iris, pp = pp_pda(0.5), cutpoint = cutpoint_mean_of_means(), stop = stop_pure_node(), binarize = binarize_largest_gap(), grouping = grouping_by_label(), leaf = leaf_majority_vote(), seed = 0)

    path_default  <- tempfile(fileext = ".json")
    path_explicit <- tempfile(fileext = ".json")
    save_json(model_default, path_default)
    save_json(model_explicit, path_explicit)

    j_default  <- jsonlite::fromJSON(readLines(path_default, warn = FALSE))
    j_explicit <- jsonlite::fromJSON(readLines(path_explicit, warn = FALSE))

    j_default$training_duration_ms  <- NULL
    j_explicit$training_duration_ms <- NULL
    expect_equal(j_default, j_explicit)
  })

  it("errors when mixing pp and lambda", {
    expect_error(
      pptr(Species ~ ., data = iris, pp = pp_pda(0.5), lambda = 0.3),
      "Cannot use `pp` together with `lambda`")
  })

  it("rejects non-pp_strategy objects", {
    expect_error(
      pptr(Species ~ ., data = iris, pp = list(name = "pda", lambda = 0.5)),
      "pp_strategy")
  })

  it("rejects non-cutpoint_strategy objects", {
    expect_error(
      pptr(Species ~ ., data = iris, cutpoint = list(name = "mean_of_means")),
      "cutpoint_strategy")
  })
})

describe("pptr edge cases", {
  it("rejects NA in features via matrix interface", {
    x <- matrix(c(1, NA, 3, 4, 1, 2, 3, 4), ncol = 2)
    y <- c("a", "a", "b", "b")
    expect_error(pptr(x = x, y = y), "NA or NaN")
  })

  it("rejects NaN in features via matrix interface", {
    x <- matrix(c(1, NaN, 3, 4, 1, 2, 3, 4), ncol = 2)
    y <- c("a", "a", "b", "b")
    expect_error(pptr(x = x, y = y), "NA or NaN")
  })

  it("trains with a constant feature column", {
    df <- data.frame(x1 = rep(5, 6), x2 = c(1, 2, 3, 7, 8, 9), y = c("a", "a", "a", "b", "b", "b"))
    expect_no_error(pptr(y ~ ., data = df, seed = 0))
  })

  it("trains with single observation per group", {
    df <- data.frame(x1 = c(1, 0), x2 = c(0, 1), y = c("a", "b"))
    model <- pptr(y ~ ., data = df, seed = 0)
    expect_s3_class(model, "pptr")
    preds <- predict(model, df)
    expect_length(preds, 2)
  })

  it("trains with minimal dataset (n=2)", {
    df <- data.frame(x1 = c(1, 9), y = c("a", "b"))
    model <- pptr(y ~ ., data = df, seed = 0)
    expect_s3_class(model, "pptr")
    preds <- predict(model, df)
    expect_length(preds, 2)
  })

  it("trains with extreme class imbalance", {
    df <- data.frame(
      x1 = c(0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 0, 1, 2, 90, 91),
      x2 = c(0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 90, 91),
      y = c(rep("a", 18), rep("b", 2))
    )
    model <- pptr(y ~ ., data = df, seed = 0)
    expect_s3_class(model, "pptr")
    preds <- predict(model, df)
    expect_length(preds, 20)
  })
})

describe("pptr regression", {
  it("end-to-end on mtcars (formula + predict + summary)", {
    # Real-dataset round-trip counterpart to the simulated-data tests
    # below, mirroring the analog in `test-pprf.R`. mtcars is small
    # (32 x 11) and ships with base R so this doesn't grow the package
    # footprint, while still exercising the full single-tree pipeline
    # on real covariances: formula interface, regression auto-detection
    # from numeric `y`, training, prediction, and summary rendering.
    data(mtcars)
    model <- pptr(mpg ~ ., data = mtcars, seed = 0)

    expect_equal(model$mode, "regression")
    # `groups` is an empty character vector for regression (no class levels).
    expect_length(model$groups, 0)
    expect_s3_class(model, "pptr_regression")
    expect_s3_class(model, "pptr")
    expect_s3_class(model, "ppmodel")

    preds <- predict(model, newdata = mtcars)
    expect_true(is.numeric(preds))
    expect_length(preds, nrow(mtcars))
    expect_true(all(is.finite(preds)))

    # summary() must render the regression-shaped header + training
    # metrics block (MSE / MAE / R²) and must not render a confusion
    # matrix. The "Regression Tree" header distinguishes from the
    # classification "Decision Tree" header.
    out <- capture.output(summary(model))
    expect_true(any(grepl("Regression Tree", out)))
    expect_true(any(grepl("Training Metrics", out)))
    expect_true(any(grepl("MSE|R\\^2|R²|MAE", out)))
    expect_false(any(grepl("Confusion Matrix", out)))
  })

  it("auto-detects regression from numeric y", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)

    expect_equal(model$mode, "regression")
    expect_length(model$groups, 0)
    # Unified `y`: continuous response stored on `model$y` for both modes.
    expect_equal(length(model$y), 30)
  })

  it("still dispatches to classification when y is a factor", {
    model <- pptr(x = iris[, 1:4], y = iris$Species, seed = 0)

    expect_equal(model$mode, "classification")
    expect_equal(length(model$groups), 3)
  })

  it("predict returns a numeric vector", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)
    preds <- predict(model, d$x)

    expect_type(preds, "double")
    expect_length(preds, 30)
    expect_false(is.factor(preds))
  })

  it("predict rejects type='class' and type='prob'", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)

    expect_error(predict(model, d$x, type = "class"), "regression")
    expect_error(predict(model, d$x, type = "prob"), "regression")
  })

  it("supports the formula interface", {
    d <- make_regression_data()
    df <- data.frame(x1 = d$x[, 1], x2 = d$x[, 2], y = d$y)
    model <- pptr(y ~ ., data = df, seed = 0)

    expect_equal(model$mode, "regression")
    preds <- predict(model, df)
    expect_type(preds, "double")
  })

  it("save/load round-trip preserves predictions", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, seed = 0)

    path <- tempfile(fileext = ".json")
    save_json(model, path)
    loaded <- load_json(path)

    expect_equal(loaded$mode, "regression")

    original_preds <- predict(model, d$x)
    loaded_preds <- predict(loaded, d$x)

    expect_equal(as.numeric(loaded_preds), as.numeric(original_preds))
  })

  it("works with custom strategies", {
    d <- make_regression_data(n = 50)
    model <- pptr(
      x = d$x, y = d$y, seed = 0,
      stop = stop_min_size(10),
      leaf = leaf_mean_response(),
      grouping = grouping_by_cutpoint()
    )

    expect_equal(model$mode, "regression")
    preds <- predict(model, d$x)
    expect_length(preds, 50)
  })

  it("is reproducible with the same seed", {
    d <- make_regression_data()

    m1 <- pptr(x = d$x, y = d$y, seed = 42)
    m2 <- pptr(x = d$x, y = d$y, seed = 42)

    expect_equal(as.numeric(predict(m1, d$x)), as.numeric(predict(m2, d$x)))
  })

  # Non-finite y rejection — these values used to cross the C++ boundary
  # and either trip an out-of-range cast or violate strict-weak-order in
  # std::stable_sort. Rejecting in `validate_data` fails fast.
  it("rejects NA in y", {
    d <- make_regression_data()
    d$y[3] <- NA_real_
    expect_error(pptr(x = d$x, y = d$y), "`y` must not contain NA")
  })

  it("rejects NaN in y", {
    d <- make_regression_data()
    d$y[5] <- NaN
    expect_error(pptr(x = d$x, y = d$y), "`y` must not contain NA")
  })

  it("rejects Inf in y", {
    d <- make_regression_data()
    d$y[7] <- Inf
    expect_error(pptr(x = d$x, y = d$y), "finite values for regression")
  })

  it("rejects NA in features via matrix interface", {
    x <- matrix(c(1, NA, 3, 4, 1, 2, 3, 4), ncol = 2)
    y <- c(1.1, 1.2, 1.3, 1.4)
    expect_error(pptr(x = x, y = y), "NA or NaN")
  })

  it("rejects NaN in features via matrix interface", {
    x <- matrix(c(1, NaN, 3, 4, 1, 2, 3, 4), ncol = 2)
    y <- c(1.1, 1.2, 1.3, 1.4)
    expect_error(pptr(x = x, y = y), "NA or NaN")
  })

  it("trains with a constant feature column", {
    # First column is constant; the second carries the signal. The
    # variable selector / cutpoint must handle the degenerate column
    # without aborting training.
    x <- cbind(rep(5, 6), c(1, 2, 3, 7, 8, 9))
    y <- c(0.1, 0.2, 0.3, 0.7, 0.8, 0.9)
    expect_no_error(pptr(x = x, y = y, seed = 0))
  })

  it("trains with minimal dataset (n = 2)", {
    # ByCutpoint::compute_init handles n = 2 by splitting [0, 0] / [1, 1].
    # The R side accepts any n >= 1 (matches the C++ minimum).
    x <- matrix(c(1, 9, 0, 0), ncol = 2)
    y <- c(0.0, 1.0)
    expect_no_error(pptr(x = x, y = y, seed = 0))
  })
})

describe("pptr explicit mode parameter", {
  # The `mode` arg lets callers override the auto-detection from `y`'s
  # type (factor/character -> classification, numeric -> regression).

  it("mode = \"classification\" coerces integer 0/1 labels", {
    # Common binary-labels case: a numeric vector that would otherwise
    # auto-detect as regression. Explicit mode pins classification, and
    # the labels are coerced via `factor()`.
    x <- matrix(rnorm(40), ncol = 2)
    y <- rep(c(0L, 1L), 10)
    model <- pptr(x = x, y = y, mode = "classification", seed = 0)

    expect_equal(model$mode, "classification")
    expect_equal(length(model$groups), 2L)
    expect_s3_class(model, "pptr_classification")
  })

  it("mode = \"regression\" errors on a factor y", {
    # Type-mismatch fail-fast: factor `y` cannot be coerced to a
    # meaningful continuous response.
    expect_error(
      pptr(x = iris[, 1:4], y = iris$Species, mode = "regression", seed = 0),
      "`mode = \"regression\"` requires a numeric response"
    )
  })

  it("mode = NULL preserves auto-detection (factor)", {
    model <- pptr(x = iris[, 1:4], y = iris$Species, mode = NULL, seed = 0)
    expect_equal(model$mode, "classification")
  })

  it("mode = NULL preserves auto-detection (numeric)", {
    d <- make_regression_data()
    model <- pptr(x = d$x, y = d$y, mode = NULL, seed = 0)
    expect_equal(model$mode, "regression")
  })

  it("rejects invalid mode strings", {
    expect_error(
      pptr(x = iris[, 1:4], y = iris$Species, mode = "ranking", seed = 0),
      "`mode` must be either"
    )
  })
})

describe("pptr call capture and update()", {
  # Mirror of the same suite on `pprf` — see comments there.
  it("stores the user's call on the fitted model", {
    model <- pptr(Species ~ ., data = iris, lambda = 0.0)
    expect_false(is.null(model$call))
    expect_true(is.call(model$call))
    expect_equal(as.character(model$call)[[1L]], "pptr")
  })

  it("update() re-fits with the substituted argument", {
    fit1 <- pptr(Species ~ ., data = iris, lambda = 0.0, seed = 0)
    expect_equal(fit1$training_spec$pp$lambda, 0.0)

    fit2 <- update(fit1, lambda = 0.5)
    expect_s3_class(fit2, "pptr")
    expect_equal(fit2$training_spec$pp$lambda, 0.5)
  })

  it("print() surfaces the call when present", {
    model <- pptr(Species ~ ., data = iris, lambda = 0.0)
    out <- capture.output(print(model))
    expect_true(any(grepl("^Call:", out)))
    expect_true(any(grepl("pptr\\(", out)))
  })
})

# ======================================================================
# Mosaic layout
# ======================================================================

#' Render a composite mosaic of tree structure, importance, and boundaries.
#'
#' Arranges three plots in a 2-column grid (AB / AC design):
#' \itemize{
#'   \item Left column (full height): tree structure diagram (no legend)
#'   \item Top-right: variable importance bar chart
#'   \item Bottom-right: decision boundaries with Class legend at bottom
#' }
#'
#' Column widths adapt to the tree's x-span.  Row heights adapt to the
#' number of features.
#'
#' @param model A pptr model.
#' @param ... Passed through to \code{plot_boundaries()}.
#' @return A patchwork object (ggplot-compatible, works with ggsave).
#' @noRd
plot_mosaic <- function(model, ...) {
  check_patchwork()

  mosaic_title_size <- 11
  mosaic_margin <- ggplot2::theme(
    plot.title = ggplot2::element_text(size = mosaic_title_size),
    plot.margin = ggplot2::margin(5, 5, 5, 5)
  )

  p_structure <- plot_tree_structure(model) +
    mosaic_margin +
    ggplot2::theme(legend.position = "none")

  p_importance <- plot_importance(model) + mosaic_margin

  # Compact mode: no aspect ratio, no axis text, upper triangle blanked.
  # Pass title size so it matches the other mosaic titles.
  p_boundaries <- plot_boundaries(model, compact = mosaic_title_size, ...)
  if (inherits(p_boundaries, "gtable")) {
    p_boundaries <- patchwork::wrap_elements(full = p_boundaries)
  } else {
    p_boundaries <- p_boundaries + mosaic_margin
  }

  tree_frac <- 0.60

  n_features <- ncol(model$x)
  imp_ratio <- 1
  bnd_ratio <- if (n_features <= 2) 2 else if (n_features <= 4) 3 else 4

  p_structure + p_importance + p_boundaries +
    patchwork::plot_layout(
      design  = "AB\nAC",
      widths  = c(tree_frac, 1 - tree_frac),
      heights = c(imp_ratio, bnd_ratio)
    )
}

#' Mean-of-means split cutpoint strategy.
#'
#' Creates a split cutpoint strategy that splits at the midpoint between group
#' means. This is the default (and currently only) split cutpoint.
#'
#' @return A \code{cutpoint_strategy} object.
#' @seealso \code{\link{pptr}}, \code{\link{pprf}}, \code{\link{pp_pda}}, \code{\link{vars_uniform}}
#' @examples
#' cutpoint_mean_of_means()
#'
#' @export
cutpoint_mean_of_means <- function() {
  structure(list(name = "mean_of_means", display_name = "Mean of means"), class = "cutpoint_strategy")
}

describe("cutpoint_mean_of_means", {
  it("creates a cutpoint_strategy object", {
    s <- cutpoint_mean_of_means()
    expect_s3_class(s, "cutpoint_strategy")
    expect_equal(s$name, "mean_of_means")
  })
})

# Shared helper for regression tests.
# testthat auto-sources `helper-*.R` before any test file.

make_regression_data <- function(n = 30, seed = 0) {
  set.seed(seed)
  x <- matrix(runif(n * 2), ncol = 2)
  colnames(x) <- c("x1", "x2")
  y <- x[, 1] + x[, 2] + rnorm(n, sd = 0.1)
  list(x = x, y = y)
}

# ======================================================================
# Projection histogram plot
# ======================================================================

#' Retrieve a tree node by breadth-first (BFS) index.
#'
#' Traverses the tree in BFS order and returns the node at position
#' \code{index} (1-based).  Used to locate the target node for projection
#' plots when the user specifies a node number.
#'
#' @param root The root node of the tree (\code{$root} from a pptr model).
#' @param index 1-based BFS index (1 = root, 2 = first child, etc.).
#' @return The tree node at the specified BFS position.
#' @note Stops with an error if the index is out of range.
#' @noRd
get_node_bfs <- function(root, index) {
  queue <- list(root)
  i <- 0L

  while (length(queue) > 0) {
    current <- queue[[1]]
    queue <- queue[-1]
    i <- i + 1L

    if (i == index) return(current)

    if (!is.null(current$lower)) queue <- c(queue, list(current$lower))
    if (!is.null(current$upper)) queue <- c(queue, list(current$upper))
  }

  stop("Node index ", index, " is out of range.", call. = FALSE)
}

#' Render a projection histogram at a specified split node.
#'
#' Projects all training observations onto the node's projector vector,
#' draws overlapping histograms coloured by group, and marks the split
#' cutpoint with a dashed vertical line.  This shows how well the
#' node's projection separates the groups.
#'
#' @param model A pptr model with \code{$root}, \code{$x}, \code{$y},
#'   \code{$groups}.
#' @param node 1-based BFS index of the target node (default: root).
#' @param ... Currently unused.
#' @return A ggplot2 object.
#' @note Stops with an error if the specified node is a leaf.
#' @noRd
plot_projection <- function(model, node = 1L, ...) {
  target_node <- get_node_bfs(model$root, node)

  if (!is.null(target_node$value)) {
    stop("Node ", node, " is a leaf node and has no projector.", call. = FALSE)
  }

  projector <- target_node$projector
  projected <- as.numeric(model$x %*% projector)
  group_labels <- model$groups[model$y]

  df <- data.frame(
    projected = projected,
    group     = group_labels,
    stringsAsFactors = FALSE
  )

  cutpoint <- target_node$cutpoint

  ggplot2::ggplot(df, ggplot2::aes(x = projected, fill = group)) +
    ggplot2::geom_histogram(
      bins     = 30,
      alpha    = ppforest2_alpha_proj(),
      position = "identity"
    ) +
    ggplot2::geom_vline(
      xintercept = cutpoint,
      linetype   = "dashed",
      color      = ppforest2_col_cutpoint(),
      linewidth  = ppforest2_lw_medium()
    ) +
    ggplot2::labs(
      title = paste0("Projected Data at Node ", node),
      x     = "Projected Value",
      y     = "Count",
      fill  = "Class"
    ) +
    ppforest2_theme()
}

#' @useDynLib ppforest2
#' @importFrom Rcpp evalCpp
#' @importFrom stats model.frame model.matrix model.response formula predict sd terms update fitted residuals nobs
NULL

#' Trains a Projection-Pursuit oblique decision tree.
#'
#' This function trains a Projection-Pursuit oblique decision tree using either a formula and data frame interface or a matrix-based interface. When using the formula interface, specify the model formula and the data frame containing the variables. For the matrix-based interface, provide matrices for the features and labels directly.
#' If \code{lambda = 0}, the model is trained using Linear Discriminant Analysis (LDA). If \code{lambda > 0}, the model is trained using Penalized Discriminant Analysis (PDA).
#'
#' Mode is taken from the \code{mode} argument when explicit, and otherwise auto-detected from `y` (factor/character → classification, numeric → regression). Pass \code{mode = "classification"} to force classification on integer labels (e.g. binary 0/1), or \code{mode = "regression"} to assert intent on numeric responses.
#'
#' @param formula A formula of the form \code{y ~ x1 + x2 + ...}, where \code{y} is a vector of labels and \code{x1}, \code{x2}, ... are the features.
#' @param data A data frame containing the variables in the formula.
#' @param x A matrix containing the features for each observation.
#' @param y A matrix containing the labels for each observation.
#' @param mode Training mode: either \code{"classification"} or \code{"regression"}. When \code{NULL} (default), mode is auto-detected from \code{y}'s type — factor or character vectors trigger classification, numeric vectors trigger regression. Setting it explicitly is useful for the binary-integer-labels case (\code{mode = "classification"} with integer 0/1 labels) and for failing fast on a type mismatch (\code{mode = "regression"} with a factor \code{y} errors immediately).
#' @param lambda A regularization parameter (default: 0.5). If \code{lambda = 0}, the model is trained using Linear Discriminant Analysis (LDA). If \code{lambda > 0}, the model is trained using Penalized Discriminant Analysis (PDA). The default uses PDA because pure LDA (\code{lambda = 0}) is ill-conditioned when there are more variables than effective observations (see the "Known limitations" section of the README). Cannot be used together with \code{pp}.
#' @param seed An optional integer seed for reproducibility. If \code{NULL} (default), a seed is drawn from R's RNG, so \code{set.seed()} controls reproducibility. If an integer is provided, that value is used directly.
#' @param pp A projection pursuit strategy object created by \code{\link{pp_pda}}. Cannot be used together with \code{lambda}.
#' @param cutpoint A split cutpoint strategy object created by \code{\link{cutpoint_mean_of_means}} (default).
#' @param stop A stopping rule object. Default depends on mode:
#'   \code{\link{stop_pure_node}()} for classification, and
#'   \code{stop_any(stop_min_size(5), stop_min_variance(0.01))} for regression.
#' @param binarize A binarization strategy object. Default depends on mode:
#'   \code{\link{binarize_largest_gap}()} for classification, and
#'   \code{\link{binarize_disabled}()} for regression (regression's default
#'   grouping always yields a 2-group partition, so no binarization is needed).
#' @param grouping A grouping strategy object. Default depends on mode:
#'   \code{\link{grouping_by_label}()} for classification, and
#'   \code{\link{grouping_by_cutpoint}()} for regression.
#' @param leaf A leaf strategy object. Default depends on mode:
#'   \code{\link{leaf_majority_vote}()} for classification, and
#'   \code{\link{leaf_mean_response}()} for regression.
#' @return A \code{pptr} model. Its S3 class vector is
#'   \code{c("pptr_classification", "pptr", "ppmodel")} or
#'   \code{c("pptr_regression", "pptr", "ppmodel")} depending on the mode.
#' @seealso \code{\link{predict.pptr_classification}}, \code{\link{predict.pptr_regression}}, \code{\link{formula.ppmodel}}, \code{\link{print.pptr}}, \code{\link{save_json}}, \code{\link{load_json}}, \code{\link{pp_tree}} for parsnip integration
#' @examples
#'
#' # Example 1: formula interface with the `iris` dataset
#' pptr(Species ~ ., data = iris)
#'
#' # Example 2: formula interface with the `iris` dataset with regularization
#' pptr(Species ~ ., data = iris, lambda = 0.5)
#'
#' # Example 3: matrix interface with the `iris` dataset
#' pptr(x = iris[, 1:4], y = iris[, 5])
#'
#' @export
pptr <- function(
    formula = NULL,
    data = NULL,
    x = NULL,
    y = NULL,
    mode = NULL,
    lambda = 0.5,
    seed = NULL,
    pp = NULL,
    cutpoint = NULL,
    stop = NULL,
    binarize = NULL,
    grouping = NULL,
    leaf = NULL) {
  # See the matching comment in `pprf()` — capture the call up front so
  # `update()` can rebuild and re-evaluate it.
  cl <- match.call()

  if (!is.null(seed) && (!is.numeric(seed) || length(seed) != 1 || seed != as.integer(seed)))
    stop("`seed` must be a single integer or NULL.")

  args <- resolve_model_data(formula, data, x, y, mode = mode)
  mode <- args$mode

  strategies <- resolve_strategies(
    pp = pp, lambda = lambda, lambda_missing = missing(lambda),
    cutpoint = cutpoint, stop = stop, binarize = binarize, grouping = grouping,
    leaf = leaf)

  x <- args$x
  y <- args$y
  groups <- args$groups
  formula <- args$formula

  if (is.null(seed)) {
    seed <- sample.int(.Machine$integer.max, 1L)
  }

  training_spec <- list(
    pp = strategies$pp,
    vars = strategies$vars,
    cutpoint = strategies$cutpoint,
    stop = strategies$stop,
    binarize = strategies$binarize,
    grouping = strategies$grouping,
    leaf = strategies$leaf,
    mode = mode,
    size = 0L,
    seed = as.integer(seed),
    threads = 0L,
    max_retries = 3L)

  # `ppforest2_train` is mode-aware on the C++ side: it dispatches on
  # `training_spec$mode` and applies the appropriate index decode + sort.
  model <- ppforest2_train(training_spec, args$x, args$y)

  if (isTRUE(model$degenerate)) {
    warning("Some splits could not separate groups (degenerate nodes). ",
            "This can be caused by ill-conditioned variables in the input data. ",
            "Degenerate nodes predict the group with the most observations.",
            call. = FALSE)
  }

  model$call    <- cl
  model$seed    <- seed
  model$groups  <- groups
  model$formula <- formula
  model$mode    <- mode
  model$x       <- x
  model$y       <- y

  scale <- apply(x, 2, sd)
  scale[scale == 0] <- 1

  model$vi <- list(
    scale       = scale,
    projections = ppforest2_vi_projections_tree(model, ncol(x), scale)
  )

  model$.cache <- .new_cache()

  # Class is set by the Rcpp wrap layer (see `make_model_class` in
  # bindings/R/inst/include/ppforest2.h), which derives it from
  # `tree.training_spec->mode`. Don't reassign here — that would let the
  # R-side guess drift away from the C++ truth.

  model
}


# ---------------------------------------------------------------------------
# Prediction: split per mode.
# ---------------------------------------------------------------------------

#' Predicts labels or per-group one-hot proportions from a pptr model (classification mode).
#'
#' @param object A \code{pptr_classification} model.
#' @param new_data A data frame or matrix of new observations. If \code{NULL}, the first positional argument in \code{...} is used for backward compatibility.
#' @param type \code{"class"} (default) returns a factor of predicted labels; \code{"prob"} returns a data frame with 1.0 for the predicted group and 0.0 elsewhere.
#' @param ... Backward-compat positional `new_data`.
#' @return A factor or data frame.
#' @seealso \code{\link{pptr}}, \code{\link{predict.pptr_regression}}
#' @export
predict.pptr_classification <- function(object, new_data = NULL, type = NULL, ...) {
  x <- process_predict_arguments(object, new_data, ...)
  if (is.null(type)) type <- "class"

  if (type == "prob") {
    probs <- ppforest2_predict_tree_prob(object, x)
    df <- as.data.frame(probs)
    colnames(df) <- object$groups
    return(df)
  }

  if (type != "class") {
    stop("`type = \"", type, "\"` is not supported for classification trees. ",
         "Use \"class\" (default) or \"prob\".", call. = FALSE)
  }

  y <- ppforest2_predict_tree(object, x)
  as.factor(object$groups[y])
}

#' Predicts numeric responses from a pptr model (regression mode).
#'
#' @param object A \code{pptr_regression} model.
#' @param new_data A data frame or matrix of new observations.
#' @param type Must be \code{"response"} (default).
#' @param ... Backward-compat positional `new_data`.
#' @return A numeric vector.
#' @seealso \code{\link{pptr}}, \code{\link{predict.pptr_classification}}
#' @export
predict.pptr_regression <- function(object, new_data = NULL, type = NULL, ...) {
  x <- process_predict_arguments(object, new_data, ...)
  if (is.null(type)) type <- "response"

  if (type %in% c("class", "prob")) {
    stop("`type = \"", type, "\"` is not available for regression models. ",
         "Use `type = \"response\"`.", call. = FALSE)
  }

  if (type != "response") {
    stop("`type = \"", type, "\"` is not recognised. Use \"response\".", call. = FALSE)
  }

  as.numeric(ppforest2_predict_tree(object, x))
}


# ---------------------------------------------------------------------------
# print.pptr -- tree structure. Leaf label formatting is mode-specific via
# the `print_node` generic dispatched on the model's class.
# ---------------------------------------------------------------------------

#' Prints the structure of a pptr tree.
#' @param x A \code{pptr} model.
#' @param ... Unused.
#' @return Invisibly returns the input \code{pptr} model \code{x} (unchanged).
#'   Called for its side effect of printing the tree structure -- the oblique
#'   split rules and leaf predictions -- to the console.
#' @export
print.pptr <- function(x, ...) {
  cat("\n")
  if (!is.null(x$call)) {
    cat("Call: ", paste(deparse(x$call, width.cutoff = 80L), collapse = "\n      "), "\n\n", sep = "")
  }
  cat("Projection-Pursuit Oblique Decision Tree:\n")
  print_node(x, x$root)
  cat("\n")
  invisible(x)
}

# Internal generic: walks the tree recursively, dispatching on the model's
# class to format leaf values differently per mode.
print_node <- function(model, node, depth = 0) UseMethod("print_node")

#' @export
print_node.pptr_classification <- function(model, node, depth = 0) {
  .print_node_impl(model, node, depth, function(value) model$groups[value])
}

#' @export
print_node.pptr_regression <- function(model, node, depth = 0) {
  .print_node_impl(model, node, depth, function(value) format(as.numeric(value), digits = 4))
}

# Shared recursion. `format_leaf` turns the raw leaf value into a display string.
.print_node_impl <- function(model, node, depth, format_leaf) {
  indent <- paste(rep(" ", depth), collapse = "")

  if (!is.null(node$value)) {
    cat(indent, "Predict:", format_leaf(node$value), "\n")
    return(invisible(NULL))
  }

  projection_str <- paste(
    "[", paste(round(node$projector, 2), collapse = " "), "] * x",
    collapse = ""
  )

  cat(indent, "If (", projection_str, ") < ", node$cutpoint, ":\n", sep = "")

  if (!is.null(node$lower)) {
    print_node(model, node$lower, depth + 1)
  }

  cat(indent, "Else:\n", sep = "")

  if (!is.null(node$upper)) {
    print_node(model, node$upper, depth + 1)
  }
}


# ---------------------------------------------------------------------------
# summary -- layered via NextMethod:
#   summary.pptr_classification / summary.pptr_regression
#     -> summary.pptr (tree-level header + VI table)
#       -> summary.ppmodel (data summary block)
# ---------------------------------------------------------------------------

#' @export
summary.pptr <- function(object, ...) {
  model <- object
  if (is.null(model$x)) {
    cat("\n(Empty pptr model -- no training data available.)\n")
    return(invisible(model))
  }

  cat("\n")
  cat(if (identical(model$mode, "regression")) {
    "Projection-Pursuit Oblique Regression Tree\n"
  } else {
    "Projection-Pursuit Oblique Decision Tree\n"
  })
  cat("\n")
  print_training_spec(model$training_spec)

  NextMethod()  # summary.ppmodel

  invisible(model)
}

#' @export
summary.pptr_classification <- function(object, ...) {
  NextMethod()
  model <- object

  cat("Confusion Matrix:\n\n")
  print_confusion_matrix(ppforest2_predict_tree(model, model$x), model)
  cat("\n")

  .print_vi_table(model, include_oob_importances = FALSE)
  invisible(model)
}

#' @export
summary.pptr_regression <- function(object, ...) {
  NextMethod()
  model <- object

  preds <- ppforest2_predict_tree(model, model$x)
  y <- model$y
  mse <- mean((preds - y)^2)
  mae <- mean(abs(preds - y))
  ss_tot <- sum((y - mean(y))^2)
  r2 <- if (ss_tot > 0) 1 - sum((preds - y)^2) / ss_tot else 0
  cat("Training Metrics:\n")
  cat("  MSE:", format(mse, nsmall = 6), "\n")
  cat("  MAE:", format(mae, nsmall = 6), "\n")
  cat("  R\u00b2: ", format(r2, nsmall = 6), "\n\n")

  .print_vi_table(model, include_oob_importances = FALSE)
  invisible(model)
}

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

# Coverage for the standard R fit-class generics: fitted(), residuals(),
# nobs(). These are what downstream tools (broom::glance, step(), AIC(),
# users' ggplot diagnostic plots) call without thinking. Also locks in
# the `predict(fit, matrix)` path, which used to error whenever the model
# was fit via the formula interface and a caller wanted to predict on a
# raw numeric matrix (e.g. `model$x`).
describe("standard R accessors", {
  describe("nobs()", {
    it("returns the training-row count for pprf", {
      m <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
      expect_equal(nobs(m), nrow(iris))
    })

    it("returns the training-row count for pptr", {
      m <- pptr(Species ~ ., data = iris)
      expect_equal(nobs(m), nrow(iris))
    })

    it("returns NA when the model has no training data attached", {
      m <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
      m$x <- NULL
      expect_true(is.na(nobs(m)))
    })
  })

  describe("fitted()", {
    it("returns a factor matching the training response for classification", {
      set.seed(0)
      m <- pprf(Species ~ ., data = iris, size = 5, threads = 1)
      fv <- fitted(m)
      expect_s3_class(fv, "factor")
      expect_equal(length(fv), nrow(iris))
      # In-sample accuracy on iris should be near-perfect for a 5-tree forest
      # (this is a sanity check, not a quality claim — see oob_predictions()
      # for an honest estimate).
      expect_gt(mean(fv == iris$Species), 0.9)
    })

    it("returns a numeric vector for regression", {
      set.seed(0)
      df <- data.frame(x1 = runif(40), x2 = runif(40))
      df$y <- df$x1 + 0.5 * df$x2 + rnorm(40, sd = 0.1)
      m <- pprf(y ~ ., data = df, size = 5, threads = 1)
      fv <- fitted(m)
      expect_true(is.numeric(fv))
      expect_equal(length(fv), nrow(df))
    })

    it("works on pptr too", {
      m <- pptr(Species ~ ., data = iris)
      expect_s3_class(fitted(m), "factor")
    })
  })

  describe("residuals()", {
    it("returns y - fitted for regression models", {
      set.seed(0)
      df <- data.frame(x1 = runif(40), x2 = runif(40))
      df$y <- df$x1 + rnorm(40, sd = 0.1)
      m <- pprf(y ~ ., data = df, size = 5, threads = 1)
      r <- residuals(m)
      expect_true(is.numeric(r))
      expect_equal(length(r), nrow(df))
      # `pprf` sorts rows by y in regression mode, so `model$y` is NOT in the
      # original input order. The residual contract is "training y minus
      # in-sample prediction" — both are taken from the (sorted) training
      # state, not from the caller's pre-sort data frame.
      expect_equal(r, as.numeric(m$y) - as.numeric(fitted(m)), tolerance = 1e-10)
    })

    it("errors with a clear message on classification models", {
      m <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
      expect_error(residuals(m), "only defined for regression")
    })
  })

  describe("predict() accepts a numeric matrix", {
    # Regression: previously `predict(fit, fit$x)` errored with "data must
    # be a data.frame" whenever the model was fit via the formula interface.
    # Now matrix input bypasses model.matrix.
    it("accepts model$x (matrix) directly for formula-fitted pprf", {
      m <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
      p <- predict(m, m$x)
      expect_s3_class(p, "factor")
      expect_equal(length(p), nrow(iris))
    })

    it("accepts a matrix for formula-fitted pptr", {
      m <- pptr(Species ~ ., data = iris)
      p <- predict(m, m$x)
      expect_s3_class(p, "factor")
      expect_equal(length(p), nrow(iris))
    })

    it("still accepts a data.frame (regression path)", {
      m <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
      expect_s3_class(predict(m, iris), "factor")
    })
  })
})

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

describe("predict.pptr", {
  describe("on an object created with the formula interface", {
    it("returns a factor with the same length as the input matrix", {
      model <- pptr(Species ~ ., data = iris)
      predictions <- predict(model, iris)
      expect_equal(length(predictions), nrow(iris))
    })

    it("returns a factor with the same levels as the groups in the model", {
      model <- pptr(Species ~ ., data = iris)
      predictions <- predict(model, iris)
      expect_equal(levels(predictions), levels(iris$Species))
    })

    it("with new_data parameter returns the same result as positional", {
      model <- pptr(Species ~ ., data = iris)
      pred_positional <- predict(model, iris)
      pred_named <- predict(model, new_data = iris)
      expect_equal(pred_positional, pred_named)
    })
  })

  describe("on an object created with the matrix interface", {
    it("returns a factor with the same length as the input matrix", {
      x <- crabs[, 2:5]
      x$sex <- as.numeric(as.factor(crabs$sex))
      model <- pptr(x = x, y = crabs$Type)
      predictions <- predict(model, x)
      expect_equal(length(predictions), nrow(x))
    })

    it("returns a factor with the same levels as the groups in the model", {
      x <- crabs[, 2:5]
      x$sex <- as.numeric(as.factor(crabs$sex))
      model <- pptr(x = x, y = crabs$Type)
      predictions <- predict(model, x)
      expect_equal(levels(predictions), levels(crabs$Type))
    })
  })

  describe("with type = 'prob'", {
    it("returns a data frame with one column per group", {
      model <- pptr(Species ~ ., data = iris)
      probs <- predict(model, iris, type = "prob")
      expect_true(is.data.frame(probs))
      expect_equal(ncol(probs), length(levels(iris$Species)))
      expect_equal(colnames(probs), levels(iris$Species))
    })

    it("returns exactly one 1.0 per row and the rest 0.0", {
      model <- pptr(Species ~ ., data = iris)
      probs <- predict(model, iris, type = "prob")
      row_sums <- rowSums(probs)
      expect_equal(row_sums, rep(1.0, nrow(iris)))
      expect_true(all(probs == 0 | probs == 1))
    })

    it("the 1.0 column matches the group prediction", {
      model <- pptr(Species ~ ., data = iris)
      group_preds <- predict(model, iris, type = "class")
      prob_preds <- predict(model, iris, type = "prob")
      for (i in seq_len(nrow(iris))) {
        expect_equal(prob_preds[i, as.character(group_preds[i])], 1.0)
      }
    })
  })
})

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

skip_if_not_installed("ggplot2")

describe("plot.pptr importance", {
  it("returns a ggplot object for type = 'importance'", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    p <- plot(model, type = "importance")
    expect_s3_class(p, "ggplot")
  })

  it("importance plot orders variables by value", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    p <- plot(model, type = "importance")
    pdata <- ggplot2::ggplot_build(p)$layout$panel_params[[1]]$y$get_labels()
    vi <- model$vi$projections
    vnames <- colnames(model$x)
    expected_order <- vnames[order(vi)]
    expect_equal(pdata, expected_order)
  })
})

describe("plot.pprf importance", {
  it("default plot renders importance grid without error", {
    model <- pprf(Species ~ ., data = iris, size = 5, seed = 0, threads = 1)
    expect_no_error(plot(model))
  })

  it("renders importance grid for type = 'importance' without metric", {
    model <- pprf(Species ~ ., data = iris, size = 5, seed = 0, threads = 1)
    expect_no_error(plot(model, type = "importance"))
  })

  it("returns a ggplot for a single importance metric", {
    model <- pprf(Species ~ ., data = iris, size = 5, seed = 0, threads = 1)
    p <- plot(model, metric = "projections")
    expect_s3_class(p, "ggplot")
    p <- plot(model, metric = "weighted")
    expect_s3_class(p, "ggplot")
    p <- plot(model, metric = "permuted")
    expect_s3_class(p, "ggplot")
  })

  it("importance plot orders variables by selected metric", {
    model <- pprf(Species ~ ., data = iris, size = 5, seed = 0, threads = 1)
    p <- plot(model, metric = "permuted")
    pdata <- ggplot2::ggplot_build(p)$layout$panel_params[[1]]$y$get_labels()
    vi <- permuted_importance(model)
    vnames <- colnames(model$x)
    expected_order <- vnames[order(vi)]
    expect_equal(pdata, expected_order)
  })
})

describe("plot.pptr importance snapshots", {
  skip_if_not_installed("vdiffr")
  skip_on_ci()  # vdiffr SVGs aren't byte-identical across platform LAPACK/BLAS

  model <- pptr(Species ~ ., data = iris, seed = 0)

  it("pptr-importance", {
    vdiffr::expect_doppelganger("pptr-importance", plot(model, type = "importance"))
  })
})

describe("plot.pprf importance snapshots", {
  skip_if_not_installed("vdiffr")
  skip_on_ci()  # vdiffr SVGs aren't byte-identical across platform LAPACK/BLAS

  # Pin to all 4 features so the snapshot is stable under the default
  # variable subsampling (p_vars = 0.5).
  model <- pprf(Species ~ ., data = iris, size = 5, n_vars = 4, seed = 0, threads = 1)

  it("pprf-importance-projections", {
    vdiffr::expect_doppelganger("pprf-importance-projections", plot(model, metric = "projections"))
  })

  it("pprf-importance-permuted", {
    vdiffr::expect_doppelganger("pprf-importance-permuted", plot(model, metric = "permuted"))
  })
})

describe("plot.pprf regression importance snapshot", {
  # Regression-side VI share the three measures with classification
  # (projections / permuted / weighted), but the underlying numbers come
  # from MSE-increase (regression) rather than accuracy-drop
  # (classification). The snapshot fences the end-to-end render for a
  # regression model so future refactors of the VI-computation path
  # (unit scale, sign handling, label formatting) can't silently alter
  # the plotted output.
  skip_if_not_installed("vdiffr")
  skip_on_ci()  # vdiffr SVGs aren't byte-identical across platform LAPACK/BLAS

  data(mtcars)
  # Pin to all 10 features so the snapshot is stable under the default
  # variable subsampling (p_vars = 0.5).
  model <- pprf(mpg ~ ., data = mtcars, size = 5, n_vars = 10, seed = 0, threads = 1)

  it("pprf-regression-importance-projections", {
    vdiffr::expect_doppelganger(
      "pprf-regression-importance-projections",
      plot(model, metric = "projections")
    )
  })
})

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

skip_if_not_installed("ggplot2")

describe("plot.pptr structure", {
  it("returns a ggplot object for type = 'structure'", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    p <- plot(model, type = "structure")
    expect_s3_class(p, "ggplot")
  })

  it("tree plot uses C++ node data with histograms", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    nodes <- ppforest2:::ppforest2_tree_node_data(model, model$x, model$y)
    expect_true(length(nodes) > 0)
    expect_false(nodes[[1]]$is_leaf)
    expect_true(length(nodes[[1]]$projected) > 0)
    expect_true(length(nodes[[1]]$groups) > 0)
  })

  it("default plot renders mosaic without error", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    expect_no_error(plot(model))
  })
})

describe("plot.pprf structure", {
  it("returns a ggplot object for type = 'structure'", {
    model <- pprf(Species ~ ., data = iris, size = 5, seed = 0, threads = 1)
    p <- plot(model, type = "structure", tree_index = 1)
    expect_s3_class(p, "ggplot")
  })
})

describe("plot.pptr structure snapshots", {
  skip_if_not_installed("vdiffr")
  skip_on_ci()  # vdiffr SVGs aren't byte-identical across platform LAPACK/BLAS

  model <- pptr(Species ~ ., data = iris, seed = 0)

  it("pptr-structure", {
    vdiffr::expect_doppelganger("pptr-structure", plot(model, type = "structure"))
  })
})

describe("plot.pprf structure snapshots", {
  skip_if_not_installed("vdiffr")
  skip_on_ci()  # vdiffr SVGs aren't byte-identical across platform LAPACK/BLAS

  # Pin to all 4 features so the snapshot is stable under the default
  # variable subsampling (p_vars = 0.5).
  model <- pprf(Species ~ ., data = iris, size = 5, n_vars = 4, seed = 0, threads = 1)

  it("pprf-structure", {
    vdiffr::expect_doppelganger("pprf-structure", plot(model, type = "structure", tree_index = 1))
  })
})

describe("plot.pptr regression structure snapshot", {
  # Regression trees render their leaves as mean-response values (numeric
  # labels) rather than majority-vote class names. The snapshot fences
  # that difference: if a future change accidentally re-routes regression
  # leaves through the classification renderer (or vice versa), the
  # rendered text labels will diverge from the saved image.
  skip_if_not_installed("vdiffr")
  skip_on_ci()  # vdiffr SVGs aren't byte-identical across platform LAPACK/BLAS

  data(mtcars)
  model <- pptr(mpg ~ ., data = mtcars, seed = 0)

  it("pptr-regression-structure", {
    vdiffr::expect_doppelganger("pptr-regression-structure", plot(model, type = "structure"))
  })
})

describe("format_projector", {
  it("formats top-k terms by magnitude", {
    fmt <- ppforest2:::format_projector(
      c(0.1, -0.7, 0.3, 0.05),
      c("a", "b", "c", "d"),
      max_terms = 2L
    )
    expect_true(grepl("b", fmt))
    expect_true(grepl("c", fmt))
    expect_true(grepl("\u2026", fmt))
  })

  it("shows all terms when fewer than max_terms", {
    fmt <- ppforest2:::format_projector(
      c(0.5, -0.3),
      c("x1", "x2"),
      max_terms = 3L
    )
    expect_true(grepl("x1", fmt))
    expect_true(grepl("x2", fmt))
    expect_false(grepl("\u2026", fmt))
  })

  it("returns '0' for a zero projector", {
    fmt <- ppforest2:::format_projector(
      c(0, 0, 0),
      c("a", "b", "c"),
      max_terms = 3L
    )
    expect_equal(fmt, "0")
  })
})

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

skip_if_not_installed("ggplot2")

describe("plot.pptr projection", {
  it("returns a ggplot object for type = 'projection'", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    p <- plot(model, type = "projection")
    expect_s3_class(p, "ggplot")
  })

  it("errors for out-of-range node index", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    expect_error(
      plot(model, type = "projection", node = 100L),
      "out of range"
    )
  })
})

describe("plot.pprf projection", {
  it("returns a ggplot object for type = 'projection'", {
    model <- pprf(Species ~ ., data = iris, size = 5, seed = 0, threads = 1)
    p <- plot(model, type = "projection", tree_index = 1)
    expect_s3_class(p, "ggplot")
  })
})

describe("plot.pptr projection snapshots", {
  skip_if_not_installed("vdiffr")
  skip_on_ci()  # vdiffr SVGs aren't byte-identical across platform LAPACK/BLAS

  model <- pptr(Species ~ ., data = iris, seed = 0)

  it("pptr-projection-root", {
    vdiffr::expect_doppelganger("pptr-projection-root", plot(model, type = "projection", node = 1L))
  })
})

# ======================================================================
# Decision boundary plots
# ======================================================================

#' Dispatch to the appropriate boundary plot based on feature count.
#'
#' Selects the visualization strategy based on the number of features (p):
#' \describe{
#'   \item{p = 1}{1D number-line plot.  Background rectangles show decision
#'     regions; solid vertical lines mark boundaries; jittered points show
#'     the data.}
#'   \item{p = 2}{Single 2D scatterplot.  Convex region polygons (from C++
#'     Sutherland-Hodgman clipping) fill the background; boundary line
#'     segments are drawn on top; data points are overlaid.  Uses
#'     \code{coord_fixed(ratio = 1)} for undistorted angles.}
#'   \item{p >= 3}{Pairwise scatterplot matrix (lower triangle).  Each panel
#'     shows a 2D projection with non-displayed variables held at their
#'     medians.  \code{facet_grid} with free scales and zero expansion
#'     ensures full region coverage per panel.}
#' }
#'
#' @param model A pptr model with \code{$root}, \code{$x}, \code{$y},
#'   \code{$groups}.
#' @param ... Passed through to the specific boundary plot function.
#' @return A ggplot2 object.
#' @noRd
plot_boundaries <- function(model, compact = FALSE, ...) {
  p <- ncol(model$x)

  if (p == 1L) {
    plot_boundaries_1d(model)
  } else if (p == 2L) {
    plot_boundaries_2d(model, ...)
  } else {
    plot_boundaries_pairwise(model, compact = compact, ...)
  }
}

#' Convert C++ region polygon data to a ggplot2-ready data frame.
#'
#' Transforms the list-of-lists output from \code{ppforest2_decision_regions()}
#' (C++ RegionVisitor) into a flat data frame suitable for
#' \code{ggplot2::geom_polygon()}.  Each polygon becomes a group of rows
#' sharing a unique region_id, with vertices in the order produced by the
#' Sutherland-Hodgman algorithm (consistent winding for correct polygon
#' rendering).
#'
#' Group indices from C++ are 1-based (matching R's convention for the
#' response vector) and are mapped to human-readable labels via the
#' model's group vector.
#'
#' @param regions List of region lists, each with \code{$x}, \code{$y}
#'   (numeric vectors of vertex coordinates) and \code{$group} (integer
#'   group index, 1-based from C++).
#' @param group_labels Character vector of group names (model$groups).
#' @return Data frame with columns x, y, region_id, region_group, or
#'   NULL if no valid polygons exist.  Degenerate regions with fewer
#'   than 3 vertices are silently dropped.
#' @noRd
build_region_df <- function(regions, group_labels) {
  if (length(regions) == 0) return(NULL)

  dfs <- list()
  for (i in seq_along(regions)) {
    r <- regions[[i]]
    if (length(r$x) >= 3) {
      dfs[[length(dfs) + 1L]] <- data.frame(
        x = r$x, y = r$y,
        region_id = i,
        region_group = group_labels[r$group],
        stringsAsFactors = FALSE
      )
    }
  }

  if (length(dfs) == 0) return(NULL)
  do.call(rbind, dfs)
}

#' Render decision boundaries for a 1-feature model.
#'
#' Draws a number-line plot with three layers (back to front):
#' \enumerate{
#'   \item Coloured background rectangles for each decision region
#'     (predicted group determined by running \code{ppforest2_predict_tree} on each
#'     region's midpoint)
#'   \item Jittered data points coloured by true group (y is meaningless,
#'     jittered around 0 for visual separation)
#'   \item Solid vertical lines at each split's decision boundary
#' }
#'
#' Boundary positions are computed as \code{cutpoint / projector[1]} for
#' each internal node.  For a 1D projector \code{a = [a1]}, the split
#' condition \code{a^T x < t} becomes \code{a1 * x < t}, i.e.
#' \code{x < t/a1}.  The sorted boundary positions partition the x-axis
#' into contiguous regions, each assigned the group predicted by the tree
#' for that region's midpoint.
#'
#' The y-axis is suppressed (no title, no ticks, no labels) since the
#' vertical dimension has no meaning — it exists only to spread out the
#' jittered points for readability.
#'
#' @param model A pptr model with 1 feature column in \code{$x}.
#' @return A ggplot2 object (no y-axis).
#' @noRd
plot_boundaries_1d <- function(model) {
  vnames <- get_variable_names(model)
  all_group_labels <- model$groups
  group_colors <- get_group_colors(all_group_labels)
  obs_group_labels <- all_group_labels[model$y]

  nodes <- ppforest2_tree_node_data(model, model$x, model$y)

  boundary_x <- vapply(
    Filter(function(nd) !nd$is_leaf, nodes),
    function(nd) nd$cutpoint / nd$projector[1],
    numeric(1)
  )
  boundary_x <- sort(boundary_x)

  x_range <- pad_range(range(model$x[, 1]))

  # Build region rectangles from sorted boundaries
  # Predict the group for the midpoint of each region
  edges <- c(x_range[1], boundary_x, x_range[2])
  region_rects <- list()
  for (r in seq_len(length(edges) - 1)) {
    mid <- (edges[r] + edges[r + 1]) / 2
    mid_vec <- matrix(mid, nrow = 1)
    pred_group <- all_group_labels[ppforest2_predict_tree(model, mid_vec)]
    region_rects[[r]] <- data.frame(
      xmin = edges[r], xmax = edges[r + 1],
      region_group = pred_group,
      stringsAsFactors = FALSE
    )
  }
  region_df <- do.call(rbind, region_rects)

  obs_data <- data.frame(
    x     = model$x[, 1],
    group = obs_group_labels,
    stringsAsFactors = FALSE
  )

  # Make the plot square: y-range equals x-range width, centred at 0
  x_span <- diff(x_range)
  y_range <- c(-x_span / 2, x_span / 2)

  ggplot2::ggplot() +
    ggplot2::geom_rect(
      data = region_df,
      ggplot2::aes(xmin = xmin, xmax = xmax, ymin = y_range[1], ymax = y_range[2],
                   fill = region_group),
      alpha = ppforest2_alpha_region()
    ) +
    ggplot2::geom_jitter(
      data   = obs_data,
      ggplot2::aes(x = x, y = 0, color = group),
      height = 0.3, size = ppforest2_pt_medium()
    ) +
    ggplot2::geom_vline(
      xintercept = boundary_x,
      linetype   = "solid",
      color      = ppforest2_col_boundary(),
      linewidth  = ppforest2_lw_medium()
    ) +
    ggplot2::scale_fill_manual(values = group_colors, name = "Class") +
    ggplot2::scale_color_manual(values = group_colors, name = "Class") +
    ggplot2::scale_x_continuous(expand = ggplot2::expansion(0)) +
    ggplot2::scale_y_continuous(expand = ggplot2::expansion(0)) +
    ggplot2::labs(
      title = "Decision Boundaries",
      x     = vnames[1]
    ) +
    ggplot2::coord_cartesian(xlim = x_range, ylim = y_range) +
    ppforest2_theme() +
    ggplot2::theme(
      aspect.ratio = 1,
      axis.title.y = ggplot2::element_blank(),
      axis.text.y  = ggplot2::element_blank(),
      axis.ticks.y = ggplot2::element_blank()
    )
}

#' Render decision boundaries for a 2-feature model.
#'
#' Draws a single scatterplot with:
#' \itemize{
#'   \item Convex region polygons (Sutherland-Hodgman clipped) coloured by
#'     predicted group, computed by \code{ppforest2_decision_regions()} in C++
#'   \item Boundary line segments (parametric-clipped) from
#'     \code{ppforest2_boundary_segments()} in C++
#'   \item Data points coloured by true group
#' }
#'
#' The plot uses \code{coord_fixed(ratio = 1)} with equalized ranges so
#' the decision boundaries appear at their true angles.  Region polygons
#' are computed on an extended bounding box (0.5x padding) so they reach
#' the plot edges.
#'
#' @param model A pptr model with 2 feature columns in \code{$x}.
#' @param ... Passed to \code{geom_segment} (e.g. additional aesthetics).
#' @return A ggplot2 object.
#' @noRd
plot_boundaries_2d <- function(model, ...) {
  vnames <- get_variable_names(model)
  group_labels <- model$groups
  group_colors <- get_group_colors(group_labels)

  obs_group_labels <- group_labels[model$y]

  x_range <- pad_range(range(model$x[, 1]))
  y_range <- pad_range(range(model$x[, 2]))

  # Equalize ranges so the plot is square with coord_fixed
  eq <- equalize_ranges(x_range, y_range)
  x_range <- eq$x
  y_range <- eq$y

  # Use a larger bounding box for regions so they extend to the plot edges
  x_ext <- diff(x_range) * 0.5
  y_ext <- diff(y_range) * 0.5
  region_bbox <- c(x_range[1] - x_ext, x_range[2] + x_ext,
                   y_range[1] - y_ext, y_range[2] + y_ext)

  # Decision regions (computed over extended bbox, clipped visually by coord)
  regions <- ppforest2_decision_regions(
    model, c(0L, 1L), numeric(0),
    region_bbox[1], region_bbox[2], region_bbox[3], region_bbox[4]
  )
  region_df <- build_region_df(regions, group_labels)

  # Boundary segments
  segments <- ppforest2_boundary_segments(
    model, c(0L, 1L), numeric(0),
    region_bbox[1], region_bbox[2], region_bbox[3], region_bbox[4]
  )

  obs_data <- data.frame(
    x     = model$x[, 1],
    y     = model$x[, 2],
    group = obs_group_labels,
    stringsAsFactors = FALSE
  )

  p <- ggplot2::ggplot()

  # Regions
  if (!is.null(region_df)) {
    p <- p + ggplot2::geom_polygon(
      data = region_df,
      ggplot2::aes(x = x, y = y, group = region_id, fill = region_group),
      alpha = ppforest2_alpha_region()
    )
  }

  # Boundary lines
  if (nrow(segments) > 0) {
    p <- p + ggplot2::geom_segment(
      data = segments,
      ggplot2::aes(x = x_start, y = y_start, xend = x_end, yend = y_end),
      linewidth = ppforest2_lw_medium(), color = ppforest2_col_boundary(), ...
    )
  }

  # Data points
  p <- p + ggplot2::geom_point(
    data = obs_data,
    ggplot2::aes(x = x, y = y, color = group),
    size = ppforest2_pt_medium()
  )

  p +
    ggplot2::scale_fill_manual(values = group_colors, name = "Class") +
    ggplot2::scale_color_manual(values = group_colors, name = "Class") +
    ggplot2::labs(
      title = "Decision Boundaries",
      x     = vnames[1],
      y     = vnames[2]
    ) +
    ggplot2::coord_fixed(ratio = 1, xlim = x_range, ylim = y_range) +
    ppforest2_theme()
}

#' Render pairwise decision boundaries for a model with 3+ features.
#'
#' Creates a lower-triangular scatterplot matrix using \code{facet_grid()}.
#' For each pair of features (i, j), all other features are held at
#' their median values and the tree's decision regions and boundary
#' lines are projected onto the (i, j) plane.
#'
#' Layout uses \code{facet_grid(y_var ~ x_var, switch = "both")} so that
#' column strips appear at the bottom (serving as x-axis labels) and
#' row strips appear on the left (serving as y-axis labels), similar
#' to a \code{pairs()} plot.  All panels use free scales and
#' \code{aspect.ratio = 1}.
#'
#' @section Full region coverage:
#' The region bounding box is set to exactly the padded data range
#' (data range +/- 5\%).  Scale expansion is disabled via
#' \code{expansion(0)}, so the visible coordinate range equals the
#' polygon bounding box.  Since the Sutherland-Hodgman algorithm
#' partitions the entire bbox across all leaves, the region polygons
#' tile the panel completely with no whitespace gaps.
#'
#' @section Axis labels:
#' The overall x/y axis titles are suppressed.  The strip labels
#' (placed at the axis positions via \code{switch = "both"}) serve as
#' the axis labels, showing the variable name for each row and column.
#'
#' Remove empty upper-triangle panels from a pairwise facet_grid plot.
#'
#' Converts a ggplot with \code{facet_grid(y_var ~ x_var)} to a gtable
#' and replaces panel grobs in the upper triangle (where the column
#' index exceeds the row index) with \code{nullGrob()}.  This prevents
#' empty panels from showing grid lines and background.
#'
#' @param p A ggplot object produced by the pairwise boundary builder.
#' @return A gtable with upper-triangle panel grobs blanked out.
#' @noRd
strip_upper_triangle <- function(p) {
  g <- ggplot2::ggplotGrob(p)

  panels <- g$layout[grepl("^panel", g$layout$name), ]
  panel_rows <- sort(unique(panels$t))
  panel_cols <- sort(unique(panels$l))

  # Collect upper-triangle positions (facet_col > facet_row)
  upper_tri <- data.frame(
    gtable_row = integer(0), gtable_col = integer(0)
  )
  for (i in seq_len(nrow(panels))) {
    facet_row <- match(panels$t[i], panel_rows)
    facet_col <- match(panels$l[i], panel_cols)
    if (facet_col > facet_row) {
      upper_tri <- rbind(upper_tri, data.frame(
        gtable_row = panels$t[i], gtable_col = panels$l[i]
      ))
    }
  }

  if (nrow(upper_tri) == 0) return(g)

  # Blank all grobs that overlap upper-triangle panel cells
  for (j in seq_len(nrow(g$layout))) {
    row_j <- g$layout$t[j]
    col_j <- g$layout$l[j]
    for (k in seq_len(nrow(upper_tri))) {
      if (row_j == upper_tri$gtable_row[k] &&
          col_j == upper_tri$gtable_col[k]) {
        g$grobs[[j]] <- grid::nullGrob()
      }
    }
  }

  g
}

#' @param model A pptr model with 3+ feature columns in \code{$x}.
#' @param ... Passed to \code{geom_segment} (e.g. additional aesthetics).
#' @return A gtable with empty upper-triangle panels removed.
#' @noRd
plot_boundaries_pairwise <- function(model, compact = FALSE, ...) {
  vnames <- get_variable_names(model)
  group_labels <- model$groups
  group_colors <- get_group_colors(group_labels)
  n_vars <- ncol(model$x)

  obs_group_labels <- group_labels[model$y]
  medians <- apply(model$x, 2, stats::median)

  # Generate all pairwise combinations (lower triangle of scatterplot matrix)
  pairs <- utils::combn(n_vars, 2, simplify = FALSE)

  # Factor levels for grid layout: x_var = vars 1..(p-1), y_var = vars 2..p
  x_var_levels <- vnames[seq_len(n_vars - 1L)]
  y_var_levels <- vnames[seq(2L, n_vars)]

  all_segments <- list()
  all_regions <- list()
  all_obs <- list()

  for (k in seq_along(pairs)) {
    i <- pairs[[k]][1]
    j <- pairs[[k]][2]
    other <- setdiff(seq_len(n_vars), c(i, j))

    x_range <- pad_range(range(model$x[, i]))
    y_range <- pad_range(range(model$x[, j]))

    # Region bbox matches the padded data range exactly.  Combined with
    # scale expansion(0) below, this ensures region polygons tile each
    # panel completely: the polygon vertices define the scale limits, and
    # zero expansion prevents ggplot2 from adding extra whitespace beyond
    # the polygon edges.
    region_bbox <- c(x_range[1], x_range[2], y_range[1], y_range[2])

    # Decision regions (extended bbox, clipped visually by facet coord)
    regs <- ppforest2_decision_regions(
      model, c(i - 1L, j - 1L), medians[other],
      region_bbox[1], region_bbox[2], region_bbox[3], region_bbox[4]
    )
    reg_df <- build_region_df(regs, group_labels)
    if (!is.null(reg_df)) {
      # Offset region_id to be unique across panels
      offset <- if (length(all_regions) > 0) {
        max(do.call(rbind, all_regions)$region_id)
      } else {
        0
      }
      reg_df$region_id <- reg_df$region_id + offset
      reg_df$x_var <- factor(vnames[i], levels = x_var_levels)
      reg_df$y_var <- factor(vnames[j], levels = y_var_levels)
      all_regions[[k]] <- reg_df
    }

    # Boundary segments (also extended)
    segs <- ppforest2_boundary_segments(
      model, c(i - 1L, j - 1L), medians[other],
      region_bbox[1], region_bbox[2], region_bbox[3], region_bbox[4]
    )

    if (nrow(segs) > 0) {
      segs$x_var <- factor(vnames[i], levels = x_var_levels)
      segs$y_var <- factor(vnames[j], levels = y_var_levels)
      all_segments[[k]] <- segs
    }

    all_obs[[k]] <- data.frame(
      x     = model$x[, i],
      y     = model$x[, j],
      group = obs_group_labels,
      x_var = factor(vnames[i], levels = x_var_levels),
      y_var = factor(vnames[j], levels = y_var_levels),
      stringsAsFactors = FALSE
    )
  }

  obs_data <- do.call(rbind, all_obs)
  seg_data <- do.call(rbind, Filter(Negate(is.null), all_segments))
  reg_data <- do.call(rbind, Filter(Negate(is.null), all_regions))

  p <- ggplot2::ggplot()

  # Regions
  if (!is.null(reg_data) && nrow(reg_data) > 0) {
    p <- p + ggplot2::geom_polygon(
      data = reg_data,
      ggplot2::aes(x = x, y = y, group = region_id, fill = region_group),
      alpha = ppforest2_alpha_region()
    )
  }

  # Boundary lines
  if (!is.null(seg_data) && nrow(seg_data) > 0) {
    p <- p + ggplot2::geom_segment(
      data = seg_data,
      ggplot2::aes(x = x_start, y = y_start, xend = x_end, yend = y_end),
      linewidth = ppforest2_lw_medium(), color = ppforest2_col_boundary(), ...
    )
  }

  # Data points
  p <- p + ggplot2::geom_point(
    data = obs_data,
    ggplot2::aes(x = x, y = y, color = group),
    size = ppforest2_pt_small()
  )

  p <- p +
    ggplot2::facet_grid(y_var ~ x_var, scales = "free", switch = "x") +
    ggplot2::scale_fill_manual(values = group_colors, name = "Class") +
    ggplot2::scale_color_manual(values = group_colors, name = "Class") +
    ggplot2::scale_x_continuous(expand = ggplot2::expansion(0)) +
    ggplot2::scale_y_continuous(expand = ggplot2::expansion(0)) +
    ggplot2::labs(title = "Decision Boundaries") +
    ppforest2_theme() +
    ggplot2::theme(
      aspect.ratio     = 1,
      strip.placement  = "outside",
      axis.title       = ggplot2::element_blank(),
      axis.text.y      = ggplot2::element_blank(),
      axis.ticks.y     = ggplot2::element_blank(),
      legend.position  = "bottom"
    )


  # Compact mode: drop aspect ratio and axis text so panels fill all
  # available space (used by the mosaic where room is constrained).
  if (is.numeric(compact)) {
    # compact is the title size in pt
    p <- p + ggplot2::theme(
      aspect.ratio = NULL,
      axis.text    = ggplot2::element_blank(),
      axis.ticks   = ggplot2::element_blank(),
      plot.title   = ggplot2::element_text(size = compact)
    )
  } else if (isTRUE(compact)) {
    p <- p + ggplot2::theme(
      aspect.ratio = NULL,
      axis.text    = ggplot2::element_blank(),
      axis.ticks   = ggplot2::element_blank()
    )
  }

  strip_upper_triangle(p)
}

#' @useDynLib ppforest2
#' @importFrom Rcpp evalCpp
NULL

#' PDA projection pursuit strategy.
#'
#' Creates a Penalized Discriminant Analysis (PDA) projection pursuit strategy
#' for use with \code{\link{pptr}} or \code{\link{pprf}}.
#'
#' @param lambda A regularization parameter between 0 and 1.
#'   If \code{lambda = 0}, the model uses Linear Discriminant Analysis (LDA).
#'   If \code{lambda > 0}, the model uses Penalized Discriminant Analysis (PDA).
#' @return A \code{pp_strategy} object.
#' @seealso \code{\link{pptr}}, \code{\link{pprf}}, \code{\link{vars_uniform}}, \code{\link{vars_all}}, \code{\link{cutpoint_mean_of_means}}
#' @examples
#' # PDA with lambda = 0.5
#' pp_pda(0.5)
#'
#' # Use with pptr
#' pptr(Species ~ ., data = iris, pp = pp_pda(0.5))
#'
#' # Use with pprf
#' pprf(Species ~ ., data = iris, pp = pp_pda(0.5), vars = vars_uniform(n_vars = 2))
#'
#' @export
pp_pda <- function(lambda = 0) {
  if (!is.numeric(lambda) || length(lambda) != 1 || lambda < 0 || lambda > 1)
    stop("`lambda` must be a single number between 0 and 1.")
    
  structure(list(name = "pda", display_name = ifelse(lambda == 0, "LDA", "PDA"), lambda = lambda), class = "pp_strategy")
}

describe("pp_pda", {
  it("creates a pp_strategy object", {
    s <- pp_pda(0.5)
    expect_s3_class(s, "pp_strategy")
    expect_equal(s$name, "pda")
    expect_equal(s$lambda, 0.5)
  })

  it("defaults lambda to 0", {
    s <- pp_pda()
    expect_equal(s$lambda, 0)
  })

  it("rejects invalid lambda", {
    expect_error(pp_pda(-1), "between 0 and 1")
    expect_error(pp_pda(2), "between 0 and 1")
    expect_error(pp_pda("a"), "between 0 and 1")
  })
})

#' Pure-node stopping rule.
#'
#' Creates a stopping rule that stops splitting when a node contains only one
#' group. This is the default stopping rule for classification.
#'
#' @return A \code{stop_strategy} object.
#' @seealso \code{\link{pptr}}, \code{\link{pprf}}
#' @examples
#' stop_pure_node()
#'
#' @export
stop_pure_node <- function() {
  structure(list(name = "pure_node", display_name = "Pure node"), class = "stop_strategy")
}

#' Minimum-size stopping rule.
#'
#' Creates a stopping rule that stops splitting when a node has fewer than
#' \code{min_size} observations. Used primarily for regression trees.
#'
#' @param min_size Minimum node size to allow a split (default: 5).
#' @return A \code{stop_strategy} object.
#' @seealso \code{\link{stop_min_variance}}, \code{\link{stop_any}}
#' @examples
#' stop_min_size(5)
#'
#' @export
stop_min_size <- function(min_size = 5L) {
  min_size <- as.integer(min_size)
  if (min_size < 2L) {
    stop("`min_size` must be an integer >= 2 (a node with 0 or 1 observations is unsplittable).")
  }
  structure(
    list(name = "min_size", min_size = min_size, display_name = paste0("Min size (", min_size, ")")),
    class = "stop_strategy"
  )
}

#' Minimum-variance stopping rule.
#'
#' Creates a stopping rule that stops splitting when the within-node response
#' variance falls below \code{threshold}. Used primarily for regression trees;
#' requires a continuous response.
#'
#' @param threshold Variance threshold below which to stop splitting (default: 0.01).
#' @return A \code{stop_strategy} object.
#' @seealso \code{\link{stop_min_size}}, \code{\link{stop_any}}
#' @examples
#' stop_min_variance(0.01)
#'
#' @export
stop_min_variance <- function(threshold = 0.01) {
  threshold <- as.numeric(threshold)
  if (threshold < 0) {
    stop("`threshold` must be non-negative.")
  }
  # Matches `MinVariance::display_name` on the C++ side, which uses
  # `defaultfloat` with precision 6 so small thresholds like 1e-6 render
  # as "1e-06" instead of rounding to "0.0000".
  structure(
    list(name = "min_variance", threshold = threshold, display_name = sprintf("Min variance (%s)", formatC(threshold, format = "g", digits = 6))),
    class = "stop_strategy"
  )
}

#' Maximum-depth stopping rule.
#'
#' Creates a stopping rule that stops splitting when a node's depth reaches
#' \code{max_depth}. Depth is zero-based at the root, so \code{max_depth(k)}
#' allows at most \code{k + 1} levels. Mode-agnostic: useful for bounding
#' tree complexity in both classification and regression trees.
#'
#' @param max_depth Maximum depth (non-negative integer; 0 produces a
#'   root-only stump).
#' @return A \code{stop_strategy} object.
#' @seealso \code{\link{stop_min_size}}, \code{\link{stop_any}}
#' @examples
#' stop_max_depth(5)
#'
#' @export
stop_max_depth <- function(max_depth) {
  max_depth <- as.integer(max_depth)
  if (is.na(max_depth) || max_depth < 0L) {
    stop("`max_depth` must be a non-negative integer.")
  }
  structure(
    list(name = "max_depth", max_depth = max_depth, display_name = paste0("Max depth (", max_depth, ")")),
    class = "stop_strategy"
  )
}

#' Composite stopping rule (logical OR).
#'
#' Creates a composite stopping rule that fires when any of the child rules
#' fires. Useful for combining multiple criteria, e.g. \code{stop_any(stop_min_size(5),
#' stop_min_variance(0.01))} for regression.
#'
#' @param ... Two or more \code{stop_strategy} objects to combine.
#' @return A \code{stop_strategy} object.
#' @seealso \code{\link{stop_min_size}}, \code{\link{stop_min_variance}}, \code{\link{stop_pure_node}}
#' @examples
#' stop_any(stop_min_size(5), stop_min_variance(0.01))
#'
#' @export
stop_any <- function(...) {
  rules <- list(...)
  if (length(rules) < 1L) {
    stop("`stop_any` requires at least one stop rule.")
  }
  for (r in rules) {
    if (!inherits(r, "stop_strategy")) {
      stop("All arguments to `stop_any` must be stop_strategy objects.")
    }
  }
  names <- vapply(rules, function(r) r$display_name %||% r$name, character(1))
  structure(
    list(name = "any", rules = rules, display_name = paste0("Any(", paste(names, collapse = ", "), ")")),
    class = "stop_strategy"
  )
}

describe("stop_pure_node", {
  it("creates a stop_strategy object", {
    s <- stop_pure_node()
    expect_s3_class(s, "stop_strategy")
    expect_equal(s$name, "pure_node")
  })
})

describe("stop_min_size", {
  it("creates a stop_strategy object", {
    s <- stop_min_size(5L)
    expect_s3_class(s, "stop_strategy")
    expect_equal(s$name, "min_size")
    expect_equal(s$min_size, 5L)
  })

  it("rejects min_size < 2", {
    expect_error(stop_min_size(0L), ">= 2")
    expect_error(stop_min_size(1L), ">= 2")
  })
})

describe("stop_min_variance", {
  it("creates a stop_strategy object", {
    s <- stop_min_variance(0.01)
    expect_s3_class(s, "stop_strategy")
    expect_equal(s$name, "min_variance")
    expect_equal(s$threshold, 0.01)
  })

  it("rejects negative thresholds", {
    expect_error(stop_min_variance(-1), "non-negative")
  })
})

describe("stop_any", {
  it("combines rules", {
    s <- stop_any(stop_min_size(5L), stop_min_variance(0.01))
    expect_s3_class(s, "stop_strategy")
    expect_equal(s$name, "any")
    expect_length(s$rules, 2L)
  })

  it("rejects empty call", {
    expect_error(stop_any(), "at least one stop rule")
  })

  it("rejects non-stop_strategy arguments", {
    expect_error(stop_any("not a rule"), "stop_strategy")
  })
})

#' @useDynLib ppforest2
#' @importFrom Rcpp evalCpp
#' @importFrom stats model.frame model.matrix model.response formula predict sd terms update fitted residuals nobs
NULL

#' Trains a Random Forest of Projection-Pursuit oblique decision trees.
#'
#' This function trains a Random Forest of Projection-Pursuit oblique decision tree using either a formula and data frame interface or a matrix-based interface. When using the formula interface, specify the model formula and the data frame containing the variables. For the matrix-based interface, provide matrices for the features and labels directly.
#' The number of trees is controlled by the \code{size} parameter. Each tree is trained on a stratified bootstrap sample drawn from the data.
#' The number of variables to consider at each split is controlled by the \code{n_vars} parameter.
#' If \code{lambda = 0}, the model is trained using Linear Discriminant Analysis (LDA). If \code{lambda > 0}, the model is trained using Penalized Discriminant Analysis (PDA).
#'
#' Mode is taken from the \code{mode} argument when explicit, and otherwise auto-detected from `y` (factor/character → classification, numeric → regression). Pass \code{mode = "classification"} to force classification on integer labels (e.g. binary 0/1), or \code{mode = "regression"} to assert intent on numeric responses.
#'
#' OOB error, OOB predictions, permuted variable importance, and weighted variable importance are computed lazily on first access via the accessor functions (`oob_error()`, `oob_predictions()`, `permuted_importance()`, `weighted_importance()`). Training itself is fast because these OOB-based computations are deferred.
#'
#' @param formula A formula of the form \code{y ~ x1 + x2 + ...}, where \code{y} is a vector of labels and \code{x1}, \code{x2}, ... are the features.
#' @param data A data frame containing the variables in the formula.
#' @param x A matrix containing the features for each observation.
#' @param y A matrix containing the labels for each observation.
#' @param mode Training mode: either \code{"classification"} or \code{"regression"}. When \code{NULL} (default), mode is auto-detected from \code{y}'s type — factor or character vectors trigger classification, numeric vectors trigger regression. Setting it explicitly is useful for the binary-integer-labels case (\code{mode = "classification"} with integer 0/1 labels) and for failing fast on a type mismatch (\code{mode = "regression"} with a factor \code{y} errors immediately).
#' @param size The number of trees in the forest (default: 100).
#' @param lambda A regularization parameter (default: 0.5). If \code{lambda = 0}, the model is trained using Linear Discriminant Analysis (LDA). If \code{lambda > 0}, the model is trained using Penalized Discriminant Analysis (PDA). The default uses PDA because pure LDA (\code{lambda = 0}) is ill-conditioned when there are more variables than effective observations (see the "Known limitations" section of the README). Cannot be used together with \code{pp}.
#' @param n_vars The number of variables to consider at each split (integer). These are chosen uniformly in each split. By default, half of the variables are used (\code{p_vars = 0.5}). Cannot be used together with \code{p_vars} or \code{dr}.
#' @param p_vars The proportion of variables to consider at each split (number between 0 and 1, exclusive). For example, \code{p_vars = 0.5} uses half the features. Cannot be used together with \code{n_vars} or \code{dr}.
#' @param seed An optional integer seed for reproducibility. If \code{NULL} (default), a seed is drawn from R's RNG, so \code{set.seed()} controls reproducibility. If an integer is provided, that value is used directly. The same seed is used for training and for computing permuted variable importance.
#' @param max_retries Maximum number of retries for degenerate trees (default: 3). When a bootstrap sample yields a singular covariance matrix, the tree is retrained with a different seed up to this many times.
#' @param threads The number of threads to use. The default is the number of cores available.
#' @param pp A projection pursuit strategy object created by \code{\link{pp_pda}}. Cannot be used together with \code{lambda}.
#' @param vars A variable selection strategy object created by \code{\link{vars_uniform}} or \code{\link{vars_all}}. Cannot be used together with \code{n_vars} or \code{p_vars}.
#' @param cutpoint A split cutpoint strategy object created by \code{\link{cutpoint_mean_of_means}} (default).
#' @param stop A stopping rule object. Default depends on mode:
#'   \code{\link{stop_pure_node}()} for classification, and
#'   \code{stop_any(stop_min_size(5), stop_min_variance(0.01))} for regression.
#' @param binarize A binarization strategy object. Default depends on mode:
#'   \code{\link{binarize_largest_gap}()} for classification, and
#'   \code{\link{binarize_disabled}()} for regression (regression's default
#'   grouping always yields a 2-group partition, so no binarization is needed).
#' @param grouping A grouping strategy object. Default depends on mode:
#'   \code{\link{grouping_by_label}()} for classification, and
#'   \code{\link{grouping_by_cutpoint}()} for regression.
#' @param leaf A leaf strategy object. Default depends on mode:
#'   \code{\link{leaf_majority_vote}()} for classification, and
#'   \code{\link{leaf_mean_response}()} for regression.
#' @return A \code{pprf} model. Its S3 class vector is
#'   \code{c("pprf_classification", "pprf", "ppmodel")} or
#'   \code{c("pprf_regression", "pprf", "ppmodel")} depending on the mode.
#' @seealso \code{\link{predict.pprf_classification}}, \code{\link{predict.pprf_regression}}, \code{\link{formula.ppmodel}}, \code{\link{oob_error}}, \code{\link{save_json}}, \code{\link{load_json}}, \code{\link{pp_rand_forest}} for parsnip integration, \code{vignette("introduction")} for a tutorial
#' @examples
#'
#' # Example 1: formula interface with the `iris` dataset
#' pprf(Species ~ ., data = iris)
#'
#' # Example 2: formula interface with the `iris` dataset with regularization
#' pprf(Species ~ ., data = iris, lambda = 0.5)
#'
#' # Example 3: matrix interface with the `iris` dataset
#' pprf(x = iris[, 1:4], y = iris[, 5])
#'
#' # Example 4: matrix interface with the `iris` dataset with regularization
#' pprf(x = iris[, 1:4], y = iris[, 5], lambda = 0.5)
#'
#' # Example 5: formula interface with the `crabs` dataset
#' pprf(Type ~ ., data = crabs)
#'
#' # Example 6: formula interface with the `crabs` dataset with regularization
#' pprf(Type ~ ., data = crabs, lambda = 0.5)
#'
#' @export
pprf <- function(
    formula = NULL,
    data = NULL,
    x = NULL,
    y = NULL,
    mode = NULL,
    size = 100,
    lambda = 0.5,
    n_vars = NULL,
    p_vars = NULL,
    seed = NULL,
    max_retries = 3L,
    threads = NULL,
    pp = NULL,
    vars = NULL,
    cutpoint = NULL,
    stop = NULL,
    binarize = NULL,
    grouping = NULL,
    leaf = NULL) {
  # Capture the call before evaluating anything else so `update()` can
  # rebuild and re-evaluate it with the user's changes substituted in
  # (this is the standard `stats::update.default` contract).
  cl <- match.call()

  if (!is.null(seed) && (!is.numeric(seed) || length(seed) != 1 || seed != as.integer(seed)))
    stop("`seed` must be a single integer or NULL.")

  if (!is.numeric(size) || length(size) != 1 || size < 1 || size != as.integer(size))
    stop("`size` must be a positive integer.")

  if (!is.numeric(max_retries) || length(max_retries) != 1 || max_retries < 0 || max_retries != as.integer(max_retries))
    stop("`max_retries` must be a non-negative integer.")

  if (!is.null(threads) && (!is.numeric(threads) || length(threads) != 1 || threads < 1 || threads != as.integer(threads)))
    stop("`threads` must be a positive integer or NULL.")

  args <- resolve_model_data(formula, data, x, y, mode = mode)

  x <- args$x
  y <- args$y
  groups <- args$groups
  formula <- args$formula
  mode <- args$mode

  strategies <- resolve_strategies(
    pp = pp, lambda = lambda, lambda_missing = missing(lambda),
    vars = vars, n_vars = n_vars, n_vars_missing = missing(n_vars),
    p_vars = p_vars, p_vars_missing = missing(p_vars),
    cutpoint = cutpoint, stop = stop, binarize = binarize, grouping = grouping,
    leaf = leaf, default_vars = vars_uniform(p_vars = 0.5),
    n_features = ncol(x))

  if (is.null(seed)) {
    seed <- sample.int(.Machine$integer.max, 1L)
  }

  # Resolve the thread count. `threads = NULL` means "auto": use all available
  # cores for normal use, but cap at 2 when running under R CMD check (CRAN sets
  # `_R_CHECK_LIMIT_CORES_`) so examples and vignettes never use more than two
  # cores simultaneously, as required by CRAN policy. 0L = auto in the C++ core.
  under_check <- nzchar(chk <- Sys.getenv("_R_CHECK_LIMIT_CORES_")) && tolower(chk) != "false"
  resolved_threads <- if (!is.null(threads)) {
    as.integer(threads)
  } else if (under_check) {
    2L
  } else {
    0L
  }

  if (resolved_threads != 1L && !ppforest2_has_openmp()) {
    warning("OpenMP is not available. The forest will be trained using a single thread.\n",
            "On macOS, install libomp: brew install libomp", call. = FALSE)
  }

  training_spec <- list(
    pp = strategies$pp,
    vars = strategies$vars,
    cutpoint = strategies$cutpoint,
    stop = strategies$stop,
    binarize = strategies$binarize,
    grouping = strategies$grouping,
    leaf = strategies$leaf,
    mode = mode,
    size = as.integer(size),
    seed = as.integer(seed),
    threads = resolved_threads,
    max_retries = as.integer(max_retries))

  # `ppforest2_train` is mode-aware on the C++ side: it dispatches on
  # `training_spec$mode` and applies the appropriate index decode + sort.
  model <- ppforest2_train(training_spec, x, y)

  if (isTRUE(model$degenerate)) {
    warning("Some splits could not separate groups (degenerate nodes). ",
            "This can be caused by ill-conditioned variables in the input data, ",
            "or by bootstrap samples that produce singular covariance matrices. ",
            "Consider reviewing your data or adjusting `max_retries` (currently ", max_retries, "). ",
            "Degenerate nodes predict the group with the most observations. ",
            "Degenerate trees are excluded from variable importance calculations.",
            call. = FALSE)
  }

  # Regression models have no groups; this loop only applies to classification.
  if (!identical(mode, "regression")) {
    for (i in 1:size) {
      model$trees[[i]]$groups <- groups
    }
  }

  model$call    <- cl
  model$seed    <- seed
  model$groups  <- groups
  model$formula <- formula
  model$mode    <- mode
  model$x       <- x
  model$y       <- y

  # Cheap, always-available VI fields. Expensive OOB-based importances
  # (`permuted_importance()`, `weighted_importance()`) compute lazily.
  scale <- apply(x, 2, sd)
  scale[scale == 0] <- 1
  model$vi <- list(
    scale       = scale,
    projections = ppforest2_vi_projections_forest(model, ncol(x), scale)
  )

  # Lazy-compute cache for OOB metrics and permuted/weighted importance.
  model$.cache <- .new_cache()

  # Class is set by the Rcpp wrap layer (see `make_model_class` in
  # bindings/R/inst/include/ppforest2.h), which derives it from
  # `forest.training_spec->mode`. Don't reassign here — that would let the
  # R-side guess drift away from the C++ truth.

  model
}


# ---------------------------------------------------------------------------
# Prediction: split per mode.
# ---------------------------------------------------------------------------

#' Predicts labels or vote proportions from a pprf model (classification mode).
#'
#' @param object A \code{pprf_classification} model.
#' @param new_data A data frame or matrix of new observations. If \code{NULL}, the first positional argument in \code{...} is used for backward compatibility.
#' @param type The type of prediction: \code{"class"} (default) returns a factor of predicted labels, \code{"prob"} returns a data frame of vote proportions.
#' @param ... For backward compatibility, the first positional argument is treated as \code{new_data} when \code{new_data} is \code{NULL}.
#' @return If \code{type = "class"}, a factor of predicted labels. If \code{type = "prob"}, a data frame with one column per group, each row summing to 1.
#' @seealso \code{\link{pprf}}, \code{\link{predict.pprf_regression}}
#' @examples
#' model <- pprf(Species ~ ., data = iris)
#' predict(model, iris)
#' predict(model, iris, type = "prob")
#' @export
predict.pprf_classification <- function(object, new_data = NULL, type = NULL, ...) {
  x <- process_predict_arguments(object, new_data, ...)
  if (is.null(type)) type <- "class"

  if (type == "prob") {
    probs <- ppforest2_predict_forest_prob(object, x)
    df <- as.data.frame(probs)
    colnames(df) <- object$groups
    return(df)
  }

  if (type != "class") {
    stop("`type = \"", type, "\"` is not supported for classification models. ",
         "Use \"class\" (default) or \"prob\".", call. = FALSE)
  }

  y <- ppforest2_predict_forest(object, x)
  as.factor(object$groups[y])
}

#' Predicts numeric responses from a pprf model (regression mode).
#'
#' @param object A \code{pprf_regression} model.
#' @param new_data A data frame or matrix of new observations.
#' @param type Must be \code{"response"} (default).
#' @param ... For backward compatibility, the first positional argument is treated as \code{new_data} when \code{new_data} is \code{NULL}.
#' @return A numeric vector of mean predictions across the forest's trees.
#' @seealso \code{\link{pprf}}, \code{\link{predict.pprf_classification}}
#' @export
predict.pprf_regression <- function(object, new_data = NULL, type = NULL, ...) {
  x <- process_predict_arguments(object, new_data, ...)
  if (is.null(type)) type <- "response"

  if (type %in% c("class", "prob")) {
    stop("`type = \"", type, "\"` is not available for regression models. ",
         "Use `type = \"response\"`.", call. = FALSE)
  }

  if (type != "response") {
    stop("`type = \"", type, "\"` is not recognised. Use \"response\".", call. = FALSE)
  }

  as.numeric(ppforest2_predict_forest(object, x))
}


# ---------------------------------------------------------------------------
# print.pprf -- minimal, mode-agnostic. See `summary()` for the full breakdown.
# ---------------------------------------------------------------------------

#' Prints a compact summary of a pprf forest.
#' @param x A \code{pprf} model.
#' @param ... Unused.
#' @return Invisibly returns the input \code{pprf} model \code{x} (unchanged).
#'   Called for its side effect of printing a compact description of the forest
#'   (call, number of trees, mode, group names, formula) to the console.
#' @seealso \code{\link{summary.pprf}}
#' @export
print.pprf <- function(x, ...) {
  cat("\n")
  cat("Random Forest of Projection-Pursuit Oblique Decision Trees\n")
  if (!is.null(x$call)) {
    cat("  Call:        ", paste(deparse(x$call, width.cutoff = 80L), collapse = "\n               "), "\n", sep = "")
  }
  cat("  Trees:       ", length(x$trees), "\n", sep = "")
  cat("  Mode:        ", x$mode, "\n", sep = "")
  if (!is.null(x$groups) && length(x$groups) > 0L) {
    cat("  Group names: ", paste(x$groups, collapse = ", "), "\n", sep = "")
  }
  if (!is.null(x$formula)) {
    cat("  Formula:     ", deparse(x$formula), "\n", sep = "")
  }
  cat("\n")
  invisible(x)
}


# ---------------------------------------------------------------------------
# summary -- layered via NextMethod:
#   summary.pprf_classification / summary.pprf_regression
#     -> summary.pprf (forest-level header, VI table)
#       -> summary.ppmodel (data summary, config, formula)
# ---------------------------------------------------------------------------

#' @export
summary.ppmodel <- function(object, ...) {
  model <- object
  cat("\n")
  cat("Data Summary:\n")
  cat("  observations:", nrow(model$x), "\n")
  cat("  features:    ", ncol(model$x), "\n")
  if (identical(model$mode, "classification")) {
    cat("  groups:      ", length(model$groups), "\n")
    cat("  group names: ", paste(model$groups, collapse = ", "), "\n")
  }
  if (!is.null(model$formula)) {
    cat("  formula:     ", deparse(model$formula), "\n")
  }
  cat("\n")
  invisible(model)
}

#' Summary of a pprf forest (shared header + VI).
#'
#' @param object A \code{pprf} model.
#' @param ... Unused.
#' @return Invisibly returns the input \code{pprf} model \code{object}
#'   (unchanged). Called for its side effect of printing a detailed summary --
#'   the training specification, data summary, and variable-importance table
#'   (plus, for classification, the training/OOB confusion matrices) -- to the
#'   console.
#' @export
summary.pprf <- function(object, ...) {
  model <- object
  if (is.null(model$x)) {
    cat("\n(Empty pprf model -- no training data available.)\n")
    return(invisible(model))
  }

  cat("\n")
  cat(if (identical(model$mode, "regression")) {
    "Random Forest of Projection-Pursuit Oblique Regression Trees\n"
  } else {
    "Random Forest of Projection-Pursuit Oblique Decision Trees\n"
  })
  cat("\n")
  cat("Size:", length(model$trees), "trees\n")
  print_training_spec(model$training_spec)

  NextMethod()  # summary.ppmodel -- data summary block

  invisible(model)
}

#' @export
summary.pprf_classification <- function(object, ...) {
  # Let the shared scaffolding (pprf -> ppmodel) run first.
  NextMethod()
  model <- object

  cat("Training Confusion Matrix:\n\n")
  print_confusion_matrix(ppforest2_predict_forest(model, model$x), model)
  cat("\n")
  cat("OOB Confusion Matrix:\n\n")
  print_oob_confusion_matrix(model)
  cat("\n")

  .print_vi_table(model, include_oob_importances = TRUE)
  invisible(model)
}

#' @export
summary.pprf_regression <- function(object, ...) {
  NextMethod()
  model <- object

  # Training metrics
  preds <- ppforest2_predict_forest(model, model$x)
  y <- model$y
  mse <- mean((preds - y)^2)
  mae <- mean(abs(preds - y))
  ss_tot <- sum((y - mean(y))^2)
  r2 <- if (ss_tot > 0) 1 - sum((preds - y)^2) / ss_tot else 0
  cat("Training Metrics:\n")
  cat("  MSE:", format(mse, nsmall = 6), "\n")
  cat("  MAE:", format(mae, nsmall = 6), "\n")
  cat("  R\u00b2: ", format(r2, nsmall = 6), "\n\n")

  oob <- oob_error(model)
  if (!is.na(oob)) {
    cat("OOB MSE:", format(oob, nsmall = 6), "\n\n")
  } else {
    cat("OOB MSE: not available (no observation has any out-of-bag tree)\n\n")
  }

  .print_vi_table(model, include_oob_importances = TRUE)
  invisible(model)
}


# ---------------------------------------------------------------------------
# Helpers -- these are for `summary()`, not part of the public API.
# ---------------------------------------------------------------------------

# Print a VI table. For forests, `include_oob_importances` pulls the lazy
# `weighted` and `permuted` importances. For trees, these are absent.
.print_vi_table <- function(model, include_oob_importances) {
  cat("Variable Importance:\n\n")

  projections <- projection_importance(model)
  p <- length(projections)
  vnames <- if (!is.null(colnames(model$x))) colnames(model$x) else paste0("x", seq_len(p))
  ord <- order(projections, decreasing = TRUE)

  cols <- list(
    Variable   = vnames[ord],
    sigma      = model$vi$scale[ord],
    Projection = projections[ord]
  )

  if (include_oob_importances) {
    cols$Weighted <- weighted_importance(model)[ord]
    cols$Permuted <- permuted_importance(model)[ord]
  }

  tbl <- do.call(data.frame, c(cols, list(row.names = seq_len(p))))
  names(tbl)[2] <- "\u03c3"
  print(tbl)

  if (!all(model$vi$scale == 1)) {
    cat("\nNote: Variable importance was calculated using scaled coefficients (|a_j| * \u03c3_j).\n")
    cat("Variable contributions can only be theoretically interpreted as such\n")
    cat("if the model was trained on scaled data. Scaling also changes the\n")
    cat("projection-pursuit optimization, which may affect the resulting tree.\n")
  }
}

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

describe("predict.pprf", {
  describe("on an object created with the formula interface", {
    it("returns a factor with the same length as the input matrix", {
      model <- pprf(Species ~ ., data = iris, threads = 1)
      predictions <- predict(model, iris)
      expect_equal(length(predictions), nrow(iris))
    })

    it("returns a factor with the same levels as the groups in the model", {
      model <- pprf(Species ~ ., data = iris, threads = 1)
      predictions <- predict(model, iris)
      expect_equal(levels(predictions), levels(iris$Species))
    })

    it("with new_data parameter returns the same result as positional", {
      model <- pprf(Species ~ ., data = iris, threads = 1)
      pred_positional <- predict(model, iris)
      pred_named <- predict(model, new_data = iris)
      expect_equal(pred_positional, pred_named)
    })
  })

  describe("on an object created with the matrix interface", {
    it("returns a factor with the same length as the input matrix", {
      x <- crabs[, 2:5]
      x$sex <- as.numeric(as.factor(crabs$sex))
      model <- pprf(x = x, y = crabs$Type, threads = 1)
      predictions <- predict(model, x)
      expect_equal(length(predictions), nrow(x))
    })

    it("returns a factor with the same levels as the groups in the model", {
      x <- crabs[, 2:5]
      x$sex <- as.numeric(as.factor(crabs$sex))
      model <- pprf(x = x, y = crabs$Type, threads = 1)
      predictions <- predict(model, x)
      expect_equal(levels(predictions), levels(crabs$Type))
    })
  })

  describe("with type = 'prob'", {
    it("returns a data frame with one column per group", {
      model <- pprf(Species ~ ., data = iris, threads = 1)
      probs <- predict(model, iris, type = "prob")
      expect_true(is.data.frame(probs))
      expect_equal(ncol(probs), length(levels(iris$Species)))
      expect_equal(colnames(probs), levels(iris$Species))
    })

    it("returns rows that sum to 1", {
      model <- pprf(Species ~ ., data = iris, threads = 1)
      probs <- predict(model, iris, type = "prob")
      row_sums <- rowSums(probs)
      expect_equal(row_sums, rep(1.0, nrow(iris)), tolerance = 1e-6)
    })

    it("returns values between 0 and 1", {
      model <- pprf(Species ~ ., data = iris, threads = 1)
      probs <- predict(model, iris, type = "prob")
      expect_true(all(probs >= 0))
      expect_true(all(probs <= 1))
    })

    it("returns the correct number of rows", {
      model <- pprf(Species ~ ., data = iris, threads = 1)
      probs <- predict(model, iris, type = "prob")
      expect_equal(nrow(probs), nrow(iris))
    })
  })
})

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

describe("formula.pprf", {
  it("on an object created with the formula interface, returns the formula used to create the model", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expect_equal(
      formula(model),
      Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width - 1
    )
  })

  it("on an object created with the matrix interface, returns NULL", {
    model <- pprf(x = iris[, 1:4], y = iris[, 5], threads = 1)
    expect_equal(formula(model), NULL)
  })
})

#' Label-based grouping strategy.
#'
#' Creates a grouping strategy that routes all observations of a group to the
#' same child node. This is the default grouping strategy for classification.
#'
#' @return A \code{grouping_strategy} object.
#' @seealso \code{\link{pptr}}, \code{\link{pprf}}
#' @examples
#' grouping_by_label()
#'
#' @export
grouping_by_label <- function() {
  structure(list(name = "by_label", display_name = "By label"), class = "grouping_strategy")
}

#' Cutpoint-based grouping strategy (regression).
#'
#' Creates a grouping strategy for regression trees: observations are split by
#' the cutpoint in projected space, then each child's observations are sorted
#' by the continuous response and median-split into 2 new groups for the next
#' projection-pursuit step. This is the default grouping strategy for regression.
#'
#' @return A \code{grouping_strategy} object.
#' @seealso \code{\link{pptr}}, \code{\link{pprf}}, \code{\link{grouping_by_label}}
#' @examples
#' grouping_by_cutpoint()
#'
#' @export
grouping_by_cutpoint <- function() {
  structure(list(name = "by_cutpoint", display_name = "By cutpoint"), class = "grouping_strategy")
}

describe("grouping_by_label", {
  it("creates a grouping_strategy object", {
    s <- grouping_by_label()
    expect_s3_class(s, "grouping_strategy")
    expect_equal(s$name, "by_label")
  })
})

describe("grouping_by_cutpoint", {
  it("creates a grouping_strategy object", {
    s <- grouping_by_cutpoint()
    expect_s3_class(s, "grouping_strategy")
    expect_equal(s$name, "by_cutpoint")
  })
})

# ======================================================================
# S3 plot methods
# ======================================================================

#' Plot a pptr model.
#'
#' Visualizes a pptr model. By default, shows a mosaic overview with tree
#' structure, decision boundaries, and variable importance. Use \code{type}
#' to show individual plots.
#'
#' @param x A pptr model.
#' @param type Character string specifying the plot type. \code{NULL} (default)
#'   shows a mosaic overview. Other options: \code{"structure"} for tree with
#'   embedded histograms, \code{"importance"} for variable importance,
#'   \code{"projection"} for projected data at a node,
#'   \code{"boundaries"} for decision boundaries in feature space.
#' @param metric Character string selecting a single importance metric
#'   to plot: \code{"projections"}, \code{"weighted"}, or
#'   \code{"permuted"} (availability depends on the model).  \code{NULL}
#'   (default) shows all available metrics together.  Only used when
#'   \code{type = "importance"}.
#' @param node Integer index of the node for projection plots (1-based, breadth-first
#'   order). Defaults to 1 (root node). Only used when \code{type = "projection"}.
#' @param ... Additional arguments passed to the internal plotting function.
#' @return A ggplot2-compatible object (invisibly). The mosaic layout returns
#'   a patchwork object that works with \code{ggplot2::ggsave()}.
#' @examples
#' \donttest{
#' if (requireNamespace("ggplot2", quietly = TRUE) &&
#'     requireNamespace("patchwork", quietly = TRUE)) {
#'   model <- pptr(Species ~ ., data = iris)
#'   plot(model)                         # mosaic overview
#'   plot(model, type = "structure")     # tree structure only
#'   plot(model, type = "importance")    # variable importance
#'   plot(model, type = "projection")    # projection histogram
#'   plot(model, type = "boundaries")    # decision boundaries
#' }
#' }
#' @export
plot.pptr <- function(x, type = NULL, metric = NULL, node = 1L, ...) {
  check_ggplot2()

  if (is.null(type)) {
    p <- plot_mosaic(x, ...)
    print(p)
    return(invisible(p))
  }

  type <- match.arg(type, c("structure", "importance", "projection", "boundaries"))

  p <- switch(type,
    structure  = plot_tree_structure(x, ...),
    importance = plot_importance(x, metric = metric, ...),
    projection = plot_projection(x, node = node, ...),
    boundaries = plot_boundaries(x, ...)
  )

  if (inherits(p, "gtable")) {
    grid::grid.newpage()
    grid::grid.draw(p)
  } else {
    print(p)
  }
  invisible(p)
}

#' Plot a pprf model.
#'
#' Visualizes a pprf model. By default, shows variable importance with
#' one plot per metric side by side. Use \code{metric} to show a single
#' importance metric.
#'
#' @param x A pprf model.
#' @param type Character string specifying the plot type.
#'   \code{"importance"} (default) shows variable importance,
#'   \code{"structure"} shows a specific tree with embedded histograms,
#'   \code{"projection"} shows projected data at a node,
#'   \code{"boundaries"} shows decision boundaries of a specific tree.
#' @param metric Character string selecting a single importance metric
#'   to plot: \code{"projections"}, \code{"weighted"}, or
#'   \code{"permuted"}.  \code{NULL} (default) shows all available
#'   metrics side by side in separate panels.  Only used when
#'   \code{type = "importance"}.
#' @param tree_index Integer index of the tree to plot (1-based). Only used when
#'   \code{type = "structure"}, \code{type = "projection"}, or
#'   \code{type = "boundaries"}. Defaults to 1.
#' @param node Integer index of the node for projection plots. Defaults to 1 (root).
#'   Only used when \code{type = "projection"}.
#' @param ... Additional arguments passed to the internal plotting function.
#' @return A ggplot2-compatible object (invisibly). The importance grid returns
#'   a patchwork object that works with \code{ggplot2::ggsave()}.
#' @examples
#' \donttest{
#' if (requireNamespace("ggplot2", quietly = TRUE) &&
#'     requireNamespace("patchwork", quietly = TRUE)) {
#'   forest <- pprf(Species ~ ., data = iris, size = 10)
#'   plot(forest)                                    # all metrics side by side
#'   plot(forest, metric = "permuted")               # single metric
#'   plot(forest, type = "structure", tree_index = 1)
#'   plot(forest, type = "projection", tree_index = 1)
#' }
#' }
#' @export
plot.pprf <- function(x, type = "importance", metric = NULL,
                          tree_index = 1L, node = 1L, ...) {
  check_ggplot2()
  type <- match.arg(type, c("importance", "structure", "projection", "boundaries"))

  if (type %in% c("structure", "projection", "boundaries")) {
    tree <- x$trees[[tree_index]]
    tree$x <- x$x
    tree$y <- x$y
    tree$groups <- x$groups
  }

  if (type == "importance" && is.null(metric)) {
    p <- plot_importance_grid(x, ...)
    print(p)
    return(invisible(p))
  }

  p <- switch(type,
    importance = plot_importance(x, metric = metric, ...),
    structure  = plot_tree_structure(tree, ...),
    projection = plot_projection(tree, node = node, ...),
    boundaries = plot_boundaries(tree, ...)
  )

  if (inherits(p, "gtable")) {
    grid::grid.newpage()
    grid::grid.draw(p)
  } else {
    print(p)
  }
  invisible(p)
}

#' @useDynLib ppforest2
#' @importFrom Rcpp evalCpp
NULL

# ---------------------------------------------------------------------------
# Memoization cache (environment-based; mutable across S3 method calls).
#
# Models are S3 lists (copy-on-modify), which can't memoize via field writes.
# Each model carries an environment `$.cache` used as a scratchpad by the
# accessor methods below. `load_json` populates the cache with values that
# were persisted during training-time save.
# ---------------------------------------------------------------------------

# Create a fresh cache environment.
.new_cache <- function() new.env(parent = emptyenv())

# Return the cached value for `key`, computing via `compute_fn` on first access.
# Falls back to uncached compute if `model$.cache` is missing (e.g. models
# assembled manually in tests).
.cached_or_compute <- function(model, key, compute_fn) {
  cache <- model$.cache
  if (is.null(cache)) return(compute_fn())
  if (!exists(key, envir = cache, inherits = FALSE)) {
    assign(key, compute_fn(), envir = cache)
  }
  get(key, envir = cache, inherits = FALSE)
}

# Stash a pre-computed value directly into the cache (used by load_json to
# preserve OOB metrics saved during training-time serialization).
.prime_cache <- function(model, key, value) {
  cache <- model$.cache
  if (is.null(cache)) return(invisible(NULL))
  if (!is.null(value)) assign(key, value, envir = cache)
  invisible(NULL)
}


# ---------------------------------------------------------------------------
# Public OOB accessors (generics + methods).
#
# Only forests have OOB concepts; calling these on a `pptr` model errors.
# The accessors compute lazily using the training data stored on the model
# (`$x`, `$y` — `$y` is integer class labels for classification and the
# continuous response for regression) and cache the result in `$.cache`.
# ---------------------------------------------------------------------------

#' Out-of-bag error for a random forest.
#'
#' Computes (or returns the cached) OOB error using the training data stored
#' on the model. For classification, this is the misclassification rate in
#' `[0, 1]`. For regression, it is the mean squared error against the
#' continuous response.
#'
#' @param model A \code{pprf} forest model.
#' @return A numeric scalar in `[0, 1]` for classification or `[0, Inf)` for
#'   regression. Returns `NA_real_` when no observation has any out-of-bag
#'   tree (e.g. a degenerate forest where every tree saw every row). Callers
#'   should check with `is.na()` rather than comparing against a sentinel
#'   value; in earlier versions this condition was signalled as `-1`, which
#'   was not distinguishable from a (mathematically impossible but
#'   representable) error rate.
#' @seealso \code{\link{oob_predictions}}, \code{\link{oob_samples}}
#' @export
oob_error <- function(model) UseMethod("oob_error")

#' @export
oob_error.default <- function(model) {
  stop(
    "`oob_error()` is only defined for `pprf` forest models, not objects of class '",
    paste(class(model), collapse = "/"), "'.",
    call. = FALSE
  )
}

# Rcpp returns a length-1 `NumericVector` from the C++ OOB-error bindings.
# That already behaves as a length-1 R numeric for every operation callers
# care about (comparisons, `is.na()`, arithmetic, assignment), so no
# `as.numeric()` coercion is needed on either branch. The length-1 vector
# carries either the error value or `NA_real_` (never the legacy `-1`
# sentinel — the C++ side translates `std::nullopt` at the boundary).

#' @export
oob_error.pprf <- function(model) {
  # Mode dispatch lives in `ppforest2_oob_error` on the C++ side: it
  # decodes `y` from R 1-based factor codes only for classification,
  # and computes either misclassification rate or MSE based on the
  # forest's `training_spec$mode`. Both subclasses can share this
  # method as a result — no per-mode wrapper needed at the R level.
  .cached_or_compute(model, "oob_error", function() {
    .require_training_data(model, c("x", "y"))
    ppforest2_oob_error(model, model$x, model$y)
  })
}


#' Out-of-bag predictions for a random forest.
#'
#' Returns predictions for each training row using only trees that did
#' not see that row in their bootstrap sample. Observations with no OOB
#' tree are represented as `NA` in both modes: `NA` at the factor level
#' for classification, and `NA_real_` for regression. Filter with the
#' standard `is.na()` idiom.
#'
#' @param model A \code{pprf} forest model.
#' @return A factor (classification) or numeric vector (regression), length `n`.
#' @seealso \code{\link{oob_error}}, \code{\link{oob_samples}}
#' @export
oob_predictions <- function(model) UseMethod("oob_predictions")

#' @export
oob_predictions.default <- function(model) {
  stop(
    "`oob_predictions()` is only defined for `pprf` forest models, not objects of class '",
    paste(class(model), collapse = "/"), "'.",
    call. = FALSE
  )
}

#' @export
oob_predictions.pprf_classification <- function(model) {
  .cached_or_compute(model, "oob_predictions", function() {
    .require_training_data(model, "x")
    # Sentinel: C++ `oob_predict` emits `NaN` for rows that had no OOB
    # tree, in both modes. The Rcpp wrapper's `to_r_indices` shift adds
    # +1 to every element to convert 0-based C++ group ids to 1-based
    # R factor indices — `NaN + 1` is still `NaN`, so the sentinel
    # survives intact. We then remap `NaN → NA_real_` explicitly (the
    # same step the regression accessor does) so the downstream
    # `as.integer` coercion doesn't emit "NAs introduced by coercion"
    # warnings, and the factor records each no-OOB row as a missing
    # level cleanly.
    raw <- ppforest2_oob_predict(model, model$x)
    raw[is.nan(raw)] <- NA_real_
    factor(model$groups[as.integer(raw)], levels = model$groups)
  })
}

#' @export
oob_predictions.pprf_regression <- function(model) {
  .cached_or_compute(model, "oob_predictions", function() {
    .require_training_data(model, "x")
    # C++ emits NaN for rows with no OOB tree (the natural float sentinel,
    # used identically in classification — see `oob_predictions.pprf_classification`).
    # Remap to NA_real_ here so callers can use the plain `is.na()` idiom
    # without having to remember the underlying NaN.
    out <- as.numeric(ppforest2_oob_predict(model, model$x))
    out[is.nan(out)] <- NA_real_
    out
  })
}


#' Out-of-bag row indices per tree.
#'
#' Returns a list where element `i` is the integer vector of row indices
#' (1-based) that were **not** in the bootstrap sample of tree `i`.
#'
#' @param model A \code{pprf} forest model.
#' @return A list of integer vectors, one per tree.
#' @seealso \code{\link{bag_samples}}, \code{\link{oob_error}}
#' @export
oob_samples <- function(model) UseMethod("oob_samples")

#' @export
oob_samples.default <- function(model) {
  stop(
    "`oob_samples()` is only defined for `pprf` forest models, not objects of class '",
    paste(class(model), collapse = "/"), "'.",
    call. = FALSE
  )
}

#' @export
oob_samples.pprf <- function(model) {
  .require_training_data(model, "x")
  n <- nrow(model$x)
  lapply(model$trees, function(t) setdiff(seq_len(n), t$sample_indices + 1L))
}


#' In-bag row indices per tree.
#'
#' Returns a list where element `i` is the integer vector of row indices
#' (1-based, with replacement) drawn into the bootstrap sample of tree `i`.
#'
#' @param model A \code{pprf} forest model.
#' @return A list of integer vectors, one per tree.
#' @seealso \code{\link{oob_samples}}
#' @export
bag_samples <- function(model) UseMethod("bag_samples")

#' @export
bag_samples.default <- function(model) {
  stop(
    "`bag_samples()` is only defined for `pprf` forest models, not objects of class '",
    paste(class(model), collapse = "/"), "'.",
    call. = FALSE
  )
}

#' @export
bag_samples.pprf <- function(model) {
  lapply(model$trees, function(t) t$sample_indices + 1L)
}


# ---------------------------------------------------------------------------
# Variable-importance accessors (lazy for OOB-requiring variants).
# ---------------------------------------------------------------------------

#' Permuted variable importance for a random forest.
#'
#' For each feature, measures the drop in OOB accuracy (classification) or
#' the increase in normalised MSE (regression) after randomly permuting
#' that feature across the OOB observations. Computed lazily from the
#' training data stored on the model; the result is cached.
#'
#' **Sign semantics.** Entries may be **negative**. That is not an error
#' and not a sentinel: it means permuting the feature did not degrade OOB
#' fit on average — the feature's signal sits at or below the noise floor
#' of the permutation procedure. Interpret negative or near-zero entries
#' as "no evidence of importance"; rely on the ranking rather than
#' clipping at zero or normalizing. The scale is already comparable
#' within a fitted model.
#'
#' @param model A \code{pprf} forest model.
#' @return A numeric vector, one entry per feature. Negative values are
#'   meaningful (see Sign semantics above).
#' @export
permuted_importance <- function(model) UseMethod("permuted_importance")

#' @export
permuted_importance.default <- function(model) {
  stop(
    "`permuted_importance()` is only defined for `pprf` forest models.",
    call. = FALSE
  )
}

#' @export
permuted_importance.pprf <- function(model) {
  # `model$y` is mode-correct for both classification (1-based factor codes,
  # which the C++ binding shifts to 0-based when needed) and regression
  # (the continuous response). Previously the regression branch passed a
  # 0/1 median-split indicator here, making MSE/NMSE meaningless — fixed
  # by unifying `model$y` with the continuous response in `validate_data()`.
  .cached_or_compute(model, "permuted_importance", function() {
    .require_training_data(model, c("x", "y"))
    ppforest2_vi_permuted_forest(model, model$x, model$y, model$seed)
  })
}


#' Weighted projection variable importance for a random forest.
#'
#' Weights each tree's projection-based importance by a per-tree OOB
#' quality score — `1 - error_rate` in `[0, 1]` for classification, and
#' `max(0, 1 - NMSE)` in `[0, 1]` for regression — then aggregates
#' `I_s × |a_j|` over splits. Computed lazily from the training data
#' stored on the model; the result is cached.
#'
#' **Sign semantics.** Entries are non-negative by construction (weights
#' and per-split contributions are both non-negative). A zero entry means
#' "this feature never appeared in a weighted OOB-contributing split,"
#' not "within noise." Contrast with \code{\link{permuted_importance}},
#' where negative values are meaningful. Do not re-normalize — rely on
#' the ranking.
#'
#' @param model A \code{pprf} forest model.
#' @return A non-negative numeric vector, one entry per feature.
#' @export
weighted_importance <- function(model) UseMethod("weighted_importance")

#' @export
weighted_importance.default <- function(model) {
  stop(
    "`weighted_importance()` is only defined for `pprf` forest models.",
    call. = FALSE
  )
}

#' @export
weighted_importance.pprf <- function(model) {
  # See the note on `permuted_importance.pprf` — `model$y` is now
  # mode-correct for both classification and regression.
  .cached_or_compute(model, "weighted_importance", function() {
    .require_training_data(model, c("x", "y"))
    ppforest2_vi_weighted_forest(model, model$x, model$y, model$vi$scale)
  })
}


#' Projection-coefficient variable importance.
#'
#' The projection-based importance (VI2): each split's scaled absolute
#' projection coefficients (`|a_j| * sigma_j`) aggregated into a per-feature
#' score, averaged over the non-degenerate trees of a forest. Unlike
#' \code{\link{permuted_importance}} and \code{\link{weighted_importance}},
#' this measure is not OOB-based — it depends only on the fitted projector
#' geometry, is computed eagerly at fit time (cheap), and is available for
#' both single trees (\code{pptr}) and forests (\code{pprf}).
#'
#' **Sign semantics.** Entries are non-negative by construction (absolute
#' coefficients scaled by each feature's standard deviation). Rely on the
#' ranking rather than re-normalizing.
#'
#' @param model A \code{pptr} or \code{pprf} model.
#' @return A non-negative numeric vector, one entry per feature.
#' @seealso \code{\link{permuted_importance}}, \code{\link{weighted_importance}}
#' @export
projection_importance <- function(model) UseMethod("projection_importance")

#' @export
projection_importance.default <- function(model) {
  stop(
    "`projection_importance()` is only defined for ppforest2 models (`pptr` or `pprf`).",
    call. = FALSE
  )
}

#' @export
projection_importance.ppmodel <- function(model) {
  # Computed eagerly at fit time (cheap, not OOB-based) and stored on
  # `model$vi$projections`; it survives save/load (see `load_json`).
  model$vi$projections
}


# ---------------------------------------------------------------------------
# Shared S3 methods at the ppmodel level (both pptr and pprf).
# ---------------------------------------------------------------------------

#' Formula extractor for ppforest2 models.
#'
#' @param x A \code{pptr} or \code{pprf} model.
#' @param ... Unused.
#' @return The formula the model was trained with, or `NULL` for matrix-interface fits.
#' @export
formula.ppmodel <- function(x, ...) {
  x$formula
}


#' Number of observations used to fit a ppforest2 model.
#'
#' Implements the standard \code{stats::nobs()} contract so downstream tools
#' (\code{step()}, broom's \code{glance()}, information criteria) can ask for
#' the training-sample size.
#'
#' @param object A \code{pptr} or \code{pprf} model.
#' @param ... Unused.
#' @return Integer scalar. Returns \code{NA_integer_} for models loaded from
#'   JSON without their original training data.
#' @export
nobs.ppmodel <- function(object, ...) {
  if (is.null(object$x)) return(NA_integer_)
  nrow(object$x)
}


#' Fitted (in-sample) predictions from a ppforest2 model.
#'
#' Returns predictions on the training data — a factor for classification,
#' a numeric vector for regression. For forests, training predictions are
#' optimistic; use \code{\link{oob_predictions}()} for an honest estimate.
#'
#' @param object A \code{pptr} or \code{pprf} model.
#' @param ... Unused.
#' @return A factor (classification) or numeric vector (regression), length
#'   equal to the number of training observations.
#' @seealso \code{\link{residuals.ppmodel}}, \code{\link{oob_predictions}}
#' @export
fitted.ppmodel <- function(object, ...) {
  .require_training_data(object, "x")
  # Dispatches through the model's own predict method, which is mode- and
  # class-aware (pprf vs pptr × classification vs regression). For
  # classification this returns a factor; for regression a numeric vector.
  predict(object, object$x)
}


#' Residuals from a regression ppforest2 model.
#'
#' Returns \code{y - fitted(model)}. Only defined for regression models —
#' classification residuals have no canonical scalar form, so this method
#' errors on classification models rather than inventing a convention.
#'
#' @param object A \code{pprf} or \code{pptr} regression model.
#' @param ... Unused.
#' @return A numeric vector of length \code{nobs(object)}.
#' @seealso \code{\link{fitted.ppmodel}}
#' @export
residuals.ppmodel <- function(object, ...) {
  if (!identical(object$mode, "regression")) {
    stop(
      "`residuals()` is only defined for regression models. ",
      "For classification, compare `fitted(model)` against the response directly, ",
      "or use `oob_predictions(model)` for an out-of-bag misclassification view.",
      call. = FALSE
    )
  }
  .require_training_data(object, c("x", "y"))
  as.numeric(object$y) - as.numeric(fitted(object))
}


# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------

# Throws a clear error if required training-data fields are missing on the
# model (e.g. the model was loaded from JSON without the original x/y).
# Distinguishes between "loaded model without data" (JSON doesn't carry x/y,
# user should reattach) and "metrics were never computed at save time" (the
# saved JSON had `include_metrics = FALSE`, so priming the cache was a
# no-op and no recomputation is possible).
.require_training_data <- function(model, fields) {
  for (f in fields) {
    if (is.null(model[[f]])) {
      # Heuristic: if the model has a `.cache` environment and `training_spec`
      # populated, it was almost certainly loaded from JSON. Mention both
      # reattachment and the save-time-metrics case so the user can pick.
      loaded_from_json <- !is.null(model$training_spec) && !is.null(model$.cache)
      if (loaded_from_json) {
        stop(
          "Required field `", f, "` is not available on the loaded model. ",
          "Either (a) re-attach the original training data to `model$", f,
          "` before calling this accessor, or (b) re-save the source model ",
          "with `save_json(model, path, include_metrics = TRUE)` so the ",
          "relevant OOB / variable-importance values are preserved in the JSON.",
          call. = FALSE
        )
      }
      stop(
        "Required field `", f, "` is not available on the model -- ",
        "this model was probably loaded from JSON without the original training data.",
        call. = FALSE
      )
    }
  }
  invisible(NULL)
}

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

describe("pprf formula interface", {
  it("returns a object with s3 class pprf", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expect_s3_class(model, "pprf")
  })

  it("preserves the observations (x) as a matrix in the returned model", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expected <- model.matrix(
      Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width - 1,
      iris,
      response = TRUE
    )
    expect_equal(model$x, expected)
  })

  it("preserves the labels (y) as a matrix in the returned model", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expected <- as.matrix(as.numeric(as.factor(iris$Species)))
    expect_equal(model$y, expected)
  })

  it("preserves the groups in the returned model", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expected <- levels(iris$Species)
    expect_equal(model$groups, expected)
  })

  it("preserves the formula in the returned model", {
    model <- pprf(Species ~ ., data = iris, threads = 1)

    expect_equal(
      model$formula,
      Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width - 1
    )
  })
})

describe("pprf matrix interface", {
  it("returns a object with s3 class pprf", {
    model <- pprf(x = iris[, 1:4], y = iris[, 5], threads = 1)
    expect_s3_class(model, "pprf")
  })

  it("preserves the observations (x) as a matrix in the returned model", {
    model <- pprf(x = iris[, 1:4], y = iris[, 5], threads = 1)
    expected <- as.matrix(iris[, 1:4])
    expect_equal(model$x, expected)
  })

  it("preserves the labels (y) as a matrix in the returned model", {
    model <- pprf(x = iris[, 1:4], y = iris[, 5], threads = 1)
    expected <- as.matrix(as.numeric(as.factor(iris$Species)))
    expect_equal(model$y, expected)
  })

  it("preserves the groups in the returned model", {
    model <- pprf(x = iris[, 1:4], y = iris[, 5], threads = 1)
    expected <- levels(iris$Species)
    expect_equal(model$groups, expected)
  })

  it("does not have a formula", {
    model <- pprf(x = iris[, 1:4], y = iris[, 5], threads = 1)
    expect_equal(model$formula, NULL)
  })
})

describe("pprf reproducibility", {
  it("produces identical predictions with set.seed", {
    set.seed(0)
    model1 <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
    set.seed(0)
    model2 <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
    expect_equal(predict(model1, iris), predict(model2, iris))
  })

  it("produces identical predictions with explicit seed", {
    model1 <- pprf(Species ~ ., data = iris, size = 3, seed = 123L, threads = 1)
    model2 <- pprf(Species ~ ., data = iris, size = 3, seed = 123L, threads = 1)
    expect_equal(predict(model1, iris), predict(model2, iris))
  })

  it("produces reproducible VI permuted importance", {
    model1 <- pprf(Species ~ ., data = iris, size = 3, seed = 0, threads = 1)
    model2 <- pprf(Species ~ ., data = iris, size = 3, seed = 0, threads = 1)
    expect_equal(permuted_importance(model1), permuted_importance(model2))
  })

  it("stores the explicit seed in the model", {
    model <- pprf(Species ~ ., data = iris, size = 3, seed = 0, threads = 1)
    expect_equal(model$seed, 0)
  })

  it("stores the generated seed when seed is NULL", {
    set.seed(99)
    model <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
    expect_true(is.numeric(model$seed))
    expect_true(model$seed > 0)
  })
})

describe("pprf input validation", {
  it("rejects invalid lambda", {
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], lambda = -1), "between 0 and 1")
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], lambda = 2), "between 0 and 1")
  })

  it("rejects non-integer seed", {
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], seed = 1.5), "integer")
  })

  it("rejects size < 1", {
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], size = 0), "positive integer")
  })

  it("rejects non-integer size", {
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], size = 2.5), "positive integer")
  })

  it("rejects n_vars out of range", {
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], n_vars = 100), "number of features")
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], n_vars = 0), "positive integer")
  })

  it("rejects non-positive threads", {
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], threads = 0), "positive integer")
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], threads = -1), "positive integer")
  })

  it("rejects both n_vars and p_vars", {
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], n_vars = 2, p_vars = 0.5), "not both")
  })

  it("rejects invalid p_vars", {
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], p_vars = 0), "between 0 (exclusive) and 1 (inclusive)", fixed = TRUE)
    expect_error(pprf(x = iris[, 1:4], y = iris[, 5], p_vars = 1.5), "between 0 (exclusive) and 1 (inclusive)", fixed = TRUE)
  })

  it("accepts valid p_vars", {
    model <- pprf(x = iris[, 1:4], y = iris[, 5], p_vars = 0.5, size = 2, threads = 1)
    expect_equal(model$training_spec$vars$count, 2)
  })
})

describe("pprf training spec", {
  it("preserves the lambda parameter in the returned model", {
    model <- pprf(Species ~ ., data = iris, lambda = 0.5, threads = 1)
    expect_equal(model$training_spec$pp$lambda, 0.5)
  })

  it("the lambda parameter is 0.5 by default", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expect_equal(model$training_spec$pp$lambda, 0.5)
  })

  it("preserves the n_vars parameter in the returned model", {
    model <- pprf(Species ~ ., data = iris, n_vars = 2, threads = 1)
    expect_equal(model$training_spec$vars$count, 2)
  })

  it("uses half the variables by default", {
    model <- pprf(x = iris[, 1:4], y = iris[, 5], threads = 1)
    expect_equal(model$training_spec$vars$count, 2)  # 0.5 * 4 = 2
  })

  it("generates as many trees as indicated by the size parameter", {
    model <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
    expect_equal(length(model$trees), 3)
  })

  it("generates 100 trees when the size parameter is not given", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expect_equal(length(model$trees), 100)
  })

  it("the pp strategy is pda", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expect_equal(model$training_spec$pp$name, "pda")
  })

  it("the vars strategy is uniform", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expect_equal(model$training_spec$vars$name, "uniform")
  })

  it("the stop strategy is pure_node", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expect_equal(model$training_spec$stop$name, "pure_node")
  })

  it("the binarize strategy is largest_gap", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expect_equal(model$training_spec$binarize$name, "largest_gap")
  })

  it("the grouping strategy is by_label", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expect_equal(model$training_spec$grouping$name, "by_label")
  })

  it("the leaf strategy is majority_vote", {
    model <- pprf(Species ~ ., data = iris, threads = 1)
    expect_equal(model$training_spec$leaf$name, "majority_vote")
  })
})

describe("pprf regression training spec", {
  # Pins the mode-aware defaults selected by
  # `TrainingSpec::Builder::apply_defaults()` for regression. The C++
  # side is the single source of truth; this block guards against
  # silent drift if the default selection ever changes.
  it("the pp strategy is pda", {
    d <- make_regression_data()
    model <- pprf(x = d$x, y = d$y, seed = 0, threads = 1)
    expect_equal(model$training_spec$pp$name, "pda")
  })

  it("the vars strategy is uniform (forest default)", {
    # Unlike trees, forests subsample variables by default — same
    # convention as the classification path.
    d <- make_regression_data()
    model <- pprf(x = d$x, y = d$y, seed = 0, threads = 1)
    expect_equal(model$training_spec$vars$name, "uniform")
  })

  it("the cutpoint strategy is mean_of_means", {
    d <- make_regression_data()
    model <- pprf(x = d$x, y = d$y, seed = 0, threads = 1)
    expect_equal(model$training_spec$cutpoint$name, "mean_of_means")
  })

  it("the stop strategy is any(min_size(5), min_variance(0.01))", {
    d <- make_regression_data()
    model <- pprf(x = d$x, y = d$y, seed = 0, threads = 1)
    stop_spec <- model$training_spec$stop
    expect_equal(stop_spec$name, "any")
    expect_length(stop_spec$rules, 2)
    expect_equal(stop_spec$rules[[1]]$name, "min_size")
    expect_equal(stop_spec$rules[[1]]$min_size, 5)
    expect_equal(stop_spec$rules[[2]]$name, "min_variance")
    expect_equal(stop_spec$rules[[2]]$threshold, 0.01, tolerance = 1e-6)
  })

  it("the binarize strategy is disabled", {
    # Regression's `by_cutpoint` grouping always yields a 2-group
    # partition, so binarize is a no-op placeholder.
    d <- make_regression_data()
    model <- pprf(x = d$x, y = d$y, seed = 0, threads = 1)
    expect_equal(model$training_spec$binarize$name, "disabled")
  })

  it("the grouping strategy is by_cutpoint", {
    d <- make_regression_data()
    model <- pprf(x = d$x, y = d$y, seed = 0, threads = 1)
    expect_equal(model$training_spec$grouping$name, "by_cutpoint")
  })

  it("the leaf strategy is mean_response", {
    d <- make_regression_data()
    model <- pprf(x = d$x, y = d$y, seed = 0, threads = 1)
    expect_equal(model$training_spec$leaf$name, "mean_response")
  })
})

describe("pprf with strategy objects", {
  it("trains with pp and vars", {
    model <- pprf(Species ~ ., data = iris, size = 2, pp = pp_pda(0.5), vars = vars_uniform(n_vars = 2), seed = 0, threads = 1)
    expect_s3_class(model, "pprf")
    expect_equal(model$training_spec$pp$lambda, 0.5)
    expect_equal(model$training_spec$vars$count, 2)
  })

  it("trains with vars using p_vars", {
    model <- pprf(Species ~ ., data = iris, size = 2, vars = vars_uniform(p_vars = 0.5), seed = 0, threads = 1)
    expect_s3_class(model, "pprf")
    expect_equal(model$training_spec$vars$count, 2)  # 0.5 * 4 = 2
  })

  it("produces identical export as shortcut params", {
    model_shortcut <- pprf(Species ~ ., data = iris, size = 3, lambda = 0.5, n_vars = 2, seed = 0, threads = 1)
    model_strategy <- pprf(Species ~ ., data = iris, size = 3, pp = pp_pda(0.5), vars = vars_uniform(n_vars = 2), seed = 0, threads = 1)

    path_shortcut <- tempfile(fileext = ".json")
    path_strategy <- tempfile(fileext = ".json")
    save_json(model_shortcut, path_shortcut)
    save_json(model_strategy, path_strategy)

    j_shortcut <- jsonlite::fromJSON(readLines(path_shortcut, warn = FALSE))
    j_strategy <- jsonlite::fromJSON(readLines(path_strategy, warn = FALSE))

    j_shortcut$training_duration_ms <- NULL
    j_strategy$training_duration_ms <- NULL
    expect_equal(j_shortcut, j_strategy)
  })

  it("p_vars shortcut and vars_uniform(p_vars) produce identical export", {
    model_shortcut <- pprf(Species ~ ., data = iris, size = 3, p_vars = 0.5, seed = 0, threads = 1)
    model_strategy <- pprf(Species ~ ., data = iris, size = 3, vars = vars_uniform(p_vars = 0.5), seed = 0, threads = 1)

    path_shortcut <- tempfile(fileext = ".json")
    path_strategy <- tempfile(fileext = ".json")
    save_json(model_shortcut, path_shortcut)
    save_json(model_strategy, path_strategy)

    j_shortcut <- jsonlite::fromJSON(readLines(path_shortcut, warn = FALSE))
    j_strategy <- jsonlite::fromJSON(readLines(path_strategy, warn = FALSE))

    j_shortcut$training_duration_ms <- NULL
    j_strategy$training_duration_ms <- NULL
    expect_equal(j_shortcut, j_strategy)
  })

  it("errors when mixing pp and lambda", {
    expect_error(
      pprf(Species ~ ., data = iris, pp = pp_pda(0.5), lambda = 0.3),
      "Cannot use `pp` together with `lambda`")
  })

  it("errors when mixing vars and n_vars", {
    expect_error(
      pprf(Species ~ ., data = iris, vars = vars_uniform(n_vars = 2), n_vars = 3),
      "Cannot use `vars` together with `n_vars`/`p_vars`")
  })

  it("errors when mixing vars and p_vars", {
    expect_error(
      pprf(Species ~ ., data = iris, vars = vars_uniform(n_vars = 2), p_vars = 0.5),
      "Cannot use `vars` together with `n_vars`/`p_vars`")
  })

  it("rejects non-vars_strategy objects for vars", {
    expect_error(
      pprf(Species ~ ., data = iris, vars = list(name = "uniform", count = 2)),
      "vars_strategy object")
  })

  it("works with parsnip engine args pattern", {
    model <- pprf(Species ~ ., data = iris, size = 2, pp = pp_pda(0.5), vars = vars_uniform(n_vars = 2), seed = 0, threads = 1)
    preds <- predict(model, iris)
    expect_length(preds, 150)
  })

  it("trains with vars_all (no variable subsampling)", {
    model <- pprf(Species ~ ., data = iris, size = 2, vars = vars_all(), seed = 0, threads = 1)
    expect_s3_class(model, "pprf")
    expect_equal(model$training_spec$vars$name, "all")
  })

  it("training spec reflects strategy objects, not just defaults", {
    model <- pprf(Species ~ ., data = iris, size = 2, pp = pp_pda(0.7), vars = vars_uniform(n_vars = 3), cutpoint = cutpoint_mean_of_means(), seed = 0, threads = 1)
    expect_equal(model$training_spec$pp$name, "pda")
    expect_equal(model$training_spec$pp$lambda, 0.7, tolerance = 1e-5)
    expect_equal(model$training_spec$vars$name, "uniform")
    expect_equal(model$training_spec$vars$count, 3)
    expect_equal(model$training_spec$cutpoint$name, "mean_of_means")
  })

  it("trains with explicit stop, binarize, grouping, leaf", {
    model <- pprf(Species ~ ., data = iris, size = 2, stop = stop_pure_node(), binarize = binarize_largest_gap(), grouping = grouping_by_label(), leaf = leaf_majority_vote(), seed = 0, threads = 1)
    expect_s3_class(model, "pprf")
    expect_equal(model$training_spec$stop$name, "pure_node")
    expect_equal(model$training_spec$binarize$name, "largest_gap")
    expect_equal(model$training_spec$grouping$name, "by_label")
    expect_equal(model$training_spec$leaf$name, "majority_vote")
  })

  it("rejects non-stop_strategy objects", {
    expect_error(
      pprf(Species ~ ., data = iris, stop = list(name = "pure_node")),
      "stop_strategy")
  })

  it("rejects non-binarize_strategy objects", {
    expect_error(
      pprf(Species ~ ., data = iris, binarize = list(name = "largest_gap")),
      "binarize_strategy")
  })

  it("rejects non-grouping_strategy objects", {
    expect_error(
      pprf(Species ~ ., data = iris, grouping = list(name = "by_label")),
      "grouping_strategy")
  })

  it("rejects non-leaf_strategy objects", {
    expect_error(
      pprf(Species ~ ., data = iris, leaf = list(name = "majority_vote")),
      "leaf_strategy")
  })
})

describe("pprf edge cases", {
  it("rejects NA in features via matrix interface", {
    x <- matrix(c(1, NA, 3, 4, 1, 2, 3, 4), ncol = 2)
    y <- c("a", "a", "b", "b")
    expect_error(pprf(x = x, y = y, size = 3, threads = 1), "NA or NaN")
  })

  it("rejects NaN in features via matrix interface", {
    x <- matrix(c(1, NaN, 3, 4, 1, 2, 3, 4), ncol = 2)
    y <- c("a", "a", "b", "b")
    expect_error(pprf(x = x, y = y, size = 3, threads = 1), "NA or NaN")
  })

  it("trains with a constant feature column", {
    df <- data.frame(x1 = rep(5, 6), x2 = c(1, 2, 3, 7, 8, 9), y = c("a", "a", "a", "b", "b", "b"))
    expect_no_error(pprf(y ~ ., data = df, size = 3, seed = 0, threads = 1))
  })

  it("trains with single observation per group", {
    df <- data.frame(x1 = c(1, 0), x2 = c(0, 1), y = c("a", "b"))
    model <- pprf(y ~ ., data = df, size = 3, seed = 0, threads = 1)
    expect_s3_class(model, "pprf")
    preds <- predict(model, df)
    expect_length(preds, 2)
  })

  it("trains with minimal dataset (n=2)", {
    df <- data.frame(x1 = c(1, 9), y = c("a", "b"))
    model <- pprf(y ~ ., data = df, size = 3, seed = 0, threads = 1)
    expect_s3_class(model, "pprf")
    preds <- predict(model, df)
    expect_length(preds, 2)
  })

  it("trains with extreme class imbalance", {
    df <- data.frame(
      x1 = c(0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 0, 1, 2, 90, 91),
      x2 = c(0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 90, 91),
      y = c(rep("a", 18), rep("b", 2))
    )
    model <- pprf(y ~ ., data = df, size = 5, seed = 0, threads = 1)
    expect_s3_class(model, "pprf")
    preds <- predict(model, df)
    expect_length(preds, 20)
  })
})

describe("pprf regression", {
  it("end-to-end on mtcars (formula + predict + summary)", {
    # Real-dataset round-trip counterpart to the simulated-data tests
    # below. mtcars is small (32 x 11) and ships with base R so this
    # doesn't grow the package footprint, while still exercising the
    # full pipeline on real covariances: formula interface, regression
    # auto-detection from numeric `y`, training, prediction, and
    # summary rendering.
    data(mtcars)
    model <- pprf(mpg ~ ., data = mtcars, size = 10, seed = 0, threads = 1)

    expect_equal(model$mode, "regression")
    # `groups` is an empty character vector for regression (no class levels).
    expect_length(model$groups, 0)
    expect_s3_class(model, "pprf_regression")
    expect_s3_class(model, "pprf")
    expect_s3_class(model, "ppmodel")

    preds <- predict(model, newdata = mtcars)
    expect_true(is.numeric(preds))
    expect_length(preds, nrow(mtcars))
    expect_true(all(is.finite(preds)))

    # summary() must render the regression-shaped metrics block (MSE /
    # MAE / R²) and must not render a confusion matrix.
    out <- capture.output(summary(model))
    expect_true(any(grepl("Training", out)))
    expect_true(any(grepl("MSE|R\\^2|R\u00b2|MAE", out)))
    expect_false(any(grepl("Confusion Matrix", out)))

    # Sanity on OOB error: MSE is non-negative, or NA if no row has OOB.
    err <- oob_error(model)
    expect_true(is.na(err) || err >= 0)
  })

  it("auto-detects regression from numeric y", {
    d <- make_regression_data(n = 40)
    model <- pprf(x = d$x, y = d$y, size = 5, seed = 0, threads = 1)

    expect_equal(model$mode, "regression")
    expect_length(model$trees, 5)
  })

  it("predict returns numeric", {
    d <- make_regression_data(n = 40)
    model <- pprf(x = d$x, y = d$y, size = 5, seed = 0, threads = 1)
    preds <- predict(model, d$x)

    expect_type(preds, "double")
    expect_length(preds, 40)
  })

  it("produces a reasonable MSE", {
    d <- make_regression_data(n = 50)
    model <- pprf(x = d$x, y = d$y, size = 20, seed = 0, threads = 1)
    preds <- predict(model, d$x)
    mse <- mean((preds - d$y)^2)

    expect_lt(mse, 1.0)
  })

  it("oob_error is an MSE", {
    d <- make_regression_data(n = 40)
    model <- pprf(x = d$x, y = d$y, size = 10, seed = 0, threads = 1)

    expect_true(is.numeric(oob_error(model)))
    expect_gte(oob_error(model), 0)
  })

  it("oob_predictions uses NA_real_ (not NaN) for no-OOB rows", {
    # C++ emits NaN for rows that no tree left out-of-bag; the R boundary
    # remaps those to NA_real_ so the missing-value story is uniform with
    # the classification path (which returns NA at the factor level). This
    # test pins that remap: force every row to be in-bag in every tree by
    # rewriting `sample_indices` to cover `0:(n-1)`, then confirm
    #   (a) every entry is NA under `is.na()`, and
    #   (b) no entry is the raw C++ `NaN` sentinel.
    # If someone removes the remap, (b) fails loudly instead of quietly
    # leaking a NaN that `is.na()` would still catch.
    d <- make_regression_data(n = 20)
    model <- pprf(x = d$x, y = d$y, size = 3, seed = 0, threads = 1)
    n <- nrow(model$x)
    all_in_bag <- seq_len(n) - 1L
    for (i in seq_along(model$trees)) {
      model$trees[[i]]$sample_indices <- all_in_bag
    }
    if (!is.null(model$.cache) && exists("oob_predictions", envir = model$.cache, inherits = FALSE)) {
      rm("oob_predictions", envir = model$.cache)
    }
    preds <- oob_predictions(model)
    expect_true(all(is.na(preds)))
    expect_false(any(is.nan(preds)))
  })

  it("variable importance uses the continuous response", {
    # Regression VI (both permuted and weighted) needs the continuous
    # response as the OOB truth. Construct a dataset where x1 and x2 drive
    # y and x3 is pure noise; assert the informative columns outrank the
    # noise column for both measures.
    set.seed(7)
    n <- 100
    x <- matrix(runif(n * 3), ncol = 3)
    colnames(x) <- c("x1", "x2", "noise")
    y <- 2 * x[, 1] + 3 * x[, 2] + rnorm(n, sd = 0.1)

    # Use all features at each split so the signal-vs-noise VI separation is
    # tested independent of the default variable subsampling (p_vars = 0.5).
    model <- pprf(x = x, y = y, size = 20, n_vars = ncol(x), seed = 0, threads = 1)
    expect_equal(model$mode, "regression")

    vi_perm     <- permuted_importance(model)
    vi_weighted <- weighted_importance(model)

    expect_length(vi_perm, 3)
    expect_length(vi_weighted, 3)
    expect_true(all(is.finite(vi_perm)))
    expect_true(all(is.finite(vi_weighted)))

    # Informative columns must outrank the noise column. (Not asserting
    # the ordering between x1 and x2 — coefficient-to-VI isn't monotonic
    # for small forests, and flakes across RNG seeds.)
    expect_gt(vi_perm[1], vi_perm[3])
    expect_gt(vi_perm[2], vi_perm[3])
    expect_gt(vi_weighted[1], vi_weighted[3])
    expect_gt(vi_weighted[2], vi_weighted[3])
  })

  it("VI is reproducible across runs", {
    set.seed(11)
    n <- 60
    x <- matrix(runif(n * 2), ncol = 2)
    y <- x[, 1] - x[, 2] + rnorm(n, sd = 0.05)

    m1 <- pprf(x = x, y = y, size = 10, seed = 0, threads = 1)
    m2 <- pprf(x = x, y = y, size = 10, seed = 0, threads = 1)

    expect_equal(permuted_importance(m1), permuted_importance(m2))
    expect_equal(weighted_importance(m1), weighted_importance(m2))
  })

  it("save/load round-trip preserves predictions", {
    d <- make_regression_data(n = 40)
    model <- pprf(x = d$x, y = d$y, size = 5, seed = 0, threads = 1)

    path <- tempfile(fileext = ".json")
    save_json(model, path)
    loaded <- load_json(path)

    expect_equal(loaded$mode, "regression")

    original_preds <- predict(model, d$x)
    loaded_preds <- predict(loaded, d$x)

    expect_equal(as.numeric(loaded_preds), as.numeric(original_preds))
  })

  it("rejects NA in features via matrix interface", {
    x <- matrix(c(1, NA, 3, 4, 1, 2, 3, 4), ncol = 2)
    y <- c(1.1, 1.2, 1.3, 1.4)
    expect_error(pprf(x = x, y = y, size = 3, threads = 1), "NA or NaN")
  })

  it("rejects NaN in features via matrix interface", {
    x <- matrix(c(1, NaN, 3, 4, 1, 2, 3, 4), ncol = 2)
    y <- c(1.1, 1.2, 1.3, 1.4)
    expect_error(pprf(x = x, y = y, size = 3, threads = 1), "NA or NaN")
  })

  it("trains with a constant feature column", {
    x <- cbind(rep(5, 6), c(1, 2, 3, 7, 8, 9))
    y <- c(0.1, 0.2, 0.3, 0.7, 0.8, 0.9)
    expect_no_error(pprf(x = x, y = y, size = 3, seed = 0, threads = 1))
  })

  it("trains with minimal dataset (n = 2)", {
    x <- matrix(c(1, 9, 0, 0), ncol = 2)
    y <- c(0.0, 1.0)
    expect_no_error(pprf(x = x, y = y, size = 3, seed = 0, threads = 1))
  })

  it("accepts valid p_vars (resolves to count)", {
    d <- make_regression_data(n = 40)
    model <- pprf(x = d$x, y = d$y, p_vars = 0.5, size = 2, seed = 0, threads = 1)
    # `make_regression_data` produces 2 features; 0.5 → max(1, 1) = 1.
    expect_equal(model$training_spec$vars$count, 1L)
  })

  it("trains with vars_all (no variable subsampling)", {
    d <- make_regression_data(n = 40)
    model <- pprf(x = d$x, y = d$y, size = 2, vars = vars_all(), seed = 0, threads = 1)
    expect_s3_class(model, "pprf_regression")
    expect_equal(model$training_spec$vars$name, "all")
  })
})

describe("pprf explicit mode parameter", {
  # See the matching block in test-pptr.R for the rationale — these
  # tests exercise the same `mode` plumbing on the forest entry point.

  it("mode = \"classification\" coerces integer 0/1 labels", {
    x <- matrix(rnorm(120), ncol = 2)
    y <- rep(c(0L, 1L), 30)
    model <- pprf(x = x, y = y, mode = "classification", size = 3, seed = 0, threads = 1)

    expect_equal(model$mode, "classification")
    expect_equal(length(model$groups), 2L)
    expect_s3_class(model, "pprf_classification")
  })

  it("mode = \"regression\" errors on a factor y", {
    expect_error(
      pprf(x = iris[, 1:4], y = iris$Species, mode = "regression", size = 3, seed = 0, threads = 1),
      "`mode = \"regression\"` requires a numeric response"
    )
  })

  it("mode = NULL preserves auto-detection (factor)", {
    model <- pprf(x = iris[, 1:4], y = iris$Species, mode = NULL, size = 3, seed = 0, threads = 1)
    expect_equal(model$mode, "classification")
  })

  it("mode = NULL preserves auto-detection (numeric)", {
    d <- make_regression_data()
    model <- pprf(x = d$x, y = d$y, mode = NULL, size = 3, seed = 0, threads = 1)
    expect_equal(model$mode, "regression")
  })

  it("rejects invalid mode strings", {
    expect_error(
      pprf(x = iris[, 1:4], y = iris$Species, mode = "ranking", size = 3, seed = 0, threads = 1),
      "`mode` must be either"
    )
  })
})

describe("pprf call capture and update()", {
  # `model$call` powers `stats::update.default`. We stick to checking the
  # contract (call is captured, update re-fits with the substituted arg)
  # rather than the exact deparsed form, so the test survives any future
  # cosmetic tweaks to print or to `match.call`'s output.
  it("stores the user's call on the fitted model", {
    model <- pprf(Species ~ ., data = iris, size = 3, lambda = 0.0, threads = 1)
    expect_false(is.null(model$call))
    expect_true(is.call(model$call))
    expect_equal(as.character(model$call)[[1L]], "pprf")
  })

  it("update() re-fits with the substituted argument", {
    set.seed(0)
    fit1 <- pprf(Species ~ ., data = iris, size = 3, lambda = 0.0, threads = 1)
    expect_equal(fit1$training_spec$pp$lambda, 0.0)

    fit2 <- update(fit1, lambda = 0.5)
    expect_s3_class(fit2, "pprf")
    expect_equal(fit2$training_spec$pp$lambda, 0.5)
    # Unchanged args carry over from the original call.
    expect_equal(length(fit2$trees), length(fit1$trees))
  })

  it("print() surfaces the call when present", {
    model <- pprf(Species ~ ., data = iris, size = 3, lambda = 0.0, threads = 1)
    out <- capture.output(print(model))
    expect_true(any(grepl("^\\s*Call:", out)))
    expect_true(any(grepl("pprf\\(", out)))
  })
})

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

describe("save_json / load_json round-trip", {
  describe("pptr (single tree)", {
    it("preserves predictions after round-trip", {
      model <- pptr(Species ~ ., data = iris, seed = 0)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      loaded <- load_json(path)

      x <- as.matrix(iris[, 1:4])
      original_preds <- predict(model, iris)
      loaded_preds <- predict(loaded, x)
      expect_equal(loaded_preds, original_preds)
    })

    it("preserves variable importance after round-trip", {
      model <- pptr(Species ~ ., data = iris, seed = 0)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      loaded <- load_json(path)

      expect_equal(unname(loaded$vi$scale), unname(model$vi$scale), tolerance = 1e-4)
      expect_equal(loaded$vi$projections, model$vi$projections, tolerance = 1e-4)
    })

    it("preserves group labels after round-trip", {
      model <- pptr(Species ~ ., data = iris, seed = 0)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      loaded <- load_json(path)

      expect_equal(loaded$groups, model$groups)
    })

    it("preserves training spec after round-trip", {
      model <- pptr(Species ~ ., data = iris, seed = 0, lambda = 0.5)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      loaded <- load_json(path)

      expect_equal(loaded$training_spec$pp$name, "pda")
      expect_equal(loaded$training_spec$pp$lambda, 0.5, tolerance = 1e-5)
      expect_equal(loaded$seed, 0)
    })

    it("sets formula, x, y to NULL on load", {
      model <- pptr(Species ~ ., data = iris, seed = 0)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      loaded <- load_json(path)

      expect_null(loaded$formula)
      expect_null(loaded$x)
      expect_null(loaded$y)
    })
  })

  describe("pprf regression", {
    # Regression has no group concept — `meta.groups` must be `null` on
    # disk and `character(0)` on the loaded model. The wrap also exposes
    # full `oob_metrics` (mse / mae / r_squared) instead of a flattened
    # scalar, mirroring the C++ Export shape.
    it("preserves predictions and metric structure after round-trip", {
      data(mtcars)
      model <- pprf(mpg ~ ., data = mtcars, size = 5, seed = 0, threads = 1)
      path <- tempfile(fileext = ".json")
      save_json(model, path)

      # Raw JSON: meta.groups is an empty array for regression (matches the
      # in-memory empty `Names`; the discriminator is `config.mode`).
      raw <- jsonlite::fromJSON(readLines(path, warn = FALSE), simplifyVector = FALSE)
      expect_equal(length(raw$meta$groups), 0)
      expect_equal(raw$config$mode, "regression")

      loaded <- load_json(path)
      expect_s3_class(loaded, "pprf_regression")
      expect_length(loaded$groups, 0)

      # Predictions match.
      x <- as.matrix(mtcars[, -1])
      expect_equal(predict(loaded, x), predict(model, mtcars), tolerance = 1e-5)

      # Full regression metrics block is exposed (not a single scalar).
      expect_true(is.list(loaded$oob_metrics))
      expect_named(loaded$oob_metrics, c("mse", "mae", "r_squared"), ignore.order = TRUE)
      expect_false(is.null(loaded$training_metrics))
      expect_named(loaded$training_metrics, c("mse", "mae", "r_squared"), ignore.order = TRUE)

      # `oob_error()` derives the scalar from the metrics block.
      err <- oob_error(loaded)
      expect_true(is.na(err) || isTRUE(all.equal(err, loaded$oob_metrics$mse, tolerance = 1e-5)))
    })
  })

  describe("pprf (random forest)", {
    it("preserves predictions after round-trip", {
      model <- pprf(Species ~ ., data = iris, size = 3, seed = 0)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      loaded <- load_json(path)

      x <- as.matrix(iris[, 1:4])
      original_preds <- predict(model, iris)
      loaded_preds <- predict(loaded, x)
      expect_equal(loaded_preds, original_preds)
    })

    it("preserves variable importance after round-trip", {
      model <- pprf(Species ~ ., data = iris, size = 3, seed = 0)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      loaded <- load_json(path)

      expect_equal(unname(loaded$vi$scale), unname(model$vi$scale), tolerance = 1e-4)
      expect_equal(loaded$vi$projections, model$vi$projections, tolerance = 1e-4)
      expect_equal(weighted_importance(loaded), weighted_importance(model), tolerance = 1e-4)
      expect_equal(permuted_importance(loaded), permuted_importance(model), tolerance = 1e-4)
    })

    it("preserves oob_error after round-trip", {
      model <- pprf(Species ~ ., data = iris, size = 3, seed = 0)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      loaded <- load_json(path)

      expect_equal(oob_error(loaded), oob_error(model), tolerance = 1e-5)
    })

    it("exposes full classification metrics block after round-trip", {
      # Classification mirrors the regression test above: full
      # `training_metrics` / `oob_metrics` blocks (not flattened scalars),
      # each carrying the confusion matrix and error rate. The CM's labels
      # are group names (strings), and its dimensions match meta.groups
      # in the well-balanced case.
      model <- pprf(Species ~ ., data = iris, size = 3, seed = 0)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      loaded <- load_json(path)

      expect_true(is.list(loaded$training_metrics))
      expect_named(loaded$training_metrics, c("confusion_matrix", "error_rate"),
        ignore.order = TRUE)
      cm <- loaded$training_metrics$confusion_matrix
      expect_named(cm, c("matrix", "labels", "group_errors"), ignore.order = TRUE)
      expect_equal(dim(cm$matrix), c(length(loaded$groups), length(loaded$groups)))

      expect_true(is.list(loaded$oob_metrics))
      expect_named(loaded$oob_metrics, c("confusion_matrix", "error_rate"),
        ignore.order = TRUE)
    })

    it("CM labels equal meta.groups (no prediction-only labels in normal use)", {
      # Asserts the C++ contract that, in normal training/evaluation,
      # `predictions ⊆ groups` and `actual ⊆ groups`, so the CM's
      # `labels` set equals `meta.groups`. If a future change introduces
      # prediction-only labels (a class the model never predicted, or one
      # it predicted that never appeared in `y`), this assertion fires
      # and the R-side CM helper in `test-reproducibility.R` needs an
      # update to handle the asymmetric case.
      model <- pprf(Species ~ ., data = iris, size = 3, seed = 0)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      raw <- jsonlite::fromJSON(readLines(path, warn = FALSE), simplifyVector = FALSE)

      cm_labels <- vapply(raw$training_metrics$confusion_matrix$labels,
        as.character, character(1))
      expect_equal(cm_labels, unlist(raw$meta$groups))
    })

    it("preserves group labels on individual trees", {
      model <- pprf(Species ~ ., data = iris, size = 3, seed = 0)
      path <- tempfile(fileext = ".json")
      save_json(model, path)
      loaded <- load_json(path)

      for (i in seq_along(loaded$trees)) {
        expect_s3_class(loaded$trees[[i]], "pptr")
        expect_equal(loaded$trees[[i]]$groups, model$groups)
      }
    })
  })
})

describe("save_json includes data metadata", {
  it("includes observations, features, and feature_names for tree", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    path <- tempfile(fileext = ".json")
    save_json(model, path)

    j <- jsonlite::fromJSON(readLines(path, warn = FALSE))
    expect_equal(j$meta$observations, 150)
    expect_equal(j$meta$features, 4)
    expect_equal(j$meta$feature_names, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width"))
  })

  it("includes observations, features, and feature_names for forest", {
    model <- pprf(Species ~ ., data = iris, size = 3, seed = 0)
    path <- tempfile(fileext = ".json")
    save_json(model, path)

    j <- jsonlite::fromJSON(readLines(path, warn = FALSE))
    expect_equal(j$meta$observations, 150)
    expect_equal(j$meta$features, 4)
    expect_equal(j$meta$feature_names, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width"))
  })

})

describe("save_json with include_metrics = FALSE", {
  it("saves model without VI", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    path <- tempfile(fileext = ".json")
    save_json(model, path, include_metrics = FALSE)
    loaded <- load_json(path)

    expect_null(loaded$vi)

    x <- as.matrix(iris[, 1:4])
    original_preds <- predict(model, iris)
    loaded_preds <- predict(loaded, x)
    expect_equal(loaded_preds, original_preds)
  })
})

describe("load_json from golden files", {
  golden_path <- function(...) {
    system.file("golden", ..., package = "ppforest2")
  }

  it("loads a golden tree and produces correct predictions", {
    path <- golden_path("iris", "tree-pda-s0.json")
    skip_if(path == "", "Golden file not bundled")

    loaded <- load_json(path)
    expect_s3_class(loaded, "pptr")
    expect_equal(loaded$groups, c("setosa", "versicolor", "virginica"))

    x <- as.matrix(iris[, 1:4])
    preds <- predict(loaded, x)
    expect_length(preds, 150)
  })

  it("loads a golden forest", {
    path <- golden_path("iris", "forest-pda-n5-s0.json")
    skip_if(path == "", "Golden file not bundled")

    loaded <- load_json(path)
    expect_s3_class(loaded, "pprf")
    expect_equal(length(loaded$trees), 5)
  })
})

describe("R save_json meta matches golden file", {
  golden_path <- function(...) {
    system.file("golden", ..., package = "ppforest2")
  }

  golden_meta <- function(dataset, slug) {
    path <- golden_path(dataset, paste0(slug, ".json"))
    if (path == "") return(NULL)
    j <- jsonlite::fromJSON(readLines(path, warn = FALSE))
    j$meta
  }

  it("tree meta matches golden file", {
    expected <- golden_meta("iris", "tree-pda-s0")
    skip_if(is.null(expected), "Golden file not bundled")

    model <- pptr(Species ~ ., data = iris, seed = 0)
    path <- tempfile(fileext = ".json")
    save_json(model, path)
    j <- jsonlite::fromJSON(readLines(path, warn = FALSE))

    expect_equal(j$meta, expected)
  })

  it("forest meta matches golden file", {
    expected <- golden_meta("iris", "forest-pda-n5-s0")
    skip_if(is.null(expected), "Golden file not bundled")

    model <- pprf(Species ~ ., data = iris, size = 5, seed = 0)
    path <- tempfile(fileext = ".json")
    save_json(model, path)
    j <- jsonlite::fromJSON(readLines(path, warn = FALSE))

    expect_equal(j$meta, expected)
  })
})

describe("save_json / load_json with non-default strategies", {
  it("preserves tree training spec (pp, vars, cutpoint) after round-trip", {
    model <- pptr(Species ~ ., data = iris, pp = pp_pda(0.5), seed = 0)
    path <- tempfile(fileext = ".json")
    save_json(model, path)
    loaded <- load_json(path)

    expect_equal(loaded$training_spec$pp$name, "pda")
    expect_equal(loaded$training_spec$pp$lambda, 0.5, tolerance = 1e-5)
    expect_equal(loaded$training_spec$vars$name, "all")
    expect_equal(loaded$training_spec$cutpoint$name, "mean_of_means")
  })

  it("preserves forest training spec with vars_uniform after round-trip", {
    model <- pprf(Species ~ ., data = iris, size = 3, pp = pp_pda(0.3), vars = vars_uniform(n_vars = 2), seed = 0)
    path <- tempfile(fileext = ".json")
    save_json(model, path)
    loaded <- load_json(path)

    expect_equal(loaded$training_spec$pp$name, "pda")
    expect_equal(loaded$training_spec$pp$lambda, 0.3, tolerance = 1e-5)
    expect_equal(loaded$training_spec$vars$name, "uniform")
    expect_equal(loaded$training_spec$vars$count, 2)
    expect_equal(loaded$training_spec$cutpoint$name, "mean_of_means")
  })

  it("preserves forest training spec with vars_all after round-trip", {
    model <- pprf(Species ~ ., data = iris, size = 3, vars = vars_all(), seed = 0)
    path <- tempfile(fileext = ".json")
    save_json(model, path)
    loaded <- load_json(path)

    expect_equal(loaded$training_spec$vars$name, "all")
  })

  it("does not include display_name in saved JSON", {
    model <- pptr(Species ~ ., data = iris, pp = pp_pda(0.5), seed = 0)
    path <- tempfile(fileext = ".json")
    save_json(model, path)

    j <- jsonlite::fromJSON(readLines(path, warn = FALSE))
    expect_null(j$config$pp$display_name)
    expect_null(j$config$vars$display_name)
    expect_null(j$config$cutpoint$display_name)
  })

  it("predictions are correct after loading model with non-default strategies", {
    model <- pprf(Species ~ ., data = iris, size = 3, pp = pp_pda(0.3), vars = vars_uniform(n_vars = 2), seed = 0)
    path <- tempfile(fileext = ".json")
    save_json(model, path)
    loaded <- load_json(path)

    original_preds <- predict(model, iris)
    loaded_preds <- predict(loaded, model$x)
    expect_equal(loaded_preds, original_preds)
  })
})

describe("load_json error handling", {
  it("errors on invalid JSON", {
    path <- tempfile(fileext = ".json")
    writeLines("not valid json", path)
    expect_error(load_json(path))
  })
})

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

describe("pprf variable importance", {
  it("has vi with scale, projections, weighted, permuted", {
    model <- pprf(Species ~ ., data = iris, size = 5, threads = 1)
    # Forest VI: `scale` and `projections` live on `model$vi` (eager, cheap);
    # `weighted` and `permuted` are lazy accessors (not stored as fields).
    expect_true(!is.null(model$vi))
    expect_true("scale" %in% names(model$vi))
    expect_true("projections" %in% names(model$vi))
    p <- ncol(model$x)
    expect_length(model$vi$scale, p)
    expect_length(model$vi$projections, p)
    expect_length(weighted_importance(model), p)
    expect_length(permuted_importance(model), p)
  })

  it("projection_importance returns the eager VI2 vector", {
    model <- pprf(Species ~ ., data = iris, size = 5, threads = 1)
    pi <- projection_importance(model)
    expect_length(pi, ncol(model$x))
    expect_equal(pi, model$vi$projections)
    expect_true(all(pi >= 0))
  })

  it("projection_importance errors for non-ppforest2 objects", {
    expect_error(projection_importance(list()), "only defined for ppforest2")
  })

  it("has oob_error", {
    model <- pprf(Species ~ ., data = iris, size = 5, threads = 1)
    err <- oob_error(model)
    expect_true(!is.null(err))
    if (!is.na(err)) {
      expect_gte(err, 0)
      expect_lte(err, 1)
    }
  })

  it("oob_predictions returns all-NA factor when no row has any OOB tree", {
    # Symmetric counterpart to the regression-side test: force every row
    # to be in-bag in every tree, then confirm `oob_predictions` returns
    # an all-NA factor (missing at the factor level, not a raw sentinel
    # value leaking through).
    #
    # Sentinel chain (unified with regression):
    #   1. C++ `oob_predict` fills no-OOB rows with `NaN` (in both modes).
    #   2. The Rcpp wrapper's `to_r_indices` shift adds +1 to every
    #      element — `NaN + 1` is still `NaN`, so the sentinel survives.
    #   3. The R accessor remaps `NaN → NA_real_` explicitly, then
    #      `as.integer` produces `NA_integer_`, which indexes into
    #      `model$groups` to yield `NA_character_`, which `factor()`
    #      records as a missing level.
    #
    # We pin three properties to catch every leak mode:
    #   (a) `length(preds) == n` — guards against an integer-coded
    #       sentinel (e.g. a stray `0`) leaking through `model$groups[raw]`
    #       and silently dropping rows: R's zero-indexing returns
    #       `character(0)` for index 0, which `factor()` would then
    #       build into a *shorter* result that the length-blind
    #       `all(is.na(preds))` check would pass vacuously.
    #   (b) every entry is NA at the factor level (`is.na(preds)`).
    #   (c) the underlying integer codes are `NA_integer_`. With the
    #       NaN sentinel, the cast from numeric to integer is the
    #       chokepoint — `as.integer(NA_real_)` is `NA_integer_`, so a
    #       correct chain shows up here as all-NA codes. Belt-and-
    #       braces with (a) for the "leak then crash" vs "leak then
    #       mislabel" failure modes.
    #   (d) levels match `model$groups` — "all missing" must not drop
    #       levels.
    model <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
    n <- nrow(model$x)
    all_in_bag <- seq_len(n) - 1L
    for (i in seq_along(model$trees)) {
      model$trees[[i]]$sample_indices <- all_in_bag
    }
    if (!is.null(model$.cache) && exists("oob_predictions", envir = model$.cache, inherits = FALSE)) {
      rm("oob_predictions", envir = model$.cache)
    }
    preds <- oob_predictions(model)
    expect_true(is.factor(preds))
    expect_length(preds, n)
    expect_equal(levels(preds), model$groups)
    expect_true(all(is.na(preds)))
    expect_true(all(is.na(as.integer(preds))))
  })

  it("oob_error returns NA when no observation has any OOB tree", {
    # Construct a degenerate forest by manually rewriting each tree's
    # sample_indices to cover every training row (all-in-bag). This
    # removes every row from the OOB set, so `oob_error` has no data to
    # compute on and must surface `NA_real_` rather than a sentinel like
    # `-1` that callers might mistake for a (representable but
    # mathematically impossible) error rate.
    model <- pprf(Species ~ ., data = iris, size = 3, threads = 1)
    n <- nrow(model$x)
    all_in_bag <- seq_len(n) - 1L  # 0-based, as stored on the tree
    for (i in seq_along(model$trees)) {
      model$trees[[i]]$sample_indices <- all_in_bag
    }
    # Invalidate the cache (oob_error may have been eagerly primed).
    if (!is.null(model$.cache) && exists("oob_error", envir = model$.cache, inherits = FALSE)) {
      rm("oob_error", envir = model$.cache)
    }
    expect_true(is.na(oob_error(model)))
  })

  it("summary output contains Variable Importance and OOB error", {
    model <- pprf(Species ~ ., data = iris, size = 5, threads = 1)
    out <- capture.output(summary(model))
    expect_true(any(grepl("Variable Importance", out)))
    if (!is.na(oob_error(model))) {
      expect_true(any(grepl("OOB error", out)))
    }
  })
})

describe("pptr variable importance", {
  it("has vi with scale and projections only", {
    model <- pptr(Species ~ ., data = iris)
    expect_true(!is.null(model$vi))
    expect_true("scale" %in% names(model$vi))
    expect_true("projections" %in% names(model$vi))
    expect_false("weighted" %in% names(model$vi))
    expect_false("permuted" %in% names(model$vi))
    p <- ncol(model$x)
    expect_length(model$vi$scale, p)
    expect_length(model$vi$projections, p)
  })

  it("projection_importance works for single trees", {
    model <- pptr(Species ~ ., data = iris)
    pi <- projection_importance(model)
    expect_length(pi, ncol(model$x))
    expect_equal(pi, model$vi$projections)
  })

  it("summary output contains Variable Importance", {
    model <- pptr(Species ~ ., data = iris)
    out <- capture.output(summary(model))
    expect_true(any(grepl("Variable Importance", out)))
  })
})

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

if (!requireNamespace("jsonlite", quietly = TRUE)) {
  stop("jsonlite is required for reproducibility tests")
}

golden_path <- function(...) {
  system.file("golden", ..., package = "ppforest2")
}

load_golden <- function(...) {
  path <- golden_path(...)
  if (path == "" || !file.exists(path)) {
    stop(paste("Golden file not found:", ...))
  }
  # simplifyVector = FALSE preserves nested tree structure for forests.
  # Flat arrays are converted to vectors/matrices in comparison helpers.
  jsonlite::fromJSON(path, simplifyVector = FALSE)
}

# Convert a JSON list (from simplifyVector=FALSE) to a numeric/integer vector.
as_vec <- function(x) unlist(x)

# Convert a JSON list-of-lists to a matrix (row-major).
as_mat <- function(x) {
  do.call(rbind, lapply(x, unlist))
}

# Golden files use C++ group ordering (order of first appearance in CSV).
# R uses alphabetical factor levels by default. To ensure the R model is
# trained on identically-ordered data, we reorder the factor levels to
# match the golden file before training.
prepare_data <- function(data, golden, response = "Type") {
  # Response column defaults to "Type" (the convention for ppforest2's bundled
  # classification datasets) but is overridable so base R's `iris` (response
  # column "Species") works without a rename shim at every call site.
  golden_groups <- as_vec(golden$meta$groups)
  data[[response]] <- factor(data[[response]], levels = golden_groups)
  data
}

# ---------------------------------------------------------------------------
# Config (strategy) comparison
# ---------------------------------------------------------------------------

compare_strategy <- function(spec_strategy, config_strategy) {
  # Compare all fields from the golden config against the training spec.
  # The R training spec has an extra display_name field from wrap(); ignore it.
  for (key in names(config_strategy)) {
    expect_equal(spec_strategy[[key]], config_strategy[[key]], tolerance = 1e-5, info = paste("strategy field:", key))
  }
}

compare_config <- function(model, golden) {
  config <- golden$config
  spec <- model$training_spec

  # Strategies
  compare_strategy(spec$pp, config$pp)
  compare_strategy(spec$vars, config$vars)
  compare_strategy(spec$cutpoint, config$cutpoint)
  compare_strategy(spec$stop, config$stop)
  compare_strategy(spec$binarize, config$binarize)
  compare_strategy(spec$grouping, config$grouping)
  compare_strategy(spec$leaf, config$leaf)

  # Scalar fields
  expect_equal(spec$size, config$size)
  expect_equal(spec$seed, config$seed)
  expect_equal(spec$max_retries, config$max_retries)
}

# ---------------------------------------------------------------------------
# Model structure comparison
# ---------------------------------------------------------------------------

# Normalize a node to a canonical representation for comparison.
# R model nodes use 1-based group indices; golden JSON uses string labels.
# R indices are mapped to string labels via group_names for comparison.
normalize_node <- function(node, group_names = NULL) {
  if (!is.null(node$value)) {
    value <- if (!is.null(group_names)) {
      group_names[as.integer(node$value)]
    } else {
      node$value
    }

    list(value = value)
  } else {
    groups <- if (!is.null(group_names)) {
      sort(group_names[as.integer(unlist(node$groups))])
    } else {
      sort(as.character(unlist(node$groups)))
    }

    list(
      projector      = as.numeric(unlist(node$projector)),
      cutpoint      = as.numeric(node$cutpoint),
      pp_index_value = as.numeric(node$pp_index_value),
      groups         = groups,
      lower          = normalize_node(node$lower, group_names),
      upper          = normalize_node(node$upper, group_names)
    )
  }
}

compare_tree <- function(actual_root, expected_root, group_names, info = "") {
  actual <- normalize_node(actual_root, group_names = group_names)
  expected <- normalize_node(expected_root, group_names = NULL)
  expect_equal(actual, expected, tolerance = 1e-3, info = info)
}

# ---------------------------------------------------------------------------
# Output comparison helpers
# ---------------------------------------------------------------------------

compare_predictions <- function(model, golden, data) {
  predicted <- predict(model, data)
  golden_preds <- as.character(as_vec(golden$predictions))
  expected <- factor(golden_preds, levels = levels(predicted))
  expect_equal(predicted, expected)
}

compare_error_rate <- function(model, golden, data, response) {
  predicted <- predict(model, data)
  error_rate <- mean(predicted != response)
  expect_equal(error_rate, golden$error_rate, tolerance = 1e-3)
}

compare_confusion_matrix <- function(model, golden, data, response) {
  predicted <- predict(model, data)
  expected_cm <- golden$training_metrics$confusion_matrix
  golden_groups <- as.character(as_vec(golden$meta$groups))
  cm_labels <- as.character(as_vec(expected_cm$labels))

  # The saved CM is dimensioned by `union(predictions, actual)` on the
  # C++ side. In normal operation predictions ⊆ groups and actual ⊆
  # groups, so the CM labels equal the meta.groups set — assert that
  # invariant explicitly so any future "prediction-only label" case
  # surfaces here instead of silently producing a misaligned matrix.
  expect_equal(cm_labels, golden_groups, info = "CM label set must equal meta.groups (no prediction-only labels)")

  n_groups <- length(model$groups)
  # Align predicted levels with response levels (golden group ordering)
  predicted <- factor(predicted, levels = levels(response))
  actual_idx <- as.integer(response)
  pred_idx <- as.integer(predicted)

  cm <- matrix(0L, nrow = n_groups, ncol = n_groups)
  for (i in seq_along(actual_idx)) {
    cm[actual_idx[i], pred_idx[i]] <- cm[actual_idx[i], pred_idx[i]] + 1L
  }

  expect_equal(cm, as_mat(expected_cm$matrix))

  group_errors <- numeric(n_groups)
  for (i in seq_len(n_groups)) {
    row_total <- sum(cm[i, ])
    if (row_total > 0) {
      group_errors[i] <- 1 - cm[i, i] / row_total
    }
  }
  expect_equal(group_errors, as.numeric(as_vec(expected_cm$group_errors)),
    tolerance = 1e-3)
}

compare_vote_proportions <- function(model, golden, data) {
  probs <- predict(model, data, type = "prob")
  actual <- unname(as.matrix(probs))
  expect_equal(actual, as_mat(golden$vote_proportions), tolerance = 1e-3)
}

compare_vi <- function(model, golden, key, r_field) {
  expected <- as.numeric(as_vec(golden$variable_importance[[key]]))
  # The eager fields on model$vi are `scale` and `projections`; weighted and
  # permuted importances are lazy accessors.
  actual <- switch(r_field,
    weighted = weighted_importance(model),
    permuted = permuted_importance(model),
    model$vi[[r_field]]
  )
  expect_equal(actual, expected, tolerance = 1e-3)
}

# ---------------------------------------------------------------------------
# Tree: iris tree-pda-s0
# ---------------------------------------------------------------------------

describe("Reproducibility: iris tree-pda-s0", {
  golden <- load_golden("iris", "tree-pda-s0.json")
  d <- prepare_data(iris, golden, response = "Species")
  model <- pptr(Species ~ ., data = d, lambda = 0, seed = 0)

  it("training config matches golden file", {
    compare_config(model, golden)
  })

  it("model structure matches golden file", {
    compare_tree(model$root, golden$model$root, as_vec(golden$meta$groups))
  })

  it("predictions match golden file", {
    compare_predictions(model, golden, d)
  })

  it("error rate matches golden file", {
    compare_error_rate(model, golden, d, d$Species)
  })

  it("confusion matrix matches golden file", {
    compare_confusion_matrix(model, golden, d, d$Species)
  })

  it("VI projections match golden file", {
    compare_vi(model, golden, "projections", "projections")
  })
})

# ---------------------------------------------------------------------------
# Tree: crab tree-pda-s0
# ---------------------------------------------------------------------------

describe("Reproducibility: crab tree-pda-s0", {
  golden <- load_golden("crab", "tree-pda-s0.json")
  d <- prepare_data(crab, golden)
  model <- pptr(Type ~ ., data = d, lambda = 0, seed = 0)

  it("training config matches golden file", {
    compare_config(model, golden)
  })

  it("model structure matches golden file", {
    compare_tree(model$root, golden$model$root, as_vec(golden$meta$groups))
  })

  it("predictions match golden file", {
    compare_predictions(model, golden, d)
  })

  it("error rate matches golden file", {
    compare_error_rate(model, golden, d, d$Type)
  })

  it("confusion matrix matches golden file", {
    compare_confusion_matrix(model, golden, d, d$Type)
  })

  it("VI projections match golden file", {
    compare_vi(model, golden, "projections", "projections")
  })
})

# ---------------------------------------------------------------------------
# Forest: iris forest-pda-n5-s0
# ---------------------------------------------------------------------------

describe("Reproducibility: iris forest-pda-n5-s0", {
  golden <- load_golden("iris", "forest-pda-n5-s0.json")
  d <- prepare_data(iris, golden, response = "Species")
  model <- pprf(Species ~ ., data = d, size = 5, n_vars = 2, lambda = 0, seed = 0, threads = 1)

  it("training config matches golden file", {
    compare_config(model, golden)
  })

  it("model structure matches golden file", {
    for (i in seq_along(model$trees)) {
      compare_tree(model$trees[[i]]$root, golden$model$trees[[i]]$root,
        as_vec(golden$meta$groups), info = paste("tree", i))
    }
  })

  it("predictions match golden file", {
    compare_predictions(model, golden, d)
  })

  it("error rate matches golden file", {
    compare_error_rate(model, golden, d, d$Species)
  })

  it("confusion matrix matches golden file", {
    compare_confusion_matrix(model, golden, d, d$Species)
  })

  it("OOB error matches golden file", {
    expect_equal(oob_error(model), golden$oob_metrics$error_rate, tolerance = 1e-3)
  })

  it("VI projections match golden file", {
    compare_vi(model, golden, "projections", "projections")
  })

  it("VI weighted projections match golden file", {
    compare_vi(model, golden, "weighted_projections", "weighted")
  })

  it("VI permuted match golden file", {
    compare_vi(model, golden, "permuted", "permuted")
  })

  it("vote proportions match golden file", {
    compare_vote_proportions(model, golden, d)
  })
})

# ---------------------------------------------------------------------------
# Forest: iris forest-pda-l05-n5-s0
# ---------------------------------------------------------------------------

describe("Reproducibility: iris forest-pda-l05-n5-s0", {
  golden <- load_golden("iris", "forest-pda-l05-n5-s0.json")
  d <- prepare_data(iris, golden, response = "Species")
  model <- pprf(Species ~ ., data = d, size = 5, n_vars = 2, lambda = 0.5, seed = 0, threads = 1)

  it("training config matches golden file", {
    compare_config(model, golden)
  })

  it("model structure matches golden file", {
    for (i in seq_along(model$trees)) {
      compare_tree(model$trees[[i]]$root, golden$model$trees[[i]]$root,
        as_vec(golden$meta$groups), info = paste("tree", i))
    }
  })

  it("predictions match golden file", {
    compare_predictions(model, golden, d)
  })

  it("error rate matches golden file", {
    compare_error_rate(model, golden, d, d$Species)
  })

  it("confusion matrix matches golden file", {
    compare_confusion_matrix(model, golden, d, d$Species)
  })

  it("OOB error matches golden file", {
    expect_equal(oob_error(model), golden$oob_metrics$error_rate, tolerance = 1e-3)
  })

  it("VI projections match golden file", {
    compare_vi(model, golden, "projections", "projections")
  })

  it("VI weighted projections match golden file", {
    compare_vi(model, golden, "weighted_projections", "weighted")
  })

  it("VI permuted match golden file", {
    compare_vi(model, golden, "permuted", "permuted")
  })

  it("vote proportions match golden file", {
    compare_vote_proportions(model, golden, d)
  })
})

# ---------------------------------------------------------------------------
# Forest: crab forest-pda-n10-s0
# ---------------------------------------------------------------------------

describe("Reproducibility: crab forest-pda-n10-s0", {
  golden <- load_golden("crab", "forest-pda-n10-s0.json")
  d <- prepare_data(crab, golden)
  model <- pprf(Type ~ ., data = d, size = 10, n_vars = 3, lambda = 0, seed = 0, threads = 1)

  it("training config matches golden file", {
    compare_config(model, golden)
  })

  it("model structure matches golden file", {
    for (i in seq_along(model$trees)) {
      compare_tree(model$trees[[i]]$root, golden$model$trees[[i]]$root,
        as_vec(golden$meta$groups), info = paste("tree", i))
    }
  })

  it("predictions match golden file", {
    compare_predictions(model, golden, d)
  })

  it("error rate matches golden file", {
    compare_error_rate(model, golden, d, d$Type)
  })

  it("confusion matrix matches golden file", {
    compare_confusion_matrix(model, golden, d, d$Type)
  })

  it("OOB error matches golden file", {
    expect_equal(oob_error(model), golden$oob_metrics$error_rate, tolerance = 1e-3)
  })

  it("VI projections match golden file", {
    compare_vi(model, golden, "projections", "projections")
  })

  it("VI weighted projections match golden file", {
    compare_vi(model, golden, "weighted_projections", "weighted")
  })

  it("VI permuted match golden file", {
    compare_vi(model, golden, "permuted", "permuted")
  })

  it("vote proportions match golden file", {
    compare_vote_proportions(model, golden, d)
  })
})

# ---------------------------------------------------------------------------
# Forest: wine forest-pda-n10-s0
# ---------------------------------------------------------------------------

describe("Reproducibility: wine forest-pda-n10-s0", {
  golden <- load_golden("wine", "forest-pda-n10-s0.json")
  d <- prepare_data(wine, golden)
  model <- pprf(Type ~ ., data = d, size = 10, n_vars = 4, lambda = 0, seed = 0, threads = 1)

  it("training config matches golden file", {
    compare_config(model, golden)
  })

  it("model structure matches golden file", {
    for (i in seq_along(model$trees)) {
      compare_tree(model$trees[[i]]$root, golden$model$trees[[i]]$root,
        as_vec(golden$meta$groups), info = paste("tree", i))
    }
  })

  it("predictions match golden file", {
    compare_predictions(model, golden, d)
  })

  it("error rate matches golden file", {
    compare_error_rate(model, golden, d, d$Type)
  })

  it("confusion matrix matches golden file", {
    compare_confusion_matrix(model, golden, d, d$Type)
  })

  it("OOB error matches golden file", {
    expect_equal(oob_error(model), golden$oob_metrics$error_rate, tolerance = 1e-3)
  })

  it("VI projections match golden file", {
    compare_vi(model, golden, "projections", "projections")
  })

  it("VI weighted projections match golden file", {
    compare_vi(model, golden, "weighted_projections", "weighted")
  })

  it("VI permuted match golden file", {
    compare_vi(model, golden, "permuted", "permuted")
  })

  it("vote proportions match golden file", {
    compare_vote_proportions(model, golden, d)
  })
})

# ---------------------------------------------------------------------------
# Forest: glass forest-pda-n10-s0
# ---------------------------------------------------------------------------

describe("Reproducibility: glass forest-pda-n10-s0", {
  golden <- load_golden("glass", "forest-pda-n10-s0.json")
  d <- prepare_data(glass, golden)
  model <- pprf(Type ~ ., data = d, size = 10, n_vars = 3, lambda = 0, seed = 0, threads = 1)

  it("training config matches golden file", {
    compare_config(model, golden)
  })

  it("model structure matches golden file", {
    for (i in seq_along(model$trees)) {
      compare_tree(model$trees[[i]]$root, golden$model$trees[[i]]$root,
        as_vec(golden$meta$groups), info = paste("tree", i))
    }
  })

  it("predictions match golden file", {
    compare_predictions(model, golden, d)
  })

  it("error rate matches golden file", {
    compare_error_rate(model, golden, d, d$Type)
  })

  it("confusion matrix matches golden file", {
    compare_confusion_matrix(model, golden, d, d$Type)
  })

  it("OOB error matches golden file", {
    expect_equal(oob_error(model), golden$oob_metrics$error_rate, tolerance = 1e-3)
  })

  it("VI projections match golden file", {
    compare_vi(model, golden, "projections", "projections")
  })

  it("VI weighted projections match golden file", {
    compare_vi(model, golden, "weighted_projections", "weighted")
  })

  it("VI permuted match golden file", {
    compare_vi(model, golden, "permuted", "permuted")
  })

  it("vote proportions match golden file", {
    compare_vote_proportions(model, golden, d)
  })
})

#' Crabs Dataset
#'
#' Measurements of 200 crabs from the genus Leptograpsus, with 50 crabs for each of two colour forms and both sexes. Measurements include the frontal lobe size, rear width, carapace length, carapace width, and body depth.
#'
#' @format A data frame with 200 rows and 8 variables:
#' \describe{
#'   \item{Type}{Species code, with `B` indicating blue and `O` indicating orange.}
#'   \item{sex}{Sex of the crab, with `M` indicating male and `F` indicating female.}
#'   \item{FL}{Size of the frontal lobe in mm.}
#'   \item{RW}{Rear width in mm.}
#'   \item{CL}{Carapace length in mm.}
#'   \item{CW}{Carapace width in mm.}
#'   \item{BD}{Body depth in mm.}
#' }
#'
#' @docType data
#' @source Campbell, N.A., and Mahon, R.J. (1974) A multivariate study of variation in two species of rock crab of genus Leptograpsus. Australian Journal of Zoology, 22, 417-425.
#' @name crabs
#'
NULL

#' Australian Crabs Dataset
#'
#' Measurements on rock crabs of the genus Leptograpsus. The dataset contains 200 observations
#'  from two species of crab (blue and orange), there are 50 specimens of each sex of each species,
#'   collected on site at Fremantle, Western Australia.
#' \describe{
#'   \item{Type}{ is the group variable and has 4 groups with the combinations of species and sex (BlueMale, BlueFemale, OrangeMale and OrangeFemale)}.
#'   \item{FL}{the size of the frontal lobe length, in mm}
#'   \item{RW}{rear width, in mm}
#'   \item{CL}{length of midline of the carapace, in mm}
#'   \item{CW}{maximum width of carapace, in mm}
#'   \item{BD}{depth of the body; for females, measured after displacement of the abdomen, in mm}
#' }
#'
#' @docType data
#' @keywords datasets
#' @name crab
#' @usage data(crab)
#' @format A data frame with 200 rows and 6 variables
#' @source Campbell, N. A. & Mahon, R. J. (1974), A Multivariate Study of Variation in Two Species of Rock Crab of genus Leptograpsus, Australian Journal of Zoology 22(3), 417 - 425.
NULL

#' Fish Catch Dataset
#'
#' There are 159 fishes of 7 species are caught and measured. Altogether there are
#' 7 variables.  All the fishes are caught from the same lake(Laengelmavesi) near Tampere in Finland.
#' \describe{
#' 	 \item{Type}{ has 7 fish groups, with 35 cases of Bream, 11 cases of Parkki, 56 cases of Perch 17 cases of Pike, 20 cases of Roach, 14 cases of Smelt and 6 cases of Whitewish.}
#' 	 \item{weight}{ Weight of the fish (in grams)}
#' 	 \item{length1}{ Length from the nose to the beginning of the tail (in cm)}
#' 	 \item{length2}{ Length from the nose to the notch of the tail (in cm)}
#' 	 \item{length3}{ Length from the nose to the end of the tail (in cm)}
#' 	 \item{height}{ Maximal height as \% of Length3}
#' 	 \item{width}{ Maximal width as \% of Length3}
#' }
#'
#' @docType data
#' @keywords datasets
#' @name fishcatch
#' @usage data(fishcatch)
#' @format A data frame with 159 rows and 7 variables
#' @source \url{https://jse.amstat.org/jse_data_archive.htm}
NULL

#' Glass Dataset
#'
#' Contains measurements 214 observations of 6 types of glass; defined in terms of their oxide content.
#' \describe{
#'   \item{Type}{ has 6 types of glasses}
#'   \item{X1}{ refractive index}
#'   \item{X2}{ Sodium (unit measurement: weight percent in corresponding oxide).}
#'   \item{X3}{ Magnesium}
#'   \item{X4}{ Aluminum}
#'   \item{X5}{ Silicon}
#'   \item{X6}{ Potassium}
#'   \item{X7}{ Calcium}
#'   \item{X8}{ Barium}
#'   \item{X9}{ Iron}
#' }
#'
#' @docType data
#' @keywords datasets
#' @name glass
#' @usage data(glass)
#' @format A data frame with 214 rows and 10 variables
NULL

#' Image Dataset
#'
#' Contains  2310 observations of instances from 7 outdoor images
#' \describe{
#'   \item{Type}{ has 7 types of outdoor images, brickface, cement,  foliage, grass, path, sky, and window.}
#'   \item{X1}{ the column of the center pixel of the region}
#'   \item{X2}{ the row of the center pixel of the region. }
#'   \item{X3}{ the number of pixels in a region = 9. }
#'   \item{X4}{ the results of a line extraction algorithm that counts how many lines of length 5 (any orientation) with low contrast, less than or equal to 5, go through the region.}
#'   \item{X5}{ measure the contrast of horizontally adjacent pixels in the region. There are 6, the mean and standard deviation are given. This attribute is used as a vertical edge detector.}
#'   \item{X6}{ X5 sd}
#'   \item{X7}{ measures the contrast of vertically adjacent pixels. Used for horizontal line detection. }
#'   \item{X8}{ sd X7}
#'   \item{X9}{ the average over the region of (R + G + B)/3}
#'   \item{X10}{ the average over the region of the R value.}
#'   \item{X11}{ the average over the region of the B value.}
#'   \item{X12}{ the average over the region of the G value.}
#'   \item{X13}{ measure the excess red: (2R - (G + B))}
#'   \item{X14}{ measure the excess blue: (2B - (G + R))}
#'   \item{X15}{ measure the excess green: (2G - (R + B))}
#'   \item{X16}{ 3-d nonlinear transformation of RGB. (Algorithm can be found in Foley and VanDam, Fundamentals of Interactive Computer Graphics)}
#'   \item{X17}{ mean of X16}
#'   \item{X18}{ hue  mean}
#' }
#'
#' @docType data
#' @keywords datasets
#' @name image
#' @usage data(image)
#' @format A data frame contains 2310 observations and 19 variables
NULL

#' Leukemia Dataset
#'
#' This dataset comes from a study of gene expression in two types of acute leukemias, acute lymphoblastic leukemia (ALL) and acute myeloid leukemia (AML). Gene expression levels were measured using Affymetrix high density oligonucleotide arrays containing 6817 human genes. A dataset containing 72 observations from 3 leukemia type groups.
#' \describe{
#'   \item{Type}{ has 3 groups with 38 cases of B-cell ALL, 25 cases of AML and 9 cases of T-cell ALL}.
#'   \item{Gene1 to Gen 40}{ gene expression levels}
#' }
#' @docType data
#' @keywords datasets
#' @name leukemia
#' @usage data(leukemia)
#' @format A data frame with 72 rows and 41 variables
#' @source Dudoit, S., Fridlyand, J. and Speed, T. P. (2002). Comparison of Discrimination Methods for the Classification of Tumors Using Gene Expression Data. Journal of the American statistical Association 97 77-87.
NULL

#' Lymphoma Dataset
#'
#' Gene expression in the three most prevalent adult lymphoid malignancies: B-cell chronic lymphocytic leukemia (B-CLL), follicular lymphoma (FL), and diffuse large B-cell lym- phoma (DLBCL). Gene expression levels were measured using a specialized cDNA microarray, the Lymphochip, containing genes that are preferentially expressed in lymphoid cells or that are of known immunologic or oncologic importance. This dataset contain 80 observations from 3 lymphoma types.
#' \describe{
#'   \item{Type}{ Group variable has 3 groups with 29 cases of B-cell ALL (B-CLL), 42 cases of diffuse large B-cell lymphoma (DLBCL) and 9 cases of follicular lymphoma (FL)}.
#'   \item{Gene1 to Gen 50}{gene expression}
#'   }
#'
#' @docType data
#' @keywords datasets
#' @name lymphoma
#' @usage data(lymphoma)
#' @format A data frame with 80 rows and 51 variables
#' @source Dudoit, S., Fridlyand, J. and Speed, T. P. (2002). Comparison of Discrimination Methods for the Classification of Tumors Using Gene Ex- pression Data. Journal of the American statistical Association 97 77-87.
NULL

#' NCI60 Dataset
#'
#' cDNA microarrays were used to examine the variation in gene expression among the 60 cell lines.  The cell lines are derived from tumors with different sites of origin. This dataset contain 61 observations and 30 feature variables from 8 different tissue types.
#'
#' \describe{
#'   \item{Type}{ has 8 different tissue types, 9 cases of breast, 5 cases of central nervous system (CNS), 7 cases pf colon, 8 cases of leukemia, 8 cases of melanoma, 9 cases of  non-small-cell lung carcinoma (NSCLC), 6 cases of ovarian and 9 cases of renal.}
#'   \item{Gene1 to Gen 30}{ gene expression information}
#' }
#'
#' @docType data
#' @keywords datasets
#' @name NCI60
#' @usage data(NCI60)
#' @format A data frame with 61 rows and 31 variables
#' @source Dudoit, S., Fridlyand, J. and Speed, T. P. (2002). Comparison of Discrimination Methods for the Classification of Tumors Using Gene Expression Data. Journal of the American statistical Association 97 77-87.
NULL

#' Olive Dataset
#'
#' Contains  572 observations and 10 variables
#' \describe{
#'   \item{Region}{Three super-groups of Italy: North, South and the island of Sardinia }
#'   \item{area}{ Nine collection areas: three from North, four from South and 2 from Sardinia}
#'   \item{palmitic}{ fatty acids percent x 100}
#'   \item{palmitoleic}{ fatty acids percent x 100 }
#'   \item{stearic}{fatty acids percent x 100}
#'   \item{oleic}{fatty acids percent x 100 }
#'   \item{linoleic}{fatty acids percent x 100}
#'   \item{linolenic}{fatty acids percent x 100  }
#'   \item{arachidic}{ fatty acids percent x 100}
#'   \item{eicosenoic}{ fatty acids percent x 100}
#' }
#'
#' @docType data
#' @keywords datasets
#' @name olive
#' @usage data(olive)
#' @format A data frame contains 573 observations and 10 variables
NULL

#' Parkinson Dataset
#'
#' A dataset containing 195 observations from 2 parkinson types.
#' \describe{
#'  \item{Type}{ Group variable has 2 groups, there are 48 cases of healthy people and 147 cases with Parkinson. The feature variables are biomedical voice measures}.
#'   \item{X1}{ Average vocal fundamental frequency}
#'   \item{X2}{ Maximum vocal fundamental frequency}
#'   \item{X3}{ Minimum vocal fundamental frequency}
#'   \item{X4}{ MDVP:Jitter(\%) measures of variation in fundamental frequency}
#'   \item{X5}{ MDVP:Jitter(Abs) measures of variation in fundamental frequency}
#'   \item{X6}{ MDVP:RAP measures of variation in fundamental frequency}
#'   \item{X7}{ MDVP:PPQ measures of variation in fundamental frequency}
#'   \item{X8}{ Jitter:DDP measures of variation in fundamental frequency}
#'   \item{X9}{ MDVP:Shimmer measures of variation in amplitude}
#'   \item{X10}{ MDVP:Shimmer(dB) measures of variation in amplitude}
#'   \item{X11}{ Shimmer:APQ3 measures of variation in amplitude}
#'   \item{X12}{ Shimmer:APQ5 measures of variation in amplitude}
#'   \item{X13}{ MDVP:APQ measures of variation in amplitude}
#'   \item{X14}{ Shimmer:DDA measures of variation in amplitude}
#'   \item{X15}{ NHR measures of ratio of noise to tonal components in the voice}
#'   \item{X16}{ HNR measures of ratio of noise to tonal components in the voice}
#'   \item{X17}{ RPDE nonlinear dynamical complexity measures}
#'   \item{X18}{ D2 nonlinear dynamical complexity measures}
#'   \item{X19}{ DFA - Signal fractal scaling exponent}
#'   \item{X20}{ spread1 Nonlinear measures of fundamental frequency variation}
#'   \item{X21}{ spread2 Nonlinear measures of fundamental frequency variation}
#'   \item{X22}{ PPE Nonlinear measures of fundamental frequency variation}
#' }
#'
#' @docType data
#' @keywords datasets
#' @name parkinson
#' @usage data(parkinson)
#' @format A data frame with 195 rows and 23 variables
#' @source \url{https://archive.ics.uci.edu/ml/datasets/Parkinsons}
NULL

#' California Housing Dataset
#'
#' Median house values and demographic/geographic predictors for 20,433 California
#' block groups, derived from the 1990 US Census. Classic regression benchmark.
#' Originally released by Pace & Barry (1997); this version drops the categorical
#' `ocean_proximity` column (to keep the bundled schema all-numeric, matching the
#' rest of the package's datasets) and removes rows with missing `total_bedrooms`.
#' The regression target, `median_house_value`, is the last column.
#'
#' \describe{
#'   \item{longitude}{Longitude of the block group (degrees, negative = west).}
#'   \item{latitude}{Latitude of the block group (degrees).}
#'   \item{housing_median_age}{Median age of houses in the block group (years).}
#'   \item{total_rooms}{Total number of rooms.}
#'   \item{total_bedrooms}{Total number of bedrooms.}
#'   \item{population}{Block group population.}
#'   \item{households}{Number of households.}
#'   \item{median_income}{Median household income (tens of thousands of USD).}
#'   \item{median_house_value}{Regression target: median house value (USD).}
#' }
#'
#' @docType data
#' @keywords datasets
#' @name california_housing
#' @usage data(california_housing)
#' @format A data frame with 20,433 rows and 9 variables.
#' @source Pace, R.K. and Barry, R. (1997) Sparse spatial autoregressions. Statistics & Probability Letters, 33, 291-297.
NULL

#' Wine Dataset
#'
#' A dataset containing 178 observations from 3 wine grown cultivares in Italy.
#'
#' \describe{
#'   \item{Type}{ Group variable has 3 groups that are 3 different wine grown cultivars in Italy. }
#'   \item{X1 to X13}{Check vbles}
#' }
#'
#' @docType data
#' @keywords datasets
#' @name wine
#' @usage data(wine)
#' @format A data frame with 178 rows and 14 variables
NULL

#' Majority-vote leaf strategy.
#'
#' Creates a leaf strategy that assigns the majority group label as the
#' leaf prediction. When groups are tied, the smallest label wins.
#' This is the default leaf strategy for classification.
#'
#' @return A \code{leaf_strategy} object.
#' @seealso \code{\link{pptr}}, \code{\link{pprf}}
#' @examples
#' leaf_majority_vote()
#'
#' @export
leaf_majority_vote <- function() {
  structure(list(name = "majority_vote", display_name = "Majority vote"), class = "leaf_strategy")
}

#' Mean-response leaf strategy.
#'
#' Creates a leaf strategy that predicts the mean of the continuous response
#' values for the observations in the leaf. Used for regression trees; requires
#' a numeric response.
#'
#' @return A \code{leaf_strategy} object.
#' @seealso \code{\link{pptr}}, \code{\link{pprf}}, \code{\link{leaf_majority_vote}}
#' @examples
#' leaf_mean_response()
#'
#' @export
leaf_mean_response <- function() {
  structure(list(name = "mean_response", display_name = "Mean response"), class = "leaf_strategy")
}

describe("leaf_majority_vote", {
  it("creates a leaf_strategy object", {
    s <- leaf_majority_vote()
    expect_s3_class(s, "leaf_strategy")
    expect_equal(s$name, "majority_vote")
  })
})

describe("leaf_mean_response", {
  it("creates a leaf_strategy object", {
    s <- leaf_mean_response()
    expect_s3_class(s, "leaf_strategy")
    expect_equal(s$name, "mean_response")
  })
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

ppforest2_has_openmp <- function() {
    .Call('_ppforest2_ppforest2_has_openmp', PACKAGE = 'ppforest2')
}

ppforest2_proportion_to_count <- function(p, total) {
    .Call('_ppforest2_ppforest2_proportion_to_count', PACKAGE = 'ppforest2', p, total)
}

ppforest2_train <- function(spec, x, y) {
    .Call('_ppforest2_ppforest2_train', PACKAGE = 'ppforest2', spec, x, y)
}

ppforest2_predict_tree <- function(tree, data) {
    .Call('_ppforest2_ppforest2_predict_tree', PACKAGE = 'ppforest2', tree, data)
}

ppforest2_predict_forest <- function(forest, data) {
    .Call('_ppforest2_ppforest2_predict_forest', PACKAGE = 'ppforest2', forest, data)
}

ppforest2_predict_tree_prob <- function(tree, data) {
    .Call('_ppforest2_ppforest2_predict_tree_prob', PACKAGE = 'ppforest2', tree, data)
}

ppforest2_predict_forest_prob <- function(forest, data) {
    .Call('_ppforest2_ppforest2_predict_forest_prob', PACKAGE = 'ppforest2', forest, data)
}

ppforest2_vi_projections_tree <- function(tree, n_vars, scale) {
    .Call('_ppforest2_ppforest2_vi_projections_tree', PACKAGE = 'ppforest2', tree, n_vars, scale)
}

ppforest2_vi_projections_forest <- function(forest, n_vars, scale) {
    .Call('_ppforest2_ppforest2_vi_projections_forest', PACKAGE = 'ppforest2', forest, n_vars, scale)
}

ppforest2_vi_weighted_forest <- function(forest, x, y, scale) {
    .Call('_ppforest2_ppforest2_vi_weighted_forest', PACKAGE = 'ppforest2', forest, x, y, scale)
}

ppforest2_vi_permuted_forest <- function(forest, x, y, seed) {
    .Call('_ppforest2_ppforest2_vi_permuted_forest', PACKAGE = 'ppforest2', forest, x, y, seed)
}

ppforest2_oob_error <- function(forest, x, y) {
    .Call('_ppforest2_ppforest2_oob_error', PACKAGE = 'ppforest2', forest, x, y)
}

ppforest2_oob_predict <- function(forest, x) {
    .Call('_ppforest2_ppforest2_oob_predict', PACKAGE = 'ppforest2', forest, x)
}

ppforest2_tree_node_data <- function(tree, x, y) {
    .Call('_ppforest2_ppforest2_tree_node_data', PACKAGE = 'ppforest2', tree, x, y)
}

ppforest2_boundary_segments <- function(tree, var_indices, fixed_values, x_min, x_max, y_min, y_max) {
    .Call('_ppforest2_ppforest2_boundary_segments', PACKAGE = 'ppforest2', tree, var_indices, fixed_values, x_min, x_max, y_min, y_max)
}

ppforest2_decision_regions <- function(tree, var_indices, fixed_values, x_min, x_max, y_min, y_max) {
    .Call('_ppforest2_ppforest2_decision_regions', PACKAGE = 'ppforest2', tree, var_indices, fixed_values, x_min, x_max, y_min, y_max)
}

ppforest2_tree_layout <- function(tree) {
    .Call('_ppforest2_ppforest2_tree_layout', PACKAGE = 'ppforest2', tree)
}

ppforest2_save_model_json <- function(model, groups, include_metrics, x, y, feature_names) {
    .Call('_ppforest2_ppforest2_save_model_json', PACKAGE = 'ppforest2', model, groups, include_metrics, x, y, feature_names)
}

ppforest2_load_model_json <- function(path) {
    .Call('_ppforest2_ppforest2_load_model_json', PACKAGE = 'ppforest2', path)
}

Sys.setenv(R_TESTS = "")
Sys.setenv(OMP_THREAD_LIMIT = "1")
Sys.setenv(OMP_NUM_THREADS = "1")

library(testthat)
library(ppforest2)

skip_if_not_installed("ggplot2")

describe("plot.pptr boundaries", {
  it("returns a ggplot object for 2-feature boundary plot", {
    model <- pptr(
      x = iris[, c("Sepal.Length", "Sepal.Width")],
      y = iris$Species,
      seed = 0
    )
    p <- plot(model, type = "boundaries")
    expect_s3_class(p, "ggplot")
  })

  it("returns a gtable for 3-feature boundary plot (pairwise with stripped panels)", {
    model <- pptr(
      x = iris[, c("Sepal.Length", "Sepal.Width", "Petal.Length")],
      y = iris$Species,
      seed = 0
    )
    p <- plot(model, type = "boundaries")
    expect_s3_class(p, "gtable")
  })

  it("returns a gtable for pairwise boundary plot with 4 features", {
    model <- pptr(Species ~ ., data = iris, seed = 0)
    p <- plot(model, type = "boundaries")
    expect_s3_class(p, "gtable")
  })

  it("boundary plot uses C++ segment computation", {
    model <- pptr(
      x = iris[, c("Sepal.Length", "Sepal.Width")],
      y = iris$Species,
      seed = 0
    )
    segs <- ppforest2:::ppforest2_boundary_segments(
      model, c(0L, 1L), numeric(0),
      4.0, 8.0, 2.0, 4.5
    )
    expect_s3_class(segs, "data.frame")
    expect_true(nrow(segs) > 0)
    expect_true(all(c("x_start", "y_start", "x_end", "y_end", "depth") %in% names(segs)))
  })

  it("decision regions returns polygon data", {
    model <- pptr(
      x = iris[, c("Sepal.Length", "Sepal.Width")],
      y = iris$Species,
      seed = 0
    )
    regions <- ppforest2:::ppforest2_decision_regions(
      model, c(0L, 1L), numeric(0),
      4.0, 8.0, 2.0, 4.5
    )
    expect_true(is.list(regions))
    expect_true(length(regions) > 0)
    r <- regions[[1]]
    expect_true(all(c("x", "y", "group") %in% names(r)))
    expect_true(length(r$x) >= 3)
    expect_equal(length(r$x), length(r$y))
  })
})

describe("plot.pprf boundaries", {
  it("returns a ggplot object for 2-feature boundary plot", {
    model <- pprf(
      x = iris[, c("Sepal.Length", "Sepal.Width")],
      y = iris$Species,
      size = 5, seed = 0, threads = 1
    )
    p <- plot(model, type = "boundaries", tree_index = 1)
    expect_s3_class(p, "ggplot")
  })
})

describe("plot.pptr boundaries snapshots", {
  skip_if_not_installed("vdiffr")
  skip_on_ci()  # vdiffr SVGs aren't byte-identical across platform LAPACK/BLAS

  it("pptr-boundaries-2d", {
    model <- pptr(
      x = iris[, c("Sepal.Length", "Sepal.Width")],
      y = iris$Species,
      seed = 0
    )
    vdiffr::expect_doppelganger("pptr-boundaries-2d", plot(model, type = "boundaries"))
  })
})

# ======================================================================
# Tree structure plot
# ======================================================================

#' Compute histogram bar rectangles for one internal (condition) node.
#'
#' Bins the projected values by group, then maps bin coordinates from
#' projection space to the node's rectangle in the tree layout canvas.
#' This is NOT a standard ggplot2 \code{geom_histogram} — it produces raw
#' rectangle coordinates (xmin/xmax/ymin/ymax) that are drawn with
#' \code{geom_rect} so each node's histogram is embedded inside its box.
#'
#' @param nd A node data list from \code{ppforest2_tree_node_data()},
#'   with \code{$projected} (numeric), \code{$groups} (integer),
#'   and \code{$cutpoint} (numeric).
#' @param cx,cy Center position of the node on the layout canvas.
#' @param node_w,node_h Width and height of the node rectangle.
#' @param group_labels Character vector of group names (model$groups).
#' @param n_bins Number of histogram bins (default 15).
#' @return List with:
#'   \describe{
#'     \item{bars}{data.frame with xmin, xmax, ymin, ymax, group columns}
#'     \item{thr_x}{x-coordinate of the cutpoint line in canvas space}
#'     \item{thr_y_min, thr_y_max}{y-extent of the cutpoint line}
#'     \item{ticks}{data.frame with x, y, label for axis tick labels}
#'   }
#'   or NULL if no non-empty bins exist.
#' @noRd
compute_histogram_bars <- function(nd, cx, cy, node_w, node_h, group_labels, n_bins = 15) {
  vals <- nd$projected
  cls <- group_labels[nd$groups]
  cutpoint <- nd$cutpoint

  breaks <- seq(min(vals), max(vals), length.out = n_bins + 1)
  unique_groups <- unique(cls)

  bars <- list()
  for (cl in unique_groups) {
    cl_vals <- vals[cls == cl]
    h <- graphics::hist(cl_vals, breaks = breaks, plot = FALSE)
    if (any(h$counts > 0)) {
      bars[[length(bars) + 1L]] <- data.frame(
        bin_left  = h$breaks[-length(h$breaks)],
        bin_right = h$breaks[-1],
        count     = h$counts,
        group     = cl,
        stringsAsFactors = FALSE
      )
    }
  }

  if (length(bars) == 0) return(NULL)

  bar_df <- do.call(rbind, bars)
  bar_df <- bar_df[bar_df$count > 0, , drop = FALSE]

  if (nrow(bar_df) == 0) return(NULL)

  # Map projected-value-space to node coordinate space
  val_min <- min(breaks)
  val_max <- max(breaks)
  val_range <- val_max - val_min
  if (val_range == 0) val_range <- 1

  margin <- 0.05
  draw_w <- node_w * (1 - 2 * margin)
  draw_h <- node_h * (1 - 2 * margin)

  max_count <- max(bar_df$count)

  bar_df$xmin <- cx - node_w / 2 + node_w * margin + (bar_df$bin_left - val_min) / val_range * draw_w
  bar_df$xmax <- cx - node_w / 2 + node_w * margin + (bar_df$bin_right - val_min) / val_range * draw_w
  bar_df$ymin <- cy - node_h / 2 + node_h * margin
  bar_df$ymax <- bar_df$ymin + bar_df$count / max_count * draw_h

  # Threshold line in data coordinates
  thr_x <- cx - node_w / 2 + node_w * margin + (cutpoint - val_min) / val_range * draw_w

  # Axis tick data
  tick_y <- cy - node_h / 2
  tick_labels <- data.frame(
    x = c(cx - node_w / 2 + node_w * margin, cx + node_w / 2 - node_w * margin),
    y = tick_y,
    label = formatC(c(val_min, val_max), format = "g", digits = 3),
    stringsAsFactors = FALSE
  )

  list(bars = bar_df, thr_x = thr_x, thr_y_min = cy - node_h / 2 + node_h * margin, thr_y_max = cy + node_h / 2 - node_h * margin, ticks = tick_labels)
}

#' Format a projector vector as a compact equation string.
#'
#' Shows the top-k coefficients by absolute magnitude as a sum like
#' \code{"0.71\u00b7SL \u2212 0.29\u00b7PW"}.  If terms were omitted,
#' appends \code{"+ \u2026"}.
#'
#' @param projector Numeric vector of projector coefficients.
#' @param var_names Character vector of variable names (same length as projector).
#' @param max_terms Maximum number of terms to display.
#' @return A single character string.
#' @noRd
format_projector <- function(projector, var_names, max_terms = 4L) {
  ord <- order(abs(projector), decreasing = TRUE)
  n_nonzero <- sum(projector != 0)
  n_show <- min(max_terms, n_nonzero)

  if (n_show == 0) return("0")

  top_idx <- ord[seq_len(n_show)]
  coefs <- projector[top_idx]
  names_top <- var_names[top_idx]

  parts <- character(n_show)
  for (i in seq_len(n_show)) {
    coef_str <- formatC(abs(coefs[i]), format = "f", digits = 2)
    # Drop leading zero: "0.71" -> ".71"
    coef_str <- sub("^0\\.", ".", coef_str)
    term <- paste0(coef_str, "\u00b7", names_top[i])

    if (i == 1L) {
      parts[i] <- if (coefs[i] < 0) paste0("\u2212", term) else term
    } else {
      parts[i] <- if (coefs[i] < 0) paste0(" \u2212 ", term) else paste0(" + ", term)
    }
  }

  result <- paste0(parts, collapse = "")

  if (n_nonzero > n_show) {
    result <- paste0(result, " + \u2026")
  }

  result
}

#' Render the tree structure as a ggplot2 plot.
#'
#' Draws a tree diagram on a 2D canvas with:
#' \itemize{
#'   \item Edges connecting parent to child nodes (grey segments with labels)
#'   \item Internal nodes as white rectangles containing stacked group
#'     histograms of the projected values at that split, with a dashed
#'     cutpoint line
#'   \item A compact projector equation below each internal node showing
#'     the top coefficients (controlled by \code{max_terms})
#'   \item Leaf nodes as group-colored rectangles with the predicted group
#'     label
#' }
#'
#' Layout positions come from the C++ \code{compute_tree_layout()} (pre-order,
#' left-aligned).  Per-node projection data comes from
#' \code{ppforest2_tree_node_data()} which routes all training observations
#' through the tree.
#'
#' @param model A pptr model with \code{$root}, \code{$x}, \code{$y},
#'   \code{$groups}.
#' @param max_terms Maximum number of projector terms to display per node
#'   (default 3).  Only the largest-magnitude coefficients are shown.
#' @param ... Currently unused.
#' @return A ggplot2 object (theme_void, no axes).
#' @noRd
plot_tree_structure <- function(model, max_terms = 3L, ...) {
  nodes <- ppforest2_tree_node_data(model, model$x, model$y)
  group_labels <- model$groups
  group_colors <- get_group_colors(group_labels)
  var_names <- get_variable_names(model)

  node_w <- ppforest2_node_w()
  node_h <- ppforest2_node_h()
  leaf_w <- ppforest2_leaf_w()
  leaf_h <- ppforest2_leaf_h()

  layout <- ppforest2_tree_layout(model)
  node_df <- layout$nodes
  # C++ uses 0-indexed node_idx, convert to 1-indexed for R
  node_df$node_idx <- node_df$node_idx + 1L
  edge_df <- layout$edges

  # Collect all histogram bars, cutpoint lines, and axis ticks
  all_bars <- list()
  all_thr_segs <- list()
  all_ticks <- list()
  node_bg_df <- list()
  leaf_bg_df <- list()
  leaf_label_df <- list()
  proj_label_df <- list()

  for (i in seq_len(nrow(node_df))) {
    row <- node_df[i, ]
    nd <- nodes[[row$node_idx]]

    if (nd$is_leaf) {
      predicted <- group_labels[nd$value]
      leaf_bg_df[[length(leaf_bg_df) + 1L]] <- data.frame(
        xmin = row$x - leaf_w / 2, xmax = row$x + leaf_w / 2,
        ymin = row$y - leaf_h / 2, ymax = row$y + leaf_h / 2,
        group = predicted,
        stringsAsFactors = FALSE
      )
      leaf_label_df[[length(leaf_label_df) + 1L]] <- data.frame(
        x = row$x, y = row$y, label = predicted,
        stringsAsFactors = FALSE
      )
    } else {
      node_bg_df[[length(node_bg_df) + 1L]] <- data.frame(
        xmin = row$x - node_w / 2, xmax = row$x + node_w / 2,
        ymin = row$y - node_h / 2, ymax = row$y + node_h / 2,
        stringsAsFactors = FALSE
      )

      hist_data <- compute_histogram_bars(nd, row$x, row$y, node_w, node_h, group_labels)

      if (!is.null(hist_data)) {
        all_bars[[length(all_bars) + 1L]] <- hist_data$bars
        all_thr_segs[[length(all_thr_segs) + 1L]] <- data.frame(
          x = hist_data$thr_x, xend = hist_data$thr_x,
          y = hist_data$thr_y_min, yend = hist_data$thr_y_max,
          stringsAsFactors = FALSE
        )
        all_ticks[[length(all_ticks) + 1L]] <- hist_data$ticks
      }

      proj_label_df[[length(proj_label_df) + 1L]] <- data.frame(
        x = row$x,
        y = row$y - node_h / 2 - 0.15,
        label = format_projector(nd$projector, var_names, max_terms),
        stringsAsFactors = FALSE
      )
    }
  }

  bar_df <- if (length(all_bars) > 0) do.call(rbind, all_bars) else NULL
  thr_df <- if (length(all_thr_segs) > 0) do.call(rbind, all_thr_segs) else NULL
  tick_df <- if (length(all_ticks) > 0) do.call(rbind, all_ticks) else NULL
  node_bg <- if (length(node_bg_df) > 0) do.call(rbind, node_bg_df) else NULL
  leaf_bg <- if (length(leaf_bg_df) > 0) do.call(rbind, leaf_bg_df) else NULL
  leaf_labels <- if (length(leaf_label_df) > 0) do.call(rbind, leaf_label_df) else NULL
  proj_labels <- if (length(proj_label_df) > 0) do.call(rbind, proj_label_df) else NULL

  p <- ggplot2::ggplot() +
    ggplot2::theme_void()

  # Edges
  if (nrow(edge_df) > 0) {
    p <- p + ggplot2::geom_segment(
      data = edge_df,
      ggplot2::aes(x = from_x, y = from_y, xend = to_x, yend = to_y),
      color = ppforest2_col_edge(), linewidth = ppforest2_lw_light()
    )
    # Edge labels at midpoint
    edge_label_df <- data.frame(
      x = (edge_df$from_x + edge_df$to_x) / 2,
      y = (edge_df$from_y + edge_df$to_y) / 2,
      label = edge_df$edge_label,
      stringsAsFactors = FALSE
    )
    p <- p + ggplot2::geom_label(
      data = edge_label_df,
      ggplot2::aes(x = x, y = y, label = label),
      size = 2.5, label.size = 0, fill = "white", alpha = 0.8,
      label.padding = ggplot2::unit(0.15, "lines")
    )
  }

  # Node backgrounds
  if (!is.null(node_bg)) {
    p <- p + ggplot2::geom_rect(
      data = node_bg,
      ggplot2::aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax),
      fill = "white", color = ppforest2_col_border(), linewidth = ppforest2_lw_light()
    )
  }

  # Histogram bars
  if (!is.null(bar_df)) {
    p <- p + ggplot2::geom_rect(
      data = bar_df,
      ggplot2::aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax,
                   fill = group),
      alpha = ppforest2_alpha_hist(), color = NA
    )
  }

  # Threshold lines
  if (!is.null(thr_df)) {
    p <- p + ggplot2::geom_segment(
      data = thr_df,
      ggplot2::aes(x = x, y = y, xend = xend, yend = yend),
      linetype = "dashed", color = ppforest2_col_cutpoint(), linewidth = ppforest2_lw_medium()
    )
  }

  # Axis ticks
  if (!is.null(tick_df)) {
    p <- p + ggplot2::geom_text(
      data = tick_df,
      ggplot2::aes(x = x, y = y, label = label),
      size = 1.8, vjust = 1.5, color = ppforest2_col_tick()
    )
  }

  # Leaf backgrounds
  if (!is.null(leaf_bg)) {
    p <- p + ggplot2::geom_rect(
      data = leaf_bg,
      ggplot2::aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax,
                   fill = group),
      alpha = ppforest2_alpha_leaf(), color = ppforest2_col_border(), linewidth = ppforest2_lw_light()
    )
  }

  # Leaf labels
  if (!is.null(leaf_labels)) {
    p <- p + ggplot2::geom_text(
      data = leaf_labels,
      ggplot2::aes(x = x, y = y, label = label),
      size = 3, fontface = "bold"
    )
  }

  # Projector labels below internal nodes
  if (!is.null(proj_labels)) {
    p <- p + ggplot2::geom_text(
      data = proj_labels,
      ggplot2::aes(x = x, y = y, label = label),
      size = 2.0, vjust = 1, color = ppforest2_col_tick()
    )
  }

  # Color scale, title, and legend
  p <- p +
    ggplot2::scale_fill_manual(values = group_colors, name = "Class") +
    ggplot2::ggtitle("Tree Structure") +
    ggplot2::coord_cartesian(
      xlim = c(min(node_df$x) - node_w / 2 - 0.2,
               max(node_df$x) + node_w / 2 + 0.2),
      ylim = c(min(node_df$y) - node_h / 2 - 0.3,
               max(node_df$y) + node_h / 2 + 0.1)
    ) +
    ggplot2::theme(
      plot.title       = ggplot2::element_text(hjust = 0, margin = ggplot2::margin(b = 5)),
      plot.title.position = "plot",
      legend.position  = "top",
      legend.justification = "left",
      legend.margin    = ggplot2::margin(t = 0, b = 5),
      plot.margin      = ggplot2::margin(2, 5, 2, 5)
    )

  p
}

# ======================================================================
# Variable importance plot
# ======================================================================

#' Display labels for each variable importance metric.
#'
#' Keys must match names in \code{model$vi} (set during training/forest
#' construction).
#' @noRd
metric_labels <- c(
  projections = "Projection",
  weighted    = "Weighted",
  permuted    = "Permuted"
)

#' Render a horizontal bar chart of variable importance.
#'
#' Supports multiple importance metrics (projection, weighted, permuted).
#' When \code{metric} is \code{NULL} and multiple metrics are available,
#' bars are colour-coded and dodged.  When a single metric is selected
#' (or only one is available), bars use the neutral
#' \code{ppforest2_col_bar} colour.  Variables are ordered by the
#' plotted metric's values (descending).
#'
#' @param model A pptr or pprf model with \code{$vi} (list of named
#'   numeric vectors) and \code{$x} (for variable names).
#' @param metric Character string selecting a single importance metric
#'   to plot: \code{"projections"}, \code{"weighted"}, or
#'   \code{"permuted"}.  \code{NULL} (default) shows all available
#'   metrics together.
#' @param ... Currently unused.
#' @return A ggplot2 object.
#' @noRd
plot_importance <- function(model, metric = NULL, ...) {
  vnames <- get_variable_names(model)

  # For forests, weighted and permuted are lazy accessors; for trees only
  # projections is available. Build the set of available metrics per the
  # model class, not field presence on `$vi`.
  is_forest <- inherits(model, "pprf")
  available <- if (is_forest) c("projections", "weighted", "permuted") else "projections"

  get_values <- function(m) {
    switch(m,
      projections = model$vi$projections,
      weighted    = weighted_importance(model),
      permuted    = permuted_importance(model)
    )
  }

  if (!is.null(metric)) {
    metric <- match.arg(metric, names(metric_labels))
    if (!(metric %in% available)) {
      stop(
        "Metric '", metric, "' is not available for this model. ",
        "Available: ", paste(available, collapse = ", "),
        call. = FALSE
      )
    }
    available <- metric
  }

  dfs <- list()
  for (m in available) {
    values <- get_values(m)
    dfs[[m]] <- data.frame(
      variable   = vnames,
      importance = values,
      metric     = unname(metric_labels[m]),
      stringsAsFactors = FALSE
    )
  }

  df <- do.call(rbind, dfs)

  # Order variables by first (or only) metric's values
  first_values <- get_values(available[1])
  df$variable <- factor(df$variable, levels = vnames[order(first_values)])
  df$metric <- factor(df$metric, levels = unname(metric_labels[available]))

  if (length(available) > 1) {
    p <- ggplot2::ggplot(df, ggplot2::aes(x = importance, y = variable, fill = metric)) +
      ggplot2::geom_col(position = "dodge") +
      ggplot2::labs(
        title = "Variable Importance",
        x     = "Importance",
        y     = NULL,
        fill  = "Metric"
      ) +
      ppforest2_theme() +
      ggplot2::theme(panel.grid.major.y = ggplot2::element_blank())
  } else {
    title <- paste0("Variable Importance (", metric_labels[available], ")")
    p <- ggplot2::ggplot(df, ggplot2::aes(x = importance, y = variable)) +
      ggplot2::geom_col(fill = ppforest2_col_bar()) +
      ggplot2::labs(
        title = title,
        x     = "Importance",
        y     = NULL
      ) +
      ppforest2_theme() +
      ggplot2::theme(panel.grid.major.y = ggplot2::element_blank())
  }

  p
}

#' Render a grid of individual importance plots, one per metric.
#'
#' Used by \code{plot.pprf()} to show all importance metrics side by side,
#' each ordered independently.
#'
#' @param model A pprf model.
#' @param ... Currently unused.
#' @return A patchwork object (ggplot-compatible, works with ggsave).
#' @noRd
plot_importance_grid <- function(model, ...) {
  check_patchwork()
  available <- intersect(names(metric_labels), names(model$vi))

  plots <- lapply(available, function(m) {
    # Use short subtitle per panel instead of full title
    plot_importance(model, metric = m) +
      ggplot2::labs(title = metric_labels[m]) +
      ggplot2::theme(plot.margin = ggplot2::margin(5, 10, 5, 10))
  })

  Reduce(`+`, plots) +
    patchwork::plot_layout(nrow = 1) +
    patchwork::plot_annotation(title = "Variable Importance")
}

#' Largest-gap binarization strategy.
#'
#' Creates a binarization strategy that reduces multiclass nodes to binary
#' by projecting group means and splitting at the largest gap. Default
#' for classification specs.
#'
#' @return A \code{binarize_strategy} object.
#' @seealso \code{\link{pptr}}, \code{\link{pprf}}, \code{\link{binarize_disabled}}
#' @examples
#' binarize_largest_gap()
#'
#' @export
binarize_largest_gap <- function() {
  structure(list(name = "largest_gap", display_name = "Largest gap"), class = "binarize_strategy")
}

#' Disabled binarization strategy (placeholder).
#'
#' Placeholder binarizer for specs where binarization never fires —
#' notably regression, where `grouping_by_cutpoint()` always produces
#' a 2-group partition at each node. Selecting \code{binarize_disabled()}
#' documents that intent explicitly; if binarization is ever invoked at
#' runtime with this strategy configured, training aborts with a clear
#' error rather than silently passing through. Default for regression
#' specs.
#'
#' @return A \code{binarize_strategy} object.
#' @seealso \code{\link{binarize_largest_gap}}
#' @examples
#' binarize_disabled()
#'
#' @export
binarize_disabled <- function() {
  structure(list(name = "disabled", display_name = "Disabled"), class = "binarize_strategy")
}

describe("binarize_largest_gap", {
  it("creates a binarize_strategy object", {
    s <- binarize_largest_gap()
    expect_s3_class(s, "binarize_strategy")
    expect_equal(s$name, "largest_gap")
  })
})

describe("binarize_disabled", {
  it("creates a binarize_strategy object", {
    s <- binarize_disabled()
    expect_s3_class(s, "binarize_strategy")
    expect_equal(s$name, "disabled")
  })
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----inspect------------------------------------------------------------------
library(ppforest2)

pp_pda(0.5)
vars_uniform(n_vars = 2)
cutpoint_mean_of_means()

## ----eval = FALSE-------------------------------------------------------------
# #' My custom projection pursuit strategy.
# #'
# #' @param alpha A tuning parameter.
# #' @return A \code{pp_strategy} object.
# #' @export
# pp_my_method <- function(alpha = 1.0) {
#   if (!is.numeric(alpha) || length(alpha) != 1)
#     stop("`alpha` must be a single number.")
# 
#   structure(
#     list(name = "my_method", display_name = "My method", alpha = alpha),
#     class = "pp_strategy"
#   )
# }

## ----eval = FALSE-------------------------------------------------------------
# # Single tree
# tree <- pptr(Species ~ ., data = iris, pp = pp_my_method(alpha = 0.5))
# 
# # Forest
# forest <- pprf(Species ~ ., data = iris, pp = pp_my_method(alpha = 0.5), vars = vars_uniform(n_vars = 2))
# 
# # Summary shows the strategy
# summary(tree)

---
title: "Introduction to ppforest2"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to ppforest2}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
# Cap OpenMP to 2 threads so forest training in this vignette never uses more
# than two cores when the vignette is rebuilt under R CMD check (CRAN policy).
Sys.setenv(OMP_NUM_THREADS = "2")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

ppforest2 builds oblique decision trees and random forests using projection pursuit. Instead of splitting on a single variable at each node, it finds a linear combination of variables that best separates the groups.

## Single tree

Train a projection-pursuit tree on the iris dataset:

```{r tree}
library(ppforest2)

tree <- pptr(Species ~ ., data = iris, seed = 0)
tree
```

The tree splits on linear projections of the features. Use `summary()` to see variable importance:

```{r tree-summary}
summary(tree)
```

Predict new observations:

```{r tree-predict}
preds <- predict(tree, iris)
table(predicted = preds, actual = iris$Species)
```

## Random forest

Train a forest of 100 trees, each considering 2 randomly chosen variables at each split:

```{r forest}
forest <- pprf(Species ~ ., data = iris, size = 100, n_vars = 2, seed = 0)
summary(forest)
```

The summary shows the OOB (out-of-bag) error estimate and three variable importance measures:

- **Projection**: average absolute projection coefficient across all splits
- **Weighted**: projection importance weighted by the number of observations routed through each node
- **Permuted**: decrease in OOB accuracy when each variable is permuted

Predict with class probabilities:

```{r forest-predict}
probs <- predict(forest, iris[1:5, ], type = "prob")
probs
```

## Visualization

ppforest2 provides several plot types via `plot()` (requires ggplot2):

```{r plot-structure, eval = requireNamespace("ggplot2", quietly = TRUE)}
# Tree structure with projected data histograms at each node
plot(tree, type = "structure")
```

```{r plot-importance, eval = requireNamespace("ggplot2", quietly = TRUE)}
# Variable importance bar chart
plot(tree, type = "importance")
```

```{r plot-projection, eval = requireNamespace("ggplot2", quietly = TRUE)}
# Data projected onto the first split's projection vector
plot(tree, type = "projection")
```

```{r plot-boundaries, eval = requireNamespace("ggplot2", quietly = TRUE)}
# Decision boundaries for two selected variables
plot(tree, type = "boundaries")
```

For forests, variable importance is the only global visualization available. Individual trees can be plotted by specifying a `tree_index`.

```{r plot-forest, eval = requireNamespace("ggplot2", quietly = TRUE)}
plot(forest)
```

```{r plot-forest-tree, eval = requireNamespace("ggplot2", quietly = TRUE)}
plot(forest, type = "structure", tree_index = 1)
```



## Regularization with PDA

Set `lambda > 0` to use Penalized Discriminant Analysis instead of LDA. This can help when features are highly correlated:

```{r pda}
tree_pda <- pptr(Species ~ ., data = iris, lambda = 0.5, seed = 0)
summary(tree_pda)
```

## Explicit strategy selection

For more control, you can pass strategy objects directly instead of using the shortcut parameters (`lambda`, `n_vars`, `p_vars`). This is equivalent but makes the strategy choice explicit:

```{r strategies}
# These two calls produce identical results:
forest_shortcut <- pprf(Species ~ ., data = iris, size = 10, lambda = 0.5, n_vars = 2, seed = 0)

forest_explicit <- pprf(Species ~ ., data = iris, size = 10, pp = pp_pda(0.5), vars = vars_uniform(n_vars = 2), seed = 0)

all.equal(predict(forest_shortcut, iris), predict(forest_explicit, iris))
```

Available strategy constructors:

- `pp_pda(lambda)` — PDA projection pursuit (`lambda = 0` for LDA)
- `vars_uniform(n_vars)` or `vars_uniform(p_vars)` — random variable selection
- `vars_all()` — use all variables (default for single trees)
- `cutpoint_mean_of_means()` — midpoint split rule (default)

Strategy objects can also be passed as engine arguments in parsnip:

```{r strategies-parsnip, eval = requireNamespace("parsnip", quietly = TRUE)}
library(parsnip)

spec <- pp_rand_forest(trees = 10) |>
  set_engine("ppforest2", pp = pp_pda(0.5), vars = vars_uniform(n_vars = 2)) |>
  set_mode("classification")

fit <- spec |> fit(Species ~ ., data = iris)
predict(fit, iris[1:5, ])
```

## Tidymodels integration

ppforest2 integrates with the tidymodels ecosystem via parsnip:

```{r parsnip, eval = requireNamespace("parsnip", quietly = TRUE)}
library(parsnip)

# Single tree
spec <- pp_tree(penalty = 0.5) |>
  set_engine("ppforest2") |>
  set_mode("classification")

fit <- spec |> fit(Species ~ ., data = iris)
predict(fit, iris[1:5, ])
```

```{r parsnip-forest, eval = requireNamespace("parsnip", quietly = TRUE)}
# Random forest
spec <- pp_rand_forest(trees = 50, mtry = 2, penalty = 0.5) |>
  set_engine("ppforest2") |>
  set_mode("classification")

fit <- spec |> fit(Species ~ ., data = iris)
predict(fit, iris[1:5, ], type = "prob")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{glass}
\alias{glass}
\title{Glass Dataset}
\format{
A data frame with 214 rows and 10 variables
}
\usage{
data(glass)
}
\description{
Contains measurements 214 observations of 6 types of glass; defined in terms of their oxide content.
\describe{
  \item{Type}{ has 6 types of glasses}
  \item{X1}{ refractive index}
  \item{X2}{ Sodium (unit measurement: weight percent in corresponding oxide).}
  \item{X3}{ Magnesium}
  \item{X4}{ Aluminum}
  \item{X5}{ Silicon}
  \item{X6}{ Potassium}
  \item{X7}{ Calcium}
  \item{X8}{ Barium}
  \item{X9}{ Iron}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grouping-strategy.R
\name{grouping_by_label}
\alias{grouping_by_label}
\title{Label-based grouping strategy.}
\usage{
grouping_by_label()
}
\value{
A \code{grouping_strategy} object.
}
\description{
Creates a grouping strategy that routes all observations of a group to the
same child node. This is the default grouping strategy for classification.
}
\examples{
grouping_by_label()

}
\seealso{
\code{\link{pptr}}, \code{\link{pprf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cutpoint-strategy.R
\name{cutpoint_mean_of_means}
\alias{cutpoint_mean_of_means}
\title{Mean-of-means split cutpoint strategy.}
\usage{
cutpoint_mean_of_means()
}
\value{
A \code{cutpoint_strategy} object.
}
\description{
Creates a split cutpoint strategy that splits at the midpoint between group
means. This is the default (and currently only) split cutpoint.
}
\examples{
cutpoint_mean_of_means()

}
\seealso{
\code{\link{pptr}}, \code{\link{pprf}}, \code{\link{pp_pda}}, \code{\link{vars_uniform}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{NCI60}
\alias{NCI60}
\title{NCI60 Dataset}
\format{
A data frame with 61 rows and 31 variables
}
\source{
Dudoit, S., Fridlyand, J. and Speed, T. P. (2002). Comparison of Discrimination Methods for the Classification of Tumors Using Gene Expression Data. Journal of the American statistical Association 97 77-87.
}
\usage{
data(NCI60)
}
\description{
cDNA microarrays were used to examine the variation in gene expression among the 60 cell lines.  The cell lines are derived from tumors with different sites of origin. This dataset contain 61 observations and 30 feature variables from 8 different tissue types.
}
\details{
\describe{
  \item{Type}{ has 8 different tissue types, 9 cases of breast, 5 cases of central nervous system (CNS), 7 cases pf colon, 8 cases of leukemia, 8 cases of melanoma, 9 cases of  non-small-cell lung carcinoma (NSCLC), 6 cases of ovarian and 9 cases of renal.}
  \item{Gene1 to Gen 30}{ gene expression information}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{permuted_importance}
\alias{permuted_importance}
\title{Permuted variable importance for a random forest.}
\usage{
permuted_importance(model)
}
\arguments{
\item{model}{A \code{pprf} forest model.}
}
\value{
A numeric vector, one entry per feature. Negative values are
  meaningful (see Sign semantics above).
}
\description{
For each feature, measures the drop in OOB accuracy (classification) or
the increase in normalised MSE (regression) after randomly permuting
that feature across the OOB observations. Computed lazily from the
training data stored on the model; the result is cached.
}
\details{
**Sign semantics.** Entries may be **negative**. That is not an error
and not a sentinel: it means permuting the feature did not degrade OOB
fit on average — the feature's signal sits at or below the noise floor
of the permutation procedure. Interpret negative or near-zero entries
as "no evidence of importance"; rely on the ranking rather than
clipping at zero or normalizing. The scale is already comparable
within a fitted model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stop-strategy.R
\name{stop_min_variance}
\alias{stop_min_variance}
\title{Minimum-variance stopping rule.}
\usage{
stop_min_variance(threshold = 0.01)
}
\arguments{
\item{threshold}{Variance threshold below which to stop splitting (default: 0.01).}
}
\value{
A \code{stop_strategy} object.
}
\description{
Creates a stopping rule that stops splitting when the within-node response
variance falls below \code{threshold}. Used primarily for regression trees;
requires a continuous response.
}
\examples{
stop_min_variance(0.01)

}
\seealso{
\code{\link{stop_min_size}}, \code{\link{stop_any}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vars-strategy.R
\name{vars_all}
\alias{vars_all}
\title{All-variables selection strategy.}
\usage{
vars_all()
}
\value{
A \code{vars_strategy} object.
}
\description{
Creates a variable selection strategy that uses all variables at each
split. This is the default for single trees (\code{\link{pptr}}).
}
\examples{
vars_all()

}
\seealso{
\code{\link{pptr}}, \code{\link{vars_uniform}}, \code{\link{pp_pda}}, \code{\link{cutpoint_mean_of_means}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{california_housing}
\alias{california_housing}
\title{California Housing Dataset}
\format{
A data frame with 20,433 rows and 9 variables.
}
\source{
Pace, R.K. and Barry, R. (1997) Sparse spatial autoregressions. Statistics & Probability Letters, 33, 291-297.
}
\usage{
data(california_housing)
}
\description{
Median house values and demographic/geographic predictors for 20,433 California
block groups, derived from the 1990 US Census. Classic regression benchmark.
Originally released by Pace & Barry (1997); this version drops the categorical
`ocean_proximity` column (to keep the bundled schema all-numeric, matching the
rest of the package's datasets) and removes rows with missing `total_bedrooms`.
The regression target, `median_house_value`, is the last column.
}
\details{
\describe{
  \item{longitude}{Longitude of the block group (degrees, negative = west).}
  \item{latitude}{Latitude of the block group (degrees).}
  \item{housing_median_age}{Median age of houses in the block group (years).}
  \item{total_rooms}{Total number of rooms.}
  \item{total_bedrooms}{Total number of bedrooms.}
  \item{population}{Block group population.}
  \item{households}{Number of households.}
  \item{median_income}{Median household income (tens of thousands of USD).}
  \item{median_house_value}{Regression target: median house value (USD).}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptr.R
\name{pptr}
\alias{pptr}
\title{Trains a Projection-Pursuit oblique decision tree.}
\usage{
pptr(
  formula = NULL,
  data = NULL,
  x = NULL,
  y = NULL,
  mode = NULL,
  lambda = 0.5,
  seed = NULL,
  pp = NULL,
  cutpoint = NULL,
  stop = NULL,
  binarize = NULL,
  grouping = NULL,
  leaf = NULL
)
}
\arguments{
\item{formula}{A formula of the form \code{y ~ x1 + x2 + ...}, where \code{y} is a vector of labels and \code{x1}, \code{x2}, ... are the features.}

\item{data}{A data frame containing the variables in the formula.}

\item{x}{A matrix containing the features for each observation.}

\item{y}{A matrix containing the labels for each observation.}

\item{mode}{Training mode: either \code{"classification"} or \code{"regression"}. When \code{NULL} (default), mode is auto-detected from \code{y}'s type — factor or character vectors trigger classification, numeric vectors trigger regression. Setting it explicitly is useful for the binary-integer-labels case (\code{mode = "classification"} with integer 0/1 labels) and for failing fast on a type mismatch (\code{mode = "regression"} with a factor \code{y} errors immediately).}

\item{lambda}{A regularization parameter (default: 0.5). If \code{lambda = 0}, the model is trained using Linear Discriminant Analysis (LDA). If \code{lambda > 0}, the model is trained using Penalized Discriminant Analysis (PDA). The default uses PDA because pure LDA (\code{lambda = 0}) is ill-conditioned when there are more variables than effective observations (see the "Known limitations" section of the README). Cannot be used together with \code{pp}.}

\item{seed}{An optional integer seed for reproducibility. If \code{NULL} (default), a seed is drawn from R's RNG, so \code{set.seed()} controls reproducibility. If an integer is provided, that value is used directly.}

\item{pp}{A projection pursuit strategy object created by \code{\link{pp_pda}}. Cannot be used together with \code{lambda}.}

\item{cutpoint}{A split cutpoint strategy object created by \code{\link{cutpoint_mean_of_means}} (default).}

\item{stop}{A stopping rule object. Default depends on mode:
\code{\link{stop_pure_node}()} for classification, and
\code{stop_any(stop_min_size(5), stop_min_variance(0.01))} for regression.}

\item{binarize}{A binarization strategy object. Default depends on mode:
\code{\link{binarize_largest_gap}()} for classification, and
\code{\link{binarize_disabled}()} for regression (regression's default
grouping always yields a 2-group partition, so no binarization is needed).}

\item{grouping}{A grouping strategy object. Default depends on mode:
\code{\link{grouping_by_label}()} for classification, and
\code{\link{grouping_by_cutpoint}()} for regression.}

\item{leaf}{A leaf strategy object. Default depends on mode:
\code{\link{leaf_majority_vote}()} for classification, and
\code{\link{leaf_mean_response}()} for regression.}
}
\value{
A \code{pptr} model. Its S3 class vector is
  \code{c("pptr_classification", "pptr", "ppmodel")} or
  \code{c("pptr_regression", "pptr", "ppmodel")} depending on the mode.
}
\description{
This function trains a Projection-Pursuit oblique decision tree using either a formula and data frame interface or a matrix-based interface. When using the formula interface, specify the model formula and the data frame containing the variables. For the matrix-based interface, provide matrices for the features and labels directly.
If \code{lambda = 0}, the model is trained using Linear Discriminant Analysis (LDA). If \code{lambda > 0}, the model is trained using Penalized Discriminant Analysis (PDA).
}
\details{
Mode is taken from the \code{mode} argument when explicit, and otherwise auto-detected from `y` (factor/character → classification, numeric → regression). Pass \code{mode = "classification"} to force classification on integer labels (e.g. binary 0/1), or \code{mode = "regression"} to assert intent on numeric responses.
}
\examples{

# Example 1: formula interface with the `iris` dataset
pptr(Species ~ ., data = iris)

# Example 2: formula interface with the `iris` dataset with regularization
pptr(Species ~ ., data = iris, lambda = 0.5)

# Example 3: matrix interface with the `iris` dataset
pptr(x = iris[, 1:4], y = iris[, 5])

}
\seealso{
\code{\link{predict.pptr_classification}}, \code{\link{predict.pptr_regression}}, \code{\link{formula.ppmodel}}, \code{\link{print.pptr}}, \code{\link{save_json}}, \code{\link{load_json}}, \code{\link{pp_tree}} for parsnip integration
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptr.R
\name{print.pptr}
\alias{print.pptr}
\title{Prints the structure of a pptr tree.}
\usage{
\method{print}{pptr}(x, ...)
}
\arguments{
\item{x}{A \code{pptr} model.}

\item{...}{Unused.}
}
\value{
Invisibly returns the input \code{pptr} model \code{x} (unchanged).
  Called for its side effect of printing the tree structure -- the oblique
  split rules and leaf predictions -- to the console.
}
\description{
Prints the structure of a pptr tree.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{oob_error}
\alias{oob_error}
\title{Out-of-bag error for a random forest.}
\usage{
oob_error(model)
}
\arguments{
\item{model}{A \code{pprf} forest model.}
}
\value{
A numeric scalar in `[0, 1]` for classification or `[0, Inf)` for
  regression. Returns `NA_real_` when no observation has any out-of-bag
  tree (e.g. a degenerate forest where every tree saw every row). Callers
  should check with `is.na()` rather than comparing against a sentinel
  value; in earlier versions this condition was signalled as `-1`, which
  was not distinguishable from a (mathematically impossible but
  representable) error rate.
}
\description{
Computes (or returns the cached) OOB error using the training data stored
on the model. For classification, this is the misclassification rate in
`[0, 1]`. For regression, it is the mean squared error against the
continuous response.
}
\seealso{
\code{\link{oob_predictions}}, \code{\link{oob_samples}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-methods.R
\name{plot.pprf}
\alias{plot.pprf}
\title{Plot a pprf model.}
\usage{
\method{plot}{pprf}(x, type = "importance", metric = NULL, tree_index = 1L, node = 1L, ...)
}
\arguments{
\item{x}{A pprf model.}

\item{type}{Character string specifying the plot type.
\code{"importance"} (default) shows variable importance,
\code{"structure"} shows a specific tree with embedded histograms,
\code{"projection"} shows projected data at a node,
\code{"boundaries"} shows decision boundaries of a specific tree.}

\item{metric}{Character string selecting a single importance metric
to plot: \code{"projections"}, \code{"weighted"}, or
\code{"permuted"}.  \code{NULL} (default) shows all available
metrics side by side in separate panels.  Only used when
\code{type = "importance"}.}

\item{tree_index}{Integer index of the tree to plot (1-based). Only used when
\code{type = "structure"}, \code{type = "projection"}, or
\code{type = "boundaries"}. Defaults to 1.}

\item{node}{Integer index of the node for projection plots. Defaults to 1 (root).
Only used when \code{type = "projection"}.}

\item{...}{Additional arguments passed to the internal plotting function.}
}
\value{
A ggplot2-compatible object (invisibly). The importance grid returns
  a patchwork object that works with \code{ggplot2::ggsave()}.
}
\description{
Visualizes a pprf model. By default, shows variable importance with
one plot per metric side by side. Use \code{metric} to show a single
importance metric.
}
\examples{
\donttest{
if (requireNamespace("ggplot2", quietly = TRUE) &&
    requireNamespace("patchwork", quietly = TRUE)) {
  forest <- pprf(Species ~ ., data = iris, size = 10)
  plot(forest)                                    # all metrics side by side
  plot(forest, metric = "permuted")               # single metric
  plot(forest, type = "structure", tree_index = 1)
  plot(forest, type = "projection", tree_index = 1)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{fitted.ppmodel}
\alias{fitted.ppmodel}
\title{Fitted (in-sample) predictions from a ppforest2 model.}
\usage{
\method{fitted}{ppmodel}(object, ...)
}
\arguments{
\item{object}{A \code{pptr} or \code{pprf} model.}

\item{...}{Unused.}
}
\value{
A factor (classification) or numeric vector (regression), length
  equal to the number of training observations.
}
\description{
Returns predictions on the training data — a factor for classification,
a numeric vector for regression. For forests, training predictions are
optimistic; use \code{\link{oob_predictions}()} for an honest estimate.
}
\seealso{
\code{\link{residuals.ppmodel}}, \code{\link{oob_predictions}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pp-strategy.R
\name{pp_pda}
\alias{pp_pda}
\title{PDA projection pursuit strategy.}
\usage{
pp_pda(lambda = 0)
}
\arguments{
\item{lambda}{A regularization parameter between 0 and 1.
If \code{lambda = 0}, the model uses Linear Discriminant Analysis (LDA).
If \code{lambda > 0}, the model uses Penalized Discriminant Analysis (PDA).}
}
\value{
A \code{pp_strategy} object.
}
\description{
Creates a Penalized Discriminant Analysis (PDA) projection pursuit strategy
for use with \code{\link{pptr}} or \code{\link{pprf}}.
}
\examples{
# PDA with lambda = 0.5
pp_pda(0.5)

# Use with pptr
pptr(Species ~ ., data = iris, pp = pp_pda(0.5))

# Use with pprf
pprf(Species ~ ., data = iris, pp = pp_pda(0.5), vars = vars_uniform(n_vars = 2))

}
\seealso{
\code{\link{pptr}}, \code{\link{pprf}}, \code{\link{vars_uniform}}, \code{\link{vars_all}}, \code{\link{cutpoint_mean_of_means}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{weighted_importance}
\alias{weighted_importance}
\title{Weighted projection variable importance for a random forest.}
\usage{
weighted_importance(model)
}
\arguments{
\item{model}{A \code{pprf} forest model.}
}
\value{
A non-negative numeric vector, one entry per feature.
}
\description{
Weights each tree's projection-based importance by a per-tree OOB
quality score — `1 - error_rate` in `[0, 1]` for classification, and
`max(0, 1 - NMSE)` in `[0, 1]` for regression — then aggregates
`I_s × |a_j|` over splits. Computed lazily from the training data
stored on the model; the result is cached.
}
\details{
**Sign semantics.** Entries are non-negative by construction (weights
and per-split contributions are both non-negative). A zero entry means
"this feature never appeared in a weighted OOB-contributing split,"
not "within noise." Contrast with \code{\link{permuted_importance}},
where negative values are meaningful. Do not re-normalize — rely on
the ranking.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/binarize-strategy.R
\name{binarize_largest_gap}
\alias{binarize_largest_gap}
\title{Largest-gap binarization strategy.}
\usage{
binarize_largest_gap()
}
\value{
A \code{binarize_strategy} object.
}
\description{
Creates a binarization strategy that reduces multiclass nodes to binary
by projecting group means and splitting at the largest gap. Default
for classification specs.
}
\examples{
binarize_largest_gap()

}
\seealso{
\code{\link{pptr}}, \code{\link{pprf}}, \code{\link{binarize_disabled}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json.R
\name{load_json}
\alias{load_json}
\title{Load a model from a JSON file.}
\usage{
load_json(path)
}
\arguments{
\item{path}{File path to read the JSON from.}
}
\value{
A \code{pptr} or \code{pprf} model (with the appropriate
  \code{_classification} / \code{_regression} subclass).
}
\description{
Deserializes a \code{pptr} or \code{pprf} model from a JSON file. The model
can be used for prediction immediately. If variable importance metrics and
OOB fields were saved, they are restored into the model's lazy-computation
cache so later accessor calls (\code{oob_error()}, \code{permuted_importance()},
etc.) return the stored values without recomputation.
}
\details{
Note that \code{formula}, \code{x}, and \code{y} are not stored in the JSON
and will be \code{NULL} on the loaded model. Formula-based prediction and
any accessor that requires training data will error unless the user
re-attaches those fields.
}
\examples{
model <- pptr(Species ~ ., data = iris, seed = 0)
path <- tempfile(fileext = ".json")
save_json(model, path)
loaded <- load_json(path)
predict(loaded, as.matrix(iris[, 1:4]))

}
\seealso{
\code{\link{save_json}}, \code{\link{pptr}}, \code{\link{pprf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/binarize-strategy.R
\name{binarize_disabled}
\alias{binarize_disabled}
\title{Disabled binarization strategy (placeholder).}
\usage{
binarize_disabled()
}
\value{
A \code{binarize_strategy} object.
}
\description{
Placeholder binarizer for specs where binarization never fires —
notably regression, where `grouping_by_cutpoint()` always produces
a 2-group partition at each node. Selecting \code{binarize_disabled()}
documents that intent explicitly; if binarization is ever invoked at
runtime with this strategy configured, training aborts with a clear
error rather than silently passing through. Default for regression
specs.
}
\examples{
binarize_disabled()

}
\seealso{
\code{\link{binarize_largest_gap}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pprf.R
\name{predict.pprf_classification}
\alias{predict.pprf_classification}
\title{Predicts labels or vote proportions from a pprf model (classification mode).}
\usage{
\method{predict}{pprf_classification}(object, new_data = NULL, type = NULL, ...)
}
\arguments{
\item{object}{A \code{pprf_classification} model.}

\item{new_data}{A data frame or matrix of new observations. If \code{NULL}, the first positional argument in \code{...} is used for backward compatibility.}

\item{type}{The type of prediction: \code{"class"} (default) returns a factor of predicted labels, \code{"prob"} returns a data frame of vote proportions.}

\item{...}{For backward compatibility, the first positional argument is treated as \code{new_data} when \code{new_data} is \code{NULL}.}
}
\value{
If \code{type = "class"}, a factor of predicted labels. If \code{type = "prob"}, a data frame with one column per group, each row summing to 1.
}
\description{
Predicts labels or vote proportions from a pprf model (classification mode).
}
\examples{
model <- pprf(Species ~ ., data = iris)
predict(model, iris)
predict(model, iris, type = "prob")
}
\seealso{
\code{\link{pprf}}, \code{\link{predict.pprf_regression}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stop-strategy.R
\name{stop_any}
\alias{stop_any}
\title{Composite stopping rule (logical OR).}
\usage{
stop_any(...)
}
\arguments{
\item{...}{Two or more \code{stop_strategy} objects to combine.}
}
\value{
A \code{stop_strategy} object.
}
\description{
Creates a composite stopping rule that fires when any of the child rules
fires. Useful for combining multiple criteria, e.g. \code{stop_any(stop_min_size(5),
stop_min_variance(0.01))} for regression.
}
\examples{
stop_any(stop_min_size(5), stop_min_variance(0.01))

}
\seealso{
\code{\link{stop_min_size}}, \code{\link{stop_min_variance}}, \code{\link{stop_pure_node}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{crab}
\alias{crab}
\title{Australian Crabs Dataset}
\format{
A data frame with 200 rows and 6 variables
}
\source{
Campbell, N. A. & Mahon, R. J. (1974), A Multivariate Study of Variation in Two Species of Rock Crab of genus Leptograpsus, Australian Journal of Zoology 22(3), 417 - 425.
}
\usage{
data(crab)
}
\description{
Measurements on rock crabs of the genus Leptograpsus. The dataset contains 200 observations
 from two species of crab (blue and orange), there are 50 specimens of each sex of each species,
  collected on site at Fremantle, Western Australia.
\describe{
  \item{Type}{ is the group variable and has 4 groups with the combinations of species and sex (BlueMale, BlueFemale, OrangeMale and OrangeFemale)}.
  \item{FL}{the size of the frontal lobe length, in mm}
  \item{RW}{rear width, in mm}
  \item{CL}{length of midline of the carapace, in mm}
  \item{CW}{maximum width of carapace, in mm}
  \item{BD}{depth of the body; for females, measured after displacement of the abdomen, in mm}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{fishcatch}
\alias{fishcatch}
\title{Fish Catch Dataset}
\format{
A data frame with 159 rows and 7 variables
}
\source{
\url{https://jse.amstat.org/jse_data_archive.htm}
}
\usage{
data(fishcatch)
}
\description{
There are 159 fishes of 7 species are caught and measured. Altogether there are
7 variables.  All the fishes are caught from the same lake(Laengelmavesi) near Tampere in Finland.
\describe{
	 \item{Type}{ has 7 fish groups, with 35 cases of Bream, 11 cases of Parkki, 56 cases of Perch 17 cases of Pike, 20 cases of Roach, 14 cases of Smelt and 6 cases of Whitewish.}
	 \item{weight}{ Weight of the fish (in grams)}
	 \item{length1}{ Length from the nose to the beginning of the tail (in cm)}
	 \item{length2}{ Length from the nose to the notch of the tail (in cm)}
	 \item{length3}{ Length from the nose to the end of the tail (in cm)}
	 \item{height}{ Maximal height as \% of Length3}
	 \item{width}{ Maximal width as \% of Length3}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/leaf-strategy.R
\name{leaf_mean_response}
\alias{leaf_mean_response}
\title{Mean-response leaf strategy.}
\usage{
leaf_mean_response()
}
\value{
A \code{leaf_strategy} object.
}
\description{
Creates a leaf strategy that predicts the mean of the continuous response
values for the observations in the leaf. Used for regression trees; requires
a numeric response.
}
\examples{
leaf_mean_response()

}
\seealso{
\code{\link{pptr}}, \code{\link{pprf}}, \code{\link{leaf_majority_vote}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-methods.R
\name{plot.pptr}
\alias{plot.pptr}
\title{Plot a pptr model.}
\usage{
\method{plot}{pptr}(x, type = NULL, metric = NULL, node = 1L, ...)
}
\arguments{
\item{x}{A pptr model.}

\item{type}{Character string specifying the plot type. \code{NULL} (default)
shows a mosaic overview. Other options: \code{"structure"} for tree with
embedded histograms, \code{"importance"} for variable importance,
\code{"projection"} for projected data at a node,
\code{"boundaries"} for decision boundaries in feature space.}

\item{metric}{Character string selecting a single importance metric
to plot: \code{"projections"}, \code{"weighted"}, or
\code{"permuted"} (availability depends on the model).  \code{NULL}
(default) shows all available metrics together.  Only used when
\code{type = "importance"}.}

\item{node}{Integer index of the node for projection plots (1-based, breadth-first
order). Defaults to 1 (root node). Only used when \code{type = "projection"}.}

\item{...}{Additional arguments passed to the internal plotting function.}
}
\value{
A ggplot2-compatible object (invisibly). The mosaic layout returns
  a patchwork object that works with \code{ggplot2::ggsave()}.
}
\description{
Visualizes a pptr model. By default, shows a mosaic overview with tree
structure, decision boundaries, and variable importance. Use \code{type}
to show individual plots.
}
\examples{
\donttest{
if (requireNamespace("ggplot2", quietly = TRUE) &&
    requireNamespace("patchwork", quietly = TRUE)) {
  model <- pptr(Species ~ ., data = iris)
  plot(model)                         # mosaic overview
  plot(model, type = "structure")     # tree structure only
  plot(model, type = "importance")    # variable importance
  plot(model, type = "projection")    # projection histogram
  plot(model, type = "boundaries")    # decision boundaries
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsnip.R
\name{update.pp_tree}
\alias{update.pp_tree}
\title{Update a \code{pp_tree} model specification.}
\usage{
\method{update}{pp_tree}(object, parameters = NULL, penalty = NULL, fresh = FALSE, ...)
}
\arguments{
\item{object}{A \code{pp_tree} model specification.}

\item{parameters}{A named list of parameters to update (alternative to passing them as args).}

\item{penalty}{A new value for the regularisation parameter.}

\item{fresh}{If \code{TRUE}, drop any previously-set arg whose new value is not provided.}

\item{...}{Engine-specific arguments to update.}
}
\value{
An updated \code{pp_tree} model specification.
}
\description{
Companion to \code{\link{update.pp_rand_forest}} — exists for the same
reason (\code{tune::tune_grid()} relies on \code{update()} to finalise the
spec at each grid point).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grouping-strategy.R
\name{grouping_by_cutpoint}
\alias{grouping_by_cutpoint}
\title{Cutpoint-based grouping strategy (regression).}
\usage{
grouping_by_cutpoint()
}
\value{
A \code{grouping_strategy} object.
}
\description{
Creates a grouping strategy for regression trees: observations are split by
the cutpoint in projected space, then each child's observations are sorted
by the continuous response and median-split into 2 new groups for the next
projection-pursuit step. This is the default grouping strategy for regression.
}
\examples{
grouping_by_cutpoint()

}
\seealso{
\code{\link{pptr}}, \code{\link{pprf}}, \code{\link{grouping_by_label}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{olive}
\alias{olive}
\title{Olive Dataset}
\format{
A data frame contains 573 observations and 10 variables
}
\usage{
data(olive)
}
\description{
Contains  572 observations and 10 variables
\describe{
  \item{Region}{Three super-groups of Italy: North, South and the island of Sardinia }
  \item{area}{ Nine collection areas: three from North, four from South and 2 from Sardinia}
  \item{palmitic}{ fatty acids percent x 100}
  \item{palmitoleic}{ fatty acids percent x 100 }
  \item{stearic}{fatty acids percent x 100}
  \item{oleic}{fatty acids percent x 100 }
  \item{linoleic}{fatty acids percent x 100}
  \item{linolenic}{fatty acids percent x 100  }
  \item{arachidic}{ fatty acids percent x 100}
  \item{eicosenoic}{ fatty acids percent x 100}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsnip.R
\name{pp_tree}
\alias{pp_tree}
\title{Parsnip model specification for pptr.}
\usage{
pp_tree(mode = "classification", penalty = NULL)
}
\arguments{
\item{mode}{A character string for the model type. Either \code{"classification"} or \code{"regression"}.}

\item{penalty}{The regularization parameter (maps to \code{lambda}).}
}
\value{
A parsnip model specification.
}
\description{
Creates a model specification for a single Projection Pursuit decision tree.
Use \code{set_engine("ppforest2")} to select the ppforest2 engine.
}
\examples{
\donttest{
if (requireNamespace("parsnip", quietly = TRUE)) {
  library(parsnip)
  spec <- pp_tree(penalty = 0) \%>\% set_engine("ppforest2")
  fit <- spec \%>\% fit(Species ~ ., data = iris)
  predict(fit, iris)
}
}
}
\seealso{
\code{\link{pptr}} for the underlying training function, \code{\link{pp_rand_forest}} for forests
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vars-strategy.R
\name{vars_uniform}
\alias{vars_uniform}
\title{Uniform random variable selection strategy.}
\usage{
vars_uniform(n_vars = NULL, p_vars = NULL)
}
\arguments{
\item{n_vars}{The number of variables to consider at each split (integer).
Cannot be used together with \code{p_vars}.}

\item{p_vars}{The proportion of variables to consider at each split
(number between 0 and 1, exclusive). Resolved to an integer count when
the number of features is known. Cannot be used together with \code{n_vars}.}
}
\value{
A \code{vars_strategy} object.
}
\description{
Creates a variable selection strategy that randomly selects a subset
of variables at each split. Used with \code{\link{pprf}} for random forests.
}
\details{
Exactly one of \code{n_vars} or \code{p_vars} may be specified. When
\code{p_vars} is used, it is stored as-is and resolved to an integer count
later by \code{\link{pprf}} once the number of features is known.
}
\examples{
# Select 2 variables at each split
vars_uniform(n_vars = 2)

# Select half the variables at each split
vars_uniform(p_vars = 0.5)

# Use with pprf
pprf(Species ~ ., data = iris, vars = vars_uniform(n_vars = 2))

}
\seealso{
\code{\link{pprf}}, \code{\link{vars_all}}, \code{\link{pp_pda}}, \code{\link{cutpoint_mean_of_means}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{residuals.ppmodel}
\alias{residuals.ppmodel}
\title{Residuals from a regression ppforest2 model.}
\usage{
\method{residuals}{ppmodel}(object, ...)
}
\arguments{
\item{object}{A \code{pprf} or \code{pptr} regression model.}

\item{...}{Unused.}
}
\value{
A numeric vector of length \code{nobs(object)}.
}
\description{
Returns \code{y - fitted(model)}. Only defined for regression models —
classification residuals have no canonical scalar form, so this method
errors on classification models rather than inventing a convention.
}
\seealso{
\code{\link{fitted.ppmodel}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{formula.ppmodel}
\alias{formula.ppmodel}
\title{Formula extractor for ppforest2 models.}
\usage{
\method{formula}{ppmodel}(x, ...)
}
\arguments{
\item{x}{A \code{pptr} or \code{pprf} model.}

\item{...}{Unused.}
}
\value{
The formula the model was trained with, or `NULL` for matrix-interface fits.
}
\description{
Formula extractor for ppforest2 models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{bag_samples}
\alias{bag_samples}
\title{In-bag row indices per tree.}
\usage{
bag_samples(model)
}
\arguments{
\item{model}{A \code{pprf} forest model.}
}
\value{
A list of integer vectors, one per tree.
}
\description{
Returns a list where element `i` is the integer vector of row indices
(1-based, with replacement) drawn into the bootstrap sample of tree `i`.
}
\seealso{
\code{\link{oob_samples}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pprf.R
\name{summary.pprf}
\alias{summary.pprf}
\title{Summary of a pprf forest (shared header + VI).}
\usage{
\method{summary}{pprf}(object, ...)
}
\arguments{
\item{object}{A \code{pprf} model.}

\item{...}{Unused.}
}
\value{
Invisibly returns the input \code{pprf} model \code{object}
  (unchanged). Called for its side effect of printing a detailed summary --
  the training specification, data summary, and variable-importance table
  (plus, for classification, the training/OOB confusion matrices) -- to the
  console.
}
\description{
Summary of a pprf forest (shared header + VI).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptr.R
\name{predict.pptr_classification}
\alias{predict.pptr_classification}
\title{Predicts labels or per-group one-hot proportions from a pptr model (classification mode).}
\usage{
\method{predict}{pptr_classification}(object, new_data = NULL, type = NULL, ...)
}
\arguments{
\item{object}{A \code{pptr_classification} model.}

\item{new_data}{A data frame or matrix of new observations. If \code{NULL}, the first positional argument in \code{...} is used for backward compatibility.}

\item{type}{\code{"class"} (default) returns a factor of predicted labels; \code{"prob"} returns a data frame with 1.0 for the predicted group and 0.0 elsewhere.}

\item{...}{Backward-compat positional `new_data`.}
}
\value{
A factor or data frame.
}
\description{
Predicts labels or per-group one-hot proportions from a pptr model (classification mode).
}
\seealso{
\code{\link{pptr}}, \code{\link{predict.pptr_regression}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pprf.R
\name{pprf}
\alias{pprf}
\title{Trains a Random Forest of Projection-Pursuit oblique decision trees.}
\usage{
pprf(
  formula = NULL,
  data = NULL,
  x = NULL,
  y = NULL,
  mode = NULL,
  size = 100,
  lambda = 0.5,
  n_vars = NULL,
  p_vars = NULL,
  seed = NULL,
  max_retries = 3L,
  threads = NULL,
  pp = NULL,
  vars = NULL,
  cutpoint = NULL,
  stop = NULL,
  binarize = NULL,
  grouping = NULL,
  leaf = NULL
)
}
\arguments{
\item{formula}{A formula of the form \code{y ~ x1 + x2 + ...}, where \code{y} is a vector of labels and \code{x1}, \code{x2}, ... are the features.}

\item{data}{A data frame containing the variables in the formula.}

\item{x}{A matrix containing the features for each observation.}

\item{y}{A matrix containing the labels for each observation.}

\item{mode}{Training mode: either \code{"classification"} or \code{"regression"}. When \code{NULL} (default), mode is auto-detected from \code{y}'s type — factor or character vectors trigger classification, numeric vectors trigger regression. Setting it explicitly is useful for the binary-integer-labels case (\code{mode = "classification"} with integer 0/1 labels) and for failing fast on a type mismatch (\code{mode = "regression"} with a factor \code{y} errors immediately).}

\item{size}{The number of trees in the forest (default: 100).}

\item{lambda}{A regularization parameter (default: 0.5). If \code{lambda = 0}, the model is trained using Linear Discriminant Analysis (LDA). If \code{lambda > 0}, the model is trained using Penalized Discriminant Analysis (PDA). The default uses PDA because pure LDA (\code{lambda = 0}) is ill-conditioned when there are more variables than effective observations (see the "Known limitations" section of the README). Cannot be used together with \code{pp}.}

\item{n_vars}{The number of variables to consider at each split (integer). These are chosen uniformly in each split. By default, half of the variables are used (\code{p_vars = 0.5}). Cannot be used together with \code{p_vars} or \code{dr}.}

\item{p_vars}{The proportion of variables to consider at each split (number between 0 and 1, exclusive). For example, \code{p_vars = 0.5} uses half the features. Cannot be used together with \code{n_vars} or \code{dr}.}

\item{seed}{An optional integer seed for reproducibility. If \code{NULL} (default), a seed is drawn from R's RNG, so \code{set.seed()} controls reproducibility. If an integer is provided, that value is used directly. The same seed is used for training and for computing permuted variable importance.}

\item{max_retries}{Maximum number of retries for degenerate trees (default: 3). When a bootstrap sample yields a singular covariance matrix, the tree is retrained with a different seed up to this many times.}

\item{threads}{The number of threads to use. The default is the number of cores available.}

\item{pp}{A projection pursuit strategy object created by \code{\link{pp_pda}}. Cannot be used together with \code{lambda}.}

\item{vars}{A variable selection strategy object created by \code{\link{vars_uniform}} or \code{\link{vars_all}}. Cannot be used together with \code{n_vars} or \code{p_vars}.}

\item{cutpoint}{A split cutpoint strategy object created by \code{\link{cutpoint_mean_of_means}} (default).}

\item{stop}{A stopping rule object. Default depends on mode:
\code{\link{stop_pure_node}()} for classification, and
\code{stop_any(stop_min_size(5), stop_min_variance(0.01))} for regression.}

\item{binarize}{A binarization strategy object. Default depends on mode:
\code{\link{binarize_largest_gap}()} for classification, and
\code{\link{binarize_disabled}()} for regression (regression's default
grouping always yields a 2-group partition, so no binarization is needed).}

\item{grouping}{A grouping strategy object. Default depends on mode:
\code{\link{grouping_by_label}()} for classification, and
\code{\link{grouping_by_cutpoint}()} for regression.}

\item{leaf}{A leaf strategy object. Default depends on mode:
\code{\link{leaf_majority_vote}()} for classification, and
\code{\link{leaf_mean_response}()} for regression.}
}
\value{
A \code{pprf} model. Its S3 class vector is
  \code{c("pprf_classification", "pprf", "ppmodel")} or
  \code{c("pprf_regression", "pprf", "ppmodel")} depending on the mode.
}
\description{
This function trains a Random Forest of Projection-Pursuit oblique decision tree using either a formula and data frame interface or a matrix-based interface. When using the formula interface, specify the model formula and the data frame containing the variables. For the matrix-based interface, provide matrices for the features and labels directly.
The number of trees is controlled by the \code{size} parameter. Each tree is trained on a stratified bootstrap sample drawn from the data.
The number of variables to consider at each split is controlled by the \code{n_vars} parameter.
If \code{lambda = 0}, the model is trained using Linear Discriminant Analysis (LDA). If \code{lambda > 0}, the model is trained using Penalized Discriminant Analysis (PDA).
}
\details{
Mode is taken from the \code{mode} argument when explicit, and otherwise auto-detected from `y` (factor/character → classification, numeric → regression). Pass \code{mode = "classification"} to force classification on integer labels (e.g. binary 0/1), or \code{mode = "regression"} to assert intent on numeric responses.

OOB error, OOB predictions, permuted variable importance, and weighted variable importance are computed lazily on first access via the accessor functions (`oob_error()`, `oob_predictions()`, `permuted_importance()`, `weighted_importance()`). Training itself is fast because these OOB-based computations are deferred.
}
\examples{

# Example 1: formula interface with the `iris` dataset
pprf(Species ~ ., data = iris)

# Example 2: formula interface with the `iris` dataset with regularization
pprf(Species ~ ., data = iris, lambda = 0.5)

# Example 3: matrix interface with the `iris` dataset
pprf(x = iris[, 1:4], y = iris[, 5])

# Example 4: matrix interface with the `iris` dataset with regularization
pprf(x = iris[, 1:4], y = iris[, 5], lambda = 0.5)

# Example 5: formula interface with the `crabs` dataset
pprf(Type ~ ., data = crabs)

# Example 6: formula interface with the `crabs` dataset with regularization
pprf(Type ~ ., data = crabs, lambda = 0.5)

}
\seealso{
\code{\link{predict.pprf_classification}}, \code{\link{predict.pprf_regression}}, \code{\link{formula.ppmodel}}, \code{\link{oob_error}}, \code{\link{save_json}}, \code{\link{load_json}}, \code{\link{pp_rand_forest}} for parsnip integration, \code{vignette("introduction")} for a tutorial
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json.R
\name{save_json}
\alias{save_json}
\alias{save_json.ppmodel}
\title{Save a model to a JSON file.}
\usage{
save_json(model, path, ...)

\method{save_json}{ppmodel}(model, path, include_metrics = TRUE, ...)
}
\arguments{
\item{model}{A \code{pptr} or \code{pprf} model (inheriting from \code{ppmodel}).}

\item{path}{File path to write the JSON to.}

\item{...}{Additional arguments (currently unused).}

\item{include_metrics}{If \code{TRUE} (default), include variable importance
and (for forests) OOB error in the output. Set to \code{FALSE} to save
only the model structure and metadata.}
}
\value{
Invisibly returns the file \code{path} that was written (a length-one
  character string). Called for its side effect of serializing \code{model}
  to a JSON file at \code{path}.
}
\description{
Serializes a \code{pptr} or \code{pprf} model to JSON format compatible
with the C++ CLI. The JSON includes the model structure, group labels,
training parameters, and optionally variable importance metrics.
}
\examples{
model <- pptr(Species ~ ., data = iris, seed = 0)
path <- tempfile(fileext = ".json")
save_json(model, path)

}
\seealso{
\code{\link{load_json}}, \code{\link{pptr}}, \code{\link{pprf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{wine}
\alias{wine}
\title{Wine Dataset}
\format{
A data frame with 178 rows and 14 variables
}
\usage{
data(wine)
}
\description{
A dataset containing 178 observations from 3 wine grown cultivares in Italy.
}
\details{
\describe{
  \item{Type}{ Group variable has 3 groups that are 3 different wine grown cultivars in Italy. }
  \item{X1 to X13}{Check vbles}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsnip.R
\name{update.pp_rand_forest}
\alias{update.pp_rand_forest}
\title{Update a \code{pp_rand_forest} model specification.}
\usage{
\method{update}{pp_rand_forest}(
  object,
  parameters = NULL,
  trees = NULL,
  mtry = NULL,
  mtry_prop = NULL,
  penalty = NULL,
  fresh = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{pp_rand_forest} model specification.}

\item{parameters}{A named list of parameters to update (alternative to passing them as args).}

\item{trees, mtry, mtry_prop, penalty}{New values for the corresponding parameters.}

\item{fresh}{If \code{TRUE}, drop any previously-set arg whose new value is not provided.}

\item{...}{Engine-specific arguments to update.}
}
\value{
An updated \code{pp_rand_forest} model specification.
}
\description{
Implements parsnip's \code{update} protocol so that \code{tune::tune_grid()}
(and any other caller that finalises a spec via \code{update()}) can fill
in the tuned values for \code{trees}, \code{mtry}, and \code{penalty}.
Without this method, \code{update()} falls back to \code{stats::update.default}
and fails with "need an object with call component".
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{image}
\alias{image}
\title{Image Dataset}
\format{
A data frame contains 2310 observations and 19 variables
}
\usage{
data(image)
}
\description{
Contains  2310 observations of instances from 7 outdoor images
\describe{
  \item{Type}{ has 7 types of outdoor images, brickface, cement,  foliage, grass, path, sky, and window.}
  \item{X1}{ the column of the center pixel of the region}
  \item{X2}{ the row of the center pixel of the region. }
  \item{X3}{ the number of pixels in a region = 9. }
  \item{X4}{ the results of a line extraction algorithm that counts how many lines of length 5 (any orientation) with low contrast, less than or equal to 5, go through the region.}
  \item{X5}{ measure the contrast of horizontally adjacent pixels in the region. There are 6, the mean and standard deviation are given. This attribute is used as a vertical edge detector.}
  \item{X6}{ X5 sd}
  \item{X7}{ measures the contrast of vertically adjacent pixels. Used for horizontal line detection. }
  \item{X8}{ sd X7}
  \item{X9}{ the average over the region of (R + G + B)/3}
  \item{X10}{ the average over the region of the R value.}
  \item{X11}{ the average over the region of the B value.}
  \item{X12}{ the average over the region of the G value.}
  \item{X13}{ measure the excess red: (2R - (G + B))}
  \item{X14}{ measure the excess blue: (2B - (G + R))}
  \item{X15}{ measure the excess green: (2G - (R + B))}
  \item{X16}{ 3-d nonlinear transformation of RGB. (Algorithm can be found in Foley and VanDam, Fundamentals of Interactive Computer Graphics)}
  \item{X17}{ mean of X16}
  \item{X18}{ hue  mean}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptr.R
\name{predict.pptr_regression}
\alias{predict.pptr_regression}
\title{Predicts numeric responses from a pptr model (regression mode).}
\usage{
\method{predict}{pptr_regression}(object, new_data = NULL, type = NULL, ...)
}
\arguments{
\item{object}{A \code{pptr_regression} model.}

\item{new_data}{A data frame or matrix of new observations.}

\item{type}{Must be \code{"response"} (default).}

\item{...}{Backward-compat positional `new_data`.}
}
\value{
A numeric vector.
}
\description{
Predicts numeric responses from a pptr model (regression mode).
}
\seealso{
\code{\link{pptr}}, \code{\link{predict.pptr_classification}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{parkinson}
\alias{parkinson}
\title{Parkinson Dataset}
\format{
A data frame with 195 rows and 23 variables
}
\source{
\url{https://archive.ics.uci.edu/ml/datasets/Parkinsons}
}
\usage{
data(parkinson)
}
\description{
A dataset containing 195 observations from 2 parkinson types.
\describe{
 \item{Type}{ Group variable has 2 groups, there are 48 cases of healthy people and 147 cases with Parkinson. The feature variables are biomedical voice measures}.
  \item{X1}{ Average vocal fundamental frequency}
  \item{X2}{ Maximum vocal fundamental frequency}
  \item{X3}{ Minimum vocal fundamental frequency}
  \item{X4}{ MDVP:Jitter(\%) measures of variation in fundamental frequency}
  \item{X5}{ MDVP:Jitter(Abs) measures of variation in fundamental frequency}
  \item{X6}{ MDVP:RAP measures of variation in fundamental frequency}
  \item{X7}{ MDVP:PPQ measures of variation in fundamental frequency}
  \item{X8}{ Jitter:DDP measures of variation in fundamental frequency}
  \item{X9}{ MDVP:Shimmer measures of variation in amplitude}
  \item{X10}{ MDVP:Shimmer(dB) measures of variation in amplitude}
  \item{X11}{ Shimmer:APQ3 measures of variation in amplitude}
  \item{X12}{ Shimmer:APQ5 measures of variation in amplitude}
  \item{X13}{ MDVP:APQ measures of variation in amplitude}
  \item{X14}{ Shimmer:DDA measures of variation in amplitude}
  \item{X15}{ NHR measures of ratio of noise to tonal components in the voice}
  \item{X16}{ HNR measures of ratio of noise to tonal components in the voice}
  \item{X17}{ RPDE nonlinear dynamical complexity measures}
  \item{X18}{ D2 nonlinear dynamical complexity measures}
  \item{X19}{ DFA - Signal fractal scaling exponent}
  \item{X20}{ spread1 Nonlinear measures of fundamental frequency variation}
  \item{X21}{ spread2 Nonlinear measures of fundamental frequency variation}
  \item{X22}{ PPE Nonlinear measures of fundamental frequency variation}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{nobs.ppmodel}
\alias{nobs.ppmodel}
\title{Number of observations used to fit a ppforest2 model.}
\usage{
\method{nobs}{ppmodel}(object, ...)
}
\arguments{
\item{object}{A \code{pptr} or \code{pprf} model.}

\item{...}{Unused.}
}
\value{
Integer scalar. Returns \code{NA_integer_} for models loaded from
  JSON without their original training data.
}
\description{
Implements the standard \code{stats::nobs()} contract so downstream tools
(\code{step()}, broom's \code{glance()}, information criteria) can ask for
the training-sample size.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/leaf-strategy.R
\name{leaf_majority_vote}
\alias{leaf_majority_vote}
\title{Majority-vote leaf strategy.}
\usage{
leaf_majority_vote()
}
\value{
A \code{leaf_strategy} object.
}
\description{
Creates a leaf strategy that assigns the majority group label as the
leaf prediction. When groups are tied, the smallest label wins.
This is the default leaf strategy for classification.
}
\examples{
leaf_majority_vote()

}
\seealso{
\code{\link{pptr}}, \code{\link{pprf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{crabs}
\alias{crabs}
\title{Crabs Dataset}
\format{
A data frame with 200 rows and 8 variables:
\describe{
  \item{Type}{Species code, with `B` indicating blue and `O` indicating orange.}
  \item{sex}{Sex of the crab, with `M` indicating male and `F` indicating female.}
  \item{FL}{Size of the frontal lobe in mm.}
  \item{RW}{Rear width in mm.}
  \item{CL}{Carapace length in mm.}
  \item{CW}{Carapace width in mm.}
  \item{BD}{Body depth in mm.}
}
}
\source{
Campbell, N.A., and Mahon, R.J. (1974) A multivariate study of variation in two species of rock crab of genus Leptograpsus. Australian Journal of Zoology, 22, 417-425.
}
\description{
Measurements of 200 crabs from the genus Leptograpsus, with 50 crabs for each of two colour forms and both sexes. Measurements include the frontal lobe size, rear width, carapace length, carapace width, and body depth.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stop-strategy.R
\name{stop_max_depth}
\alias{stop_max_depth}
\title{Maximum-depth stopping rule.}
\usage{
stop_max_depth(max_depth)
}
\arguments{
\item{max_depth}{Maximum depth (non-negative integer; 0 produces a
root-only stump).}
}
\value{
A \code{stop_strategy} object.
}
\description{
Creates a stopping rule that stops splitting when a node's depth reaches
\code{max_depth}. Depth is zero-based at the root, so \code{max_depth(k)}
allows at most \code{k + 1} levels. Mode-agnostic: useful for bounding
tree complexity in both classification and regression trees.
}
\examples{
stop_max_depth(5)

}
\seealso{
\code{\link{stop_min_size}}, \code{\link{stop_any}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{lymphoma}
\alias{lymphoma}
\title{Lymphoma Dataset}
\format{
A data frame with 80 rows and 51 variables
}
\source{
Dudoit, S., Fridlyand, J. and Speed, T. P. (2002). Comparison of Discrimination Methods for the Classification of Tumors Using Gene Ex- pression Data. Journal of the American statistical Association 97 77-87.
}
\usage{
data(lymphoma)
}
\description{
Gene expression in the three most prevalent adult lymphoid malignancies: B-cell chronic lymphocytic leukemia (B-CLL), follicular lymphoma (FL), and diffuse large B-cell lym- phoma (DLBCL). Gene expression levels were measured using a specialized cDNA microarray, the Lymphochip, containing genes that are preferentially expressed in lymphoid cells or that are of known immunologic or oncologic importance. This dataset contain 80 observations from 3 lymphoma types.
\describe{
  \item{Type}{ Group variable has 3 groups with 29 cases of B-cell ALL (B-CLL), 42 cases of diffuse large B-cell lymphoma (DLBCL) and 9 cases of follicular lymphoma (FL)}.
  \item{Gene1 to Gen 50}{gene expression}
  }
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsnip.R
\name{pp_rand_forest}
\alias{pp_rand_forest}
\title{Parsnip model specification for pprf.}
\usage{
pp_rand_forest(
  mode = "classification",
  trees = NULL,
  mtry = NULL,
  mtry_prop = NULL,
  penalty = NULL
)
}
\arguments{
\item{mode}{A character string for the model type. Either \code{"classification"} or \code{"regression"}.}

\item{trees}{The number of trees in the forest (maps to \code{size}).}

\item{mtry}{The number of variables to consider at each split (maps to \code{n_vars}).}

\item{mtry_prop}{The proportion of variables to consider at each split (maps to \code{p_vars}). An alternative to \code{mtry} that expresses the feature subsample as a fraction in (0, 1], tunable via the \code{dials} \code{mtry_prop()} parameter. Supply \code{mtry} or \code{mtry_prop}, not both.}

\item{penalty}{The regularization parameter (maps to \code{lambda}).}
}
\value{
A parsnip model specification.
}
\description{
Creates a model specification for a Projection Pursuit random forest.
Use \code{set_engine("ppforest2")} to select the ppforest2 engine.
}
\examples{
\donttest{
if (requireNamespace("parsnip", quietly = TRUE)) {
  library(parsnip)
  spec <- pp_rand_forest(trees = 50, mtry = 2) \%>\% set_engine("ppforest2")
  fit <- spec \%>\% fit(Species ~ ., data = iris)
  predict(fit, iris)
  predict(fit, iris, type = "prob")
}
}
}
\seealso{
\code{\link{pprf}} for the underlying training function, \code{\link{pp_tree}} for single trees
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stop-strategy.R
\name{stop_pure_node}
\alias{stop_pure_node}
\title{Pure-node stopping rule.}
\usage{
stop_pure_node()
}
\value{
A \code{stop_strategy} object.
}
\description{
Creates a stopping rule that stops splitting when a node contains only one
group. This is the default stopping rule for classification.
}
\examples{
stop_pure_node()

}
\seealso{
\code{\link{pptr}}, \code{\link{pprf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{leukemia}
\alias{leukemia}
\title{Leukemia Dataset}
\format{
A data frame with 72 rows and 41 variables
}
\source{
Dudoit, S., Fridlyand, J. and Speed, T. P. (2002). Comparison of Discrimination Methods for the Classification of Tumors Using Gene Expression Data. Journal of the American statistical Association 97 77-87.
}
\usage{
data(leukemia)
}
\description{
This dataset comes from a study of gene expression in two types of acute leukemias, acute lymphoblastic leukemia (ALL) and acute myeloid leukemia (AML). Gene expression levels were measured using Affymetrix high density oligonucleotide arrays containing 6817 human genes. A dataset containing 72 observations from 3 leukemia type groups.
\describe{
  \item{Type}{ has 3 groups with 38 cases of B-cell ALL, 25 cases of AML and 9 cases of T-cell ALL}.
  \item{Gene1 to Gen 40}{ gene expression levels}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{oob_samples}
\alias{oob_samples}
\title{Out-of-bag row indices per tree.}
\usage{
oob_samples(model)
}
\arguments{
\item{model}{A \code{pprf} forest model.}
}
\value{
A list of integer vectors, one per tree.
}
\description{
Returns a list where element `i` is the integer vector of row indices
(1-based) that were **not** in the bootstrap sample of tree `i`.
}
\seealso{
\code{\link{bag_samples}}, \code{\link{oob_error}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pprf.R
\name{print.pprf}
\alias{print.pprf}
\title{Prints a compact summary of a pprf forest.}
\usage{
\method{print}{pprf}(x, ...)
}
\arguments{
\item{x}{A \code{pprf} model.}

\item{...}{Unused.}
}
\value{
Invisibly returns the input \code{pprf} model \code{x} (unchanged).
  Called for its side effect of printing a compact description of the forest
  (call, number of trees, mode, group names, formula) to the console.
}
\description{
Prints a compact summary of a pprf forest.
}
\seealso{
\code{\link{summary.pprf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{oob_predictions}
\alias{oob_predictions}
\title{Out-of-bag predictions for a random forest.}
\usage{
oob_predictions(model)
}
\arguments{
\item{model}{A \code{pprf} forest model.}
}
\value{
A factor (classification) or numeric vector (regression), length `n`.
}
\description{
Returns predictions for each training row using only trees that did
not see that row in their bootstrap sample. Observations with no OOB
tree are represented as `NA` in both modes: `NA` at the factor level
for classification, and `NA_real_` for regression. Filter with the
standard `is.na()` idiom.
}
\seealso{
\code{\link{oob_error}}, \code{\link{oob_samples}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stop-strategy.R
\name{stop_min_size}
\alias{stop_min_size}
\title{Minimum-size stopping rule.}
\usage{
stop_min_size(min_size = 5L)
}
\arguments{
\item{min_size}{Minimum node size to allow a split (default: 5).}
}
\value{
A \code{stop_strategy} object.
}
\description{
Creates a stopping rule that stops splitting when a node has fewer than
\code{min_size} observations. Used primarily for regression trees.
}
\examples{
stop_min_size(5)

}
\seealso{
\code{\link{stop_min_variance}}, \code{\link{stop_any}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pprf.R
\name{predict.pprf_regression}
\alias{predict.pprf_regression}
\title{Predicts numeric responses from a pprf model (regression mode).}
\usage{
\method{predict}{pprf_regression}(object, new_data = NULL, type = NULL, ...)
}
\arguments{
\item{object}{A \code{pprf_regression} model.}

\item{new_data}{A data frame or matrix of new observations.}

\item{type}{Must be \code{"response"} (default).}

\item{...}{For backward compatibility, the first positional argument is treated as \code{new_data} when \code{new_data} is \code{NULL}.}
}
\value{
A numeric vector of mean predictions across the forest's trees.
}
\description{
Predicts numeric responses from a pprf model (regression mode).
}
\seealso{
\code{\link{pprf}}, \code{\link{predict.pprf_classification}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppmodel.R
\name{projection_importance}
\alias{projection_importance}
\title{Projection-coefficient variable importance.}
\usage{
projection_importance(model)
}
\arguments{
\item{model}{A \code{pptr} or \code{pprf} model.}
}
\value{
A non-negative numeric vector, one entry per feature.
}
\description{
The projection-based importance (VI2): each split's scaled absolute
projection coefficients (`|a_j| * sigma_j`) aggregated into a per-feature
score, averaged over the non-degenerate trees of a forest. Unlike
\code{\link{permuted_importance}} and \code{\link{weighted_importance}},
this measure is not OOB-based — it depends only on the fitted projector
geometry, is computed eagerly at fit time (cheap), and is available for
both single trees (\code{pptr}) and forests (\code{pprf}).
}
\details{
**Sign semantics.** Entries are non-negative by construction (absolute
coefficients scaled by each feature's standard deviation). Rely on the
ranking rather than re-normalizing.
}
\seealso{
\code{\link{permuted_importance}}, \code{\link{weighted_importance}}
}
